(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Omega.

Require Import rel_utils.
Require Import list_utils.
Require Import af.

(* Hence Ramsey can either be viewed by the preservation of af by intersection or by product *)

Section af_inter_from_product.

  Variable (X : Type) (R S : X -> X -> Prop).

  Theorem af_inter_from_product : af R -> af S -> af (fun x y => R x y /\ S x y).
  Proof.
    intros H1 H2.
    set (f := fun x : X => (x,x)).
    generalize (af_cofmap f (af_product H1 H2)).
    apply af_inc; simpl; auto.
  Qed.

End af_inter_from_product.

Section af_lexico.

  Definition flex : nat * nat -> nat.
  Proof.
    pose (T x y := fst x < fst y \/ (fst x = fst y /\ snd x < snd y)).
    pose (R x y := fst x <= fst y /\ snd x <= snd y).
    apply af_induction_type with (T := T) (R := R).
    unfold R.
    apply af_product; apply leq_af.
    admit.
    refine (fun c => 
    match c with
      | (0,_) => fun _ => 1
      | (_,0) => fun _ => 1
      | (S x, S y) => fun flex => (flex (x, S (S y)) _) + (flex (S x, y) _)
    end); unfold T; simpl; omega.
  Qed.

End af_lexico.

Section sct.

  Definition gnlex : nat * nat -> nat.
  Proof.
    set (T := fun x y => (fst x = snd y /\ snd x < snd y) \/ (fst x = snd y /\ snd x < fst y)).
    set (R := fun x y => fst x <= fst y /\ snd x <= snd y).
    apply af_induction_type with (T := T) (R := R).
    unfold R.
    apply af_product; apply leq_af.
    intros (x1,x2) (y1,y2); simpl.
    admit.
    refine (fun c => 
      match c with
        | (0,_) => fun _ => 1
        | (_,0) => fun _ => 1
        | (S x, S y) => fun gnlex => (gnlex (S y, y) _) + (gnlex (S y, x) _)
      end); red; simpl; abstract omega.
  Defined.

End sct.

Section terminator.

  Variable X : Type.

  Lemma disjunctive_wf (T R1 R2 : X -> X -> Prop) : 
        dec_rel R1
     -> dec_rel R2 
     -> well_founded R1 
     -> well_founded R2 
     -> (forall x y, clos_trans_1n _ T x y -> R1 x y \/ R2 x y) 
     -> well_founded T.
  Proof.
    intros dec1 dec2 wf1 wf2 incl.
    pose (R x y := ~ R1 y x /\ ~ R2 y x).
    apply wf_from_af with R; auto.
    intros x y H; unfold R.
    destruct (incl _ _ H); tauto.
    apply af_intersection;
    apply af_from_wf; auto.
  Qed.

End terminator.

Definition flip1 : nat * nat -> nat.
Proof.
  set (T := fun x y => fst x <= snd y /\ snd x < fst y).
  set (R := fun x y => fst x + snd x <= fst y + snd y).
  apply af_induction_type with (T := T) (R := R).
  unfold R; apply af_cofmap, leq_af.
  induction 1 as [ (x1,y1) (x2,y2) (H1 & H2) 
                 | (x1,y1) (x2,y2) (x3,y3) (H1 & H2) H IH ]; 
  unfold R; simpl in H1, H2 |- *. 
  omega.
  unfold R in IH; simpl in IH.
  omega.
 
  refine (fun c => 
    match c with
      | (0,_) => fun _ => 1
      | (_,0) => fun _ => 1
      | (S x, S y) => fun flip1 => _
    end).
  apply (flip1 (S y, x));
  unfold T; simpl; omega.
Defined.

Definition fib : nat -> nat.
Proof.
  apply af_induction_type with (T := lt) (1 := leq_af).
  intros x y H; apply clos_trans_1n_lt in H; abstract omega.

  refine (fun x => 
    match x as w return (forall y, y < w -> nat) -> nat with 
      | O         => fun frec => 1
      | 1         => fun frec => 1
      | (S (S x)) => fun frec => frec (S x) _ + frec x _
    end); abstract omega.
Defined.

