(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREER SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wellfounded.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import wft.
Require Import good_base.

Set Implicit Arguments.

(* 

   This file contains the basic informative (or "type-bounded") definitions of the 
   "Almost Full" AF predicate as defined in "Stop when your almost full" by Vytiniotis, 
   Coquand and Wahlstedt. The logical/prop-bounded version is denoted af 
   whereas the informative/type-bounded version is denoted af_t (and af_type, see below)

   Some proofs and results of this file are largely inspired by that work
   and its "Accompanying development" which is Coq code made available by
   the autors. 

   However, in the case of the type-bounded version af_t, as Vytiniotis et al.
   code, there is an alternate equivalent definition of af_t using well-founded 
   tree (wft) and securedby predicate. I did add a induction principle for af_t
   that shows it is really equivalent to the type-bounded version of af.
   With this inductive principle, you can simultaneously build wft and prove
   the securedby property. So much of the proofs scripts for af_t are nearly
   identical to those for af. This is not the case in the code from
   Vytiniotis et al.

   In my opinion, the really non-trivial part of what I borrowed from those
   authors is the *very nice* proof of Ramsey's theorem (originally by Coquand), 
   i.e. stability of the AF predicate by intersection (or isomorphically 
   by products).

   From my point of view, a quite trivial but however VERY usefull addition
   to that code is the af_t_relmap theorem which says that the AF is stable
   under surjective *relational* maps. Hence, it works very well for subtypes
   { x | P x } which is not really the case with  af_t_cofmap, af_t_fmap or
   af_t_fmap_gen. 

   Another addition is the dependently typed induction principle for AF
   named af_t_dep_rect where the ground case is NOT when the relation is
   full BUT when the type is EMPTY (a particular case of full relation).
   That concept plays a crucial role in the proof of both Higman's theorem
   and Kruskal's theorem.

   There is also a development for lexicographic products but there is
   more powerfull stuff in wft.

*)


Local Ltac squeeze := unfold lift_rel; intros ? ?; (tauto || firstorder || fail).


Section Almost_Full_type.

  Variable X : Type.
  
  Implicit Type R S : X -> X -> Prop.

  Definition af_t R := { t | af_securedby R t }.

  Fact in_af_t0 R : (forall x y, R x y) -> af_t R.
  Proof.
    exists in_wft_zt; auto.
  Qed.

  Fact in_af_t1 R : (forall a, af_t (R rlift a)) -> af_t R.
  Proof.
    intros g.
    exists (in_wft_sup (fun a => proj1_sig (g a))); simpl.
    intros a; apply (proj2_sig (g a)).
  Qed.

  Section af_t_rect.

    Variables (P : (X -> X -> Prop) -> Type) 
              (HP1 : forall R, (forall x y : X, R x y) -> P R)
              (HP2 : forall R, (forall a : X, af_t (R rlift a)) 
                            -> (forall a : X, P (R rlift a)) 
                            -> P R).
    
    Theorem af_t_rect : forall R, af_t R -> P R.
    Proof.
      intros R (t & HR).
      revert R HR P HP1 HP2.
      induction t as [ | f Hf ]; simpl; intros R HR P HP1 HP2.
      apply HP1, HR.
      apply HP2.
      intros a; exists (f a); apply HR.
      intros a; apply (Hf a); auto.
    Qed.

  End af_t_rect.
  
  Fact af_t_inc R S : R inc2 S -> af_t R -> af_t S.
  Proof.
    intros H0 H1; revert S H0.
    induction H1 as [ R HR | R HR1 HR2 ] using af_t_rect; intros S H.
    apply in_af_t0.
    intros; apply H; auto.
    apply in_af_t1.
    intros a; apply (HR2 a), lift_rel_inc; auto.
  Qed.

  Fact af_t_eq R S : R ~eq2 S -> af_t R -> af_t S.
  Proof.
    intros [ ? _ ]; apply af_t_inc; auto.
  Qed.
  
  Definition af_t_inv R : af_t R -> forall a, af_t (R rlift a).
  Proof.
    induction 1 as [ R HR | R HR IHR ] using af_t_rect; intros a.
    apply in_af_t0; left; auto.
    auto.
  Qed.

  Fact af_t_always_good R : af_t R -> forall f, { n | good R (pfx_rev f n) }.
  Proof.
    intros (t & Ht) f.
    exists (2+wft_seq_bound t f).
    apply af_sb_good_pfx; auto.
  Qed.

  Corollary af_t_inf_chain R : af_t R -> forall f, { n | exists i j, i < j <= n /\ R (f i) (f j) }.
  Proof.
    intros H f.
    destruct (af_t_always_good H f) as (n & Hn).
    apply good_eq_exists in Hn.
    exists n.
    destruct Hn as (l & a & m & b & r & H1 & H2).
    exists (length r), (length (m++b::r)); split.
    split.
    rewrite app_length; simpl; omega.
    apply f_equal with (f := @length _) in H1.
    rewrite pfx_length, app_length in H1; simpl in H1; omega.
    rewrite pfx_rev_eq with (1 := H1).
    cutrewrite (l++a::m++b::r = (l++a::m)++b::r) in H1.
    rewrite pfx_rev_eq with (1 := H1); auto.
    rewrite app_ass; auto.
  Qed.

  Corollary af_t_includes_eq R : af_t R -> @eq _ inc2 R.
  Proof.
    intros H1 x y ?; subst.
    destruct (af_t_inf_chain H1 (fun _ => y)) as ( ? & ? & ? & ? & ? ); auto.
  Qed.

  Section af_t_from_wf.

    Variable (R : X -> X -> Prop) (decR : forall x y, { R x y } + { ~ R x y }).

    Fact af_t_from_Acc z : Acc R z -> af_t (fun x y => ~ R x z \/ ~ R y x).
    Proof.
      induction 1 as [ z Hz IH ].
      apply in_af_t1.
      intros u.
      destruct (decR u z) as [ Hzu | Hzu ].
      generalize (IH _ Hzu). 
      apply af_t_inc; squeeze.
      apply in_af_t0; squeeze.
    Qed.

    Fact af_t_from_wf : well_founded R -> af_t (fun x y => ~ R y x).
    Proof.
      intros HR.
      apply in_af_t1.
      intros x.
      generalize (af_t_from_Acc (HR x)).
      apply af_t_inc; squeeze.
    Qed.
  
  End af_t_from_wf.
  
  Section wf_from_af_t.
  
    Implicit Types (T : X -> X -> Prop).

    Lemma acc_from_af_t R T y : 
       (forall x z,  clos_refl_trans _ T z y
                  -> clos_trans_1n _ T x z 
                  -> R z x 
                  -> False)
       -> af_t R  -> Acc T y.
    Proof.
      intros H1 H2; revert y T H1.
      induction H2 as [ R HR | R HR1 HR2 ] using af_t_rect; intros y T HT.

      constructor; intros x Hx.
      exfalso; apply (@HT x y); auto.
      constructor 2.
      constructor 1; auto.

      constructor; intros z Hz.
      apply HR2 with (a := y).
      intros x' z' H1 H2 [ H3 | H3 ].
      apply HT with (3 := H3); auto.
      constructor 3 with z; auto; constructor 1; auto.      
      apply HT with (3 := H3).
      constructor 2; auto.
      rewrite <- clos_trans_t1n_iff, -> clos_trans_tn1_iff.
      rewrite clos_trans_n1_clos_rt.      
      exists z; auto.
    Qed.

    Variables (R T :  X -> X -> Prop).

    Hypothesis HT : forall x y, clos_trans_1n _ T x y -> R y x -> False.
  
    Lemma wf_from_af_t : af_t R -> well_founded T.
    Proof.      
      intros H x.
      apply acc_from_af_t with (2 := H); auto.
      intros ? ? ?; apply HT.
    Qed.
      
  End wf_from_af_t.

End Almost_Full_type.

Section af_type.

  (* Just to show that we have an equivalent definition of Almost Full Type *)

  Variable (X : Type).
  
  Inductive af_type : (X -> X -> Prop) -> Type := 
    | in_af_type0 : forall R, full R -> af_type R
    | in_af_type1 : forall R, (forall a, af_type (R rlift a)) -> af_type R.

  Fact af_type_af_t R : af_type R -> af_t R.
  Proof.
    induction 1.
    apply in_af_t0; auto.
    apply in_af_t1; auto.
  Qed.
  
  Fact af_t_af_type R : af_t R -> af_type R.
  Proof.
    induction 1 using af_t_rect.
    constructor 1; auto.
    constructor 2; auto.
  Qed.

End af_type.

Section af_sb_incl.

  Variables (X : Type) (R S : relation X) (HRS : R inc2 S).

  Fact af_sb_incl t : af_securedby R t -> af_securedby S t.
  Proof.
    revert R S HRS.
    induction t as [ | f IHf ]; intros R S HRS; simpl.
    intros; auto.
    intros H x; generalize (H x); apply IHf.
    cbv; firstorder.
  Qed.

End af_sb_incl.

Fact af_t_True X : @af_t X (fun _ _ => True).
Proof.
  apply in_af_t0; auto.
Qed.

Lemma af_t_bool : af_t (@eq bool).
Proof.
  apply in_af_t1.
  intros [ | ]; apply in_af_t1.
  intros [ | ]. apply in_af_t0; firstorder.
  apply in_af_t1.
  intros [ | ].
  apply in_af_t0; firstorder.
  apply in_af_t0; firstorder.
  intro; apply in_af_t1; intros [ | ].
  apply in_af_t1; intros [ | ].
  apply in_af_t0; firstorder.
  apply in_af_t0; firstorder.
  apply in_af_t0; firstorder.
Qed.

Lemma le_af_t : af_t le.
Proof.
  generalize (af_t_from_wf lt_dec lt_wf).
  apply af_t_inc.
  intros ? ?; abstract omega.
Qed.

Section af_t_induction.

  Variables (X : Type)
            (T R : X -> X -> Prop) 
            (HR : af_t R) 
            (HTR : forall x y, clos_trans_1n _ T x y -> R y x -> False) 
            (P : X -> Type) (HP : forall x, (forall y, T y x -> P y) -> P x).

  Theorem af_t_induction_type: forall a, P a.
  Proof. 
    apply well_founded_induction_type with (R := T); auto.
    apply wf_from_af_t with (2 := HR); auto.
  Defined.

End af_t_induction.

Section af_t_union.

   Variables (X : Type) (R S : X -> X -> Prop).
   
   Fact af_t_union_left : af_t R -> af_t (fun x y => R x y \/ S x y).
   Proof.
     apply af_t_inc; squeeze.
   Qed.
   
   Fact af_t_union_right : af_t S -> af_t (fun x y => R x y \/ S x y).
   Proof.
     apply af_t_inc; squeeze.
   Qed.

End af_t_union.

(* a general projection principle with a surjective relation from R to S *)

Section af_t_relmap.

  Variables (X Y : Type) (f : X -> Y -> Prop) 
            (R : X -> X -> Prop) (S : Y -> Y -> Prop)
            (Hf : forall y, { x | f x y })
            (HRS : forall x1 x2 y1 y2, f x1 y1 -> f x2 y2 -> R x1 x2 -> S y1 y2).

  Fact af_t_relmap : af_t R -> af_t S.
  Proof.
    intros H; revert S HRS.
    induction H as [ R HR | R HR IH ] using af_t_rect; intros S H.

    apply in_af_t0.
    intros y1 y2.
    destruct (Hf y1) as (x1 & Hx1).
    destruct (Hf y2) as (x2 & Hx2).
    apply H with (3 := HR x1 x2); auto.
    
    apply in_af_t1; intros y0.
    destruct (Hf y0) as (x0 & Hx0).
    apply IH with (a := x0).
    intros x1 x2 y1 y2 H1 H2.
    generalize (H _ _ _ _ H1 H2).
    unfold lift_rel.
    firstorder.
  Qed.
    
End af_t_relmap.

(* the reverse map of a total function is surjective *)

Section af_cofmap.

  Variables (X Y : Type) (f : Y -> X).
  
  Implicit Types (R : X -> X -> Prop).

  Fact af_t_cofmap R : af_t R -> af_t (fun x y => R (f x) (f y)).
  Proof.
    apply af_t_relmap with (f := fun x y => x = f y).
    intros y; exists (f y); auto.
    intros; subst; auto.
  Qed.

End af_cofmap.

(* two instances where the relation is a surjective function *)

Section af_t_fmap_gen.

  Variables (X Y : Type) (f : X -> Y) (Hf : forall y, { x | y = f x }).

  Implicit Types (R : X -> X -> Prop) (S : Y -> Y -> Prop).  

  Fact af_t_fmap_gen R S : (forall x y, R x y -> S (f x) (f y)) -> af_t R -> af_t S.
  Proof.
    intros H.
    apply af_t_relmap with (f := fun x y => y = f x); auto.
    intros; subst; auto.
  Qed.

  Definition rel_fmap R (y1 y2 : Y) := exists x1 x2, y1 = f x1 /\ y2 = f x2 /\ R x1 x2.
  
  Fact af_t_fmap R : af_t R -> af_t (rel_fmap R).
  Proof.
    apply af_t_fmap_gen; auto.
    intros x1 x2; intros; subst; exists x1, x2; auto.
  Qed.
 
End af_t_fmap_gen.

(* This is the proof of Ramsey's theorem borrowed from Thierry Coquand's proof *)

Section af_t_intersection.

  Variable X : Type.

  Implicit Types (C : X -> X -> Prop).

  Lemma af_t_null_inter_rec A B R :      af_t R 
                           -> forall C, (forall x y, R x y <-> C x y \/ A) 
                                      -> af_t (fun x y => C x y \/ B) 
                                      -> af_t (fun x y => C x y \/ A /\ B).
  Proof.
    induction 1 as [ R HR | R HR1 HR2 ] using af_t_rect; intros C HC HB.
    
    apply af_t_inc with (2 := HB).
    intros x y [ Hxy | Hxy ]; auto.
    specialize (HR x y); rewrite HC in HR; tauto.
    
    apply in_af_t1.
    intros x.
    apply af_t_inc with (R := fun y z => (C y z \/ C x y) \/ A /\ B).
    squeeze.
    apply HR2 with (a := x).
    intros ? ?; unfold lift_rel; do 2 rewrite HC; tauto.
    apply af_t_inc with (2 := HB); squeeze.
  Qed.

  Lemma af_t_null_inter C A B : af_t (fun x y => C x y \/ A)
                             -> af_t (fun x y => C x y \/ B)
                             -> af_t (fun x y => C x y \/ (A /\ B)).
  Proof.
    intros H.
    apply af_t_null_inter_rec with (1 := H).
    intros; tauto.
  Qed.

  Lemma af_t_una_inter_rec A B R : af_t R 
                      -> forall T, af_t T 
                      -> forall C, (forall x y, R x y -> C x y \/ A x) 
                                -> (forall x y, T x y -> C x y \/ B x) 
                                -> af_t (fun x y => C x y \/ (A x /\ B x)).
  Proof.
    induction 1 as [ R HR | R HR IHR ] using af_t_rect.

    intros T HT C Req Teq.
    eapply af_t_inc with (2 := HT).
    intros x y; generalize (Req x y) (Teq x y) (HR x y); tauto.
    
    induction 1 as [ T HT | T HT IHT ] using af_t_rect.

    intros C Req Teq.
    apply af_t_inc with (R := R).
    intros x y; generalize (HT x y) (Req x y) (Teq x y); tauto.
    apply in_af_t1, HR.
  
    intros C Req Teq.
    apply in_af_t1; intros x.
    generalize (in_af_t1 HT); intros HT'.
    apply af_t_inc with (R := fun y z : X => ((C y z \/ A y /\ B y) \/ C x y) \/ A x /\ B x).
    squeeze.
    apply af_t_null_inter.

    generalize (IHR x _ HT' (fun y z => C y z \/ C x y \/ A x)); intros G0.
    eapply af_t_inc.

    Focus 2.
    apply G0.
    intros y z; generalize (Req y z) (Req x y); unfold lift_rel; tauto.
    intros y z Hyz; generalize (Teq y z); tauto.
    
    squeeze.

    generalize (IHT x (fun y z => C y z \/ C x y \/ B x)); intros G0.
    eapply af_t_inc.

    Focus 2.
    apply G0.
    intros y z; generalize (Req y z) (Req x y); tauto.
    intros y z [ H1 | H1 ]; apply Teq in H1; tauto.
    
    squeeze.
  Qed.
    
  Lemma af_t_una_inter C A B : af_t (fun x y => C x y \/ A x)
                            -> af_t (fun x y => C x y \/ B x)
                            -> af_t (fun x y => C x y \/ (A x /\ B x)).
  Proof.
    intros H1 H2.
    apply af_t_una_inter_rec with (1 := H1) (2 := H2); squeeze.
  Qed.

  Implicit Types (A B : X -> X -> Prop).
  
  Theorem af_t_inter A B : af_t A -> af_t B -> af_t (A cap2 B). 
  Proof.
    intros HA HB.
    revert A HA B HB.
    induction 1 as [ A HA | A HA IHA ] using af_t_rect.
    
    intros B HB.
    apply af_t_inc with (2 := HB).
    intros x y; specialize (HA x y); tauto.
    
    induction 1 as [ B HB | B HB IHB ] using af_t_rect.

    apply af_t_inc with (2 := in_af_t1 HA); squeeze.
    
    apply in_af_t1; intros x.
    unfold lift_rel.
    apply af_t_una_inter.
    apply af_t_inc with (2 := IHA x _ (in_af_t1 HB)); squeeze.
    apply af_t_inc with (2 := IHB x); squeeze.
  Qed.

End af_t_intersection.

Section af_t_prod.
  
  Variable (X Y : Type).
  Variable (R : X -> X -> Prop) (S : Y -> Y -> Prop).
  
  Theorem af_t_prod : af_t R
                   -> af_t S
                   -> af_t (rel_prod R S).
  Proof.
    intros HR HS.
    apply af_t_inter; apply af_t_cofmap; auto.
  Qed.

End af_t_prod.

Section af_t_sum.

  Variable (X Y : Type).

  Lemma af_t_sum_left R : af_t R -> af_t (@sum_left X Y R).
  Proof.
    induction 1 as [ R HR | R HR IH ] using af_t_rect.
    
    apply in_af_t1; intros x.
    apply in_af_t1; intros y.
    apply in_af_t0.
    intros [|] [|]; destruct x; destruct y; simpl; firstorder.
    
    apply in_af_t1; intros [ x | x ].
    apply af_t_inc with (2 := IH x).
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
    apply in_af_t1; intros [ y | y ].
    apply af_t_inc with (2 := IH y).
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
    apply in_af_t0.
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
  Qed.

  Lemma af_t_sum_right R : af_t R -> af_t (@sum_right X Y R).
  Proof.
    induction 1 as [ R HR | R HR IH ] using af_t_rect.
    
    apply in_af_t1; intros x.
    apply in_af_t1; intros y.
    apply in_af_t0.
    intros [|] [|]; destruct x; destruct y; simpl; firstorder.
    
    apply in_af_t1; intros [ x | x ].
    apply in_af_t1; intros [ y | y ].
    apply in_af_t0.
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
    apply af_t_inc with (2 := IH y).
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
    apply af_t_inc with (2 := IH x).
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
  Qed.

  Lemma af_t_sum_both R S : af_t R -> af_t S -> af_t (@sum_both X Y R S).
  Proof.
    intros HR HS.
    apply af_t_inc with (2 := af_t_inter (af_t_sum_left HR) (af_t_sum_right HS)).
    intros [|] [|] (H1 & H2); simpl in H1, H2 |- *; tauto.
  Qed.

End af_t_sum.

Section af_t_fsum.

  Variable (Y : Type).
  
  Fact af_t_rel_fsum n R : (forall k Hk, af_t (R k Hk)) -> af_t (@rel_fsum Y n R).
  Proof.
    revert R.
    induction n as [ | n IH ]; intros R HR.
    apply in_af_t0; simpl.
    intros ((? & ?) & ?); exfalso; omega.

    simpl in R, HR.

    set (S0 := R 0 (lt_0_Sn n)).
    set (Sn k Hk := R _ (lt_n_S k n Hk)).
    assert (HS0 : af_t S0). 
      apply HR. 
    assert (HSn : forall k Hk, af_t (Sn k Hk)). 
      intros; apply HR.
      
    generalize (af_t_sum_both HS0 (IH Sn HSn)).
    
    set (f (c : Y + { k | k < n } * Y) (d : { k | k < S n } * Y) :=
      match c, d with
        | inl y1              , (exist 0 _,y2)     => y1 = y2    
        | inr (exist k _,y1)  , (exist (S l) _,y2) => k = l /\ y1 = y2
        | _                   , _                  => False
      end).   
    apply af_t_relmap with f.
    
    intros (([ | a] & Ha) & y).
    exists (inl y); simpl; auto.
    exists (inr (exist _ a (lt_S_n _ _ Ha), y)); simpl; auto.
    
    intros [ y1 | ((k1 & Hk1) & y1) ] 
           [ y2 | ((k2 & Hk2) & y2) ]
           (([|k3] & Hk3) & y3) 
           (([|k4] & Hk4) & y4); simpl; try tauto.
    intros ? ?; subst; auto.
    unfold S0; rewrite (lt_pirr _ Hk3); auto.
    intros (? & ?) (? & ?); subst k3 k4 y3 y4.
    destruct (eq_nat_dec k1 k2); simpl.
    unfold Sn; rewrite (lt_pirr _ Hk3); auto.
    auto.
  Defined.

End af_t_fsum.

Section af_t_indexed_prod.

  Variable (X : Type) (ll : list X) 
           (t : forall x, In_t x ll -> Type) 
           (R : forall x Hx, relation (t x Hx))
           (HR : forall x Hx, af_t (@R x Hx)).
           
  Fact af_t_idx_prod : af_t (fun f g => forall x Hx, @R x Hx (f x Hx) (g x Hx)).
  Proof.
    revert t R HR.
    induction ll as [ | x lx IHlx ]; intros t R HR.
    
    apply in_af_t0.
    intros _ _ ? [].
    
    set (R0 := @R _ (inl eq_refl)).
    assert (HR0 : af_t R0).
      apply HR.
    set (t' := fun x Hx => @t x (inr Hx)).
    set (R' := fun x Hx => @R x (inr Hx)).
    assert (HR' : forall x Hx, af_t (R' x Hx)).
      intros; apply HR.
    
    generalize (af_t_prod HR0 (IHlx t' _ HR')).
    set (f (c : @t _ (inl eq_refl)  * forall x Hx, @t x (inr Hx)) (g : forall z Hz, @t z Hz) := 
      match c with 
        | (x,f) => x = g _ (inl eq_refl) /\ forall Y HY, f Y HY = g Y (inr HY)
      end). 
    apply af_t_relmap with (f := f).
    intros g; exists (g _ (inl eq_refl),fun X HX => g X (inr HX)); simpl; auto.
    intros (x1,f1) (x2,f2) g1 g2; unfold f; intros (H1 & H2) (H3 & H4).
    unfold rel_prod; simpl.
    intros (H5 & H6).
    intros U [ HU | HU ].
    subst U.
    unfold R0 in H5; replace_with H5; auto.
    rewrite <- H2.
    rewrite <- H4.
    apply H6.
  Defined.    

End af_t_indexed_prod.

Section af_t_list_prod.

  Variables (ll : list Type) (R : forall X, In_t X ll -> relation X) (HR : forall X HX, af_t (@R X HX)).
  
  Fact af_t_list_prod : af_t (fun f g => forall X HX, @R X HX (f X HX) (g X HX)).
  Proof.
    apply af_t_idx_prod with (t := fun x _ => x); auto.
  Defined.
 
End af_t_list_prod. 

Section af_t_dep_rect.

  Variable (P : forall X, relation X -> Type).

  Hypothesis HP0 : P ER.
  Hypothesis HP1 : forall X R, full R -> P ER -> @P X R.
  Hypothesis HP2 : forall X R, (forall x, af_t (R rlift x)) -> (forall x, P (R rlift x)) -> @P X R.

  Theorem af_t_dep_rect : forall X R, af_t R -> @P X R.
  Proof.
    intros X.
    induction 1 as [ R HR | R HR IHR ] using af_t_rect.
    apply HP1; auto.
    apply HP2; auto.
  Defined.

End af_t_dep_rect.

Section af_t_dep_rect'.

  Variable (P : forall X, relation X -> Type).

  Hypothesis HP0 : forall X R, (X -> False) -> @P X R.
  Hypothesis HP1 : forall X R, full R -> (forall Y R, (Y -> False) -> @P Y R) -> @P X R.
  Hypothesis HP2 : forall X R, (forall x, af_t (R rlift x))-> (forall x, P (R rlift x)) -> @P X R.

  Theorem af_t_dep_rect' : forall X R, af_t R -> @P X R.
  Proof.
    intros X.
    induction 1 as [ R HR | R HR IHR ] using af_t_rect.
    apply HP1; auto.
    apply HP2; auto.
  Defined.

End af_t_dep_rect'.



