(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import good_base.
Require Import af.

Set Implicit Arguments.

(* we can prove properties about AF on subsets { x | P x }
   w/o actually manipulating subsets 
   for this, use AFS instead of AF

   AND YES, you don't want to end up having to prove that
   two proof of P x are equal ;-)
*)

Section afs.

  Variables (X : Type) (P : X -> Prop).

  Implicit Type R S : X -> X -> Prop.

  Inductive afs : (X -> X -> Prop) -> Prop := 
    | in_afs_0 : forall R, (forall x y, P x -> P y -> R x y) -> afs R
    | in_afs_1 : forall R, (forall a, P a -> afs (R rlift a)) -> afs R.

  Fact afs_inv R : afs R -> forall a, P a -> afs (R rlift a).
  Proof.
    induction 1 as [ R HR | R HR IHR ]; auto.
    intros; constructor 1; left; auto.
  Qed.    

  Fact afs_inc R S : R inc2 S -> afs R -> afs S.
  Proof.
    intros H1 H; revert S H1; induction H as [ R HR | R HR IH ]; intros S H1.
    constructor 1; auto.
    constructor 2; auto.
    intros x Hx; apply (IH x Hx).
    intros y z [|]; [ left | right ]; auto.
  Defined.

  Fact afs_binary_infinite_chain R f : (forall n, P (f n)) -> afs R -> forall k, exists i j, k <= i < j /\ R (f i) (f j).
  Proof.
    intros Hf.
    induction 1 as [ R' HR' | R' HR' IH ]; intros k.
    exists k, (S k); split; auto; apply HR.
    destruct (IH _ (Hf k) (S k)) as ( i & j & Hij & [ H | H ] ); 
      [ exists i, j | exists k, i ]; split; auto; omega.
  Qed.

  Corollary afs_inf_chain R f : (forall n, P (f n))-> afs R -> exists i j, i < j /\ R (f i) (f j).
  Proof.
    intros Hf H.
    destruct (afs_binary_infinite_chain f Hf H 0) as ( i & j & ( _ & ? ) & ? ).
    exists i, j; auto.
  Qed.

  Corollary afs_contains_eq R : afs R -> forall x, P x -> R x x.
  Proof.
    intros H1 x Hx.
    destruct (@afs_inf_chain R (fun _ => x)); auto.
    destruct H; tauto.
  Qed.

  Let afs_af_rec K : af K -> forall R, K inc2 R <# P #> -> afs R.
  Proof.
    induction 1 as [ S HS | S HS IHS ]; intros R HR.
    constructor 1.
    intros x y Hx Hy.
    specialize (HS (exist _ x Hx) (exist _ y Hy)).
    apply HR in HS; auto.
    constructor 2; intros a Ha.
    specialize (IHS (exist _ a Ha)).
    apply IHS.
    intros [ x Hx ] [ y Hy ].
    generalize (HR (exist _ x Hx) (exist _ y Hy))
               (HR (exist _ a Ha) (exist _ x Hx)).
    cbv; tauto.
  Qed. 

  Fact afs_af_rel_restr_eq R : afs R <-> af (R <# P #>).
  Proof.
    split.

    induction 1 as [ R HR | R HR IHR ].
    constructor 1.
    intros [ x Hx ] [ y Hy ]; red; simpl; auto.
    constructor 2.
    intros (a & Ha).
    specialize (IHR _ Ha).
    revert IHR.
    apply af_inc.
    intros [ x Hx ] [ y Hy ]; auto.
    
    intros H.
    apply afs_af_rec with (1 := H).
    auto.
  Qed.

End afs.

Fact af_any_afs X P R : af R -> @afs X P R.
Proof.
  induction 1 as [ R G1 | R G1 G2 ].
  apply in_afs_0; intros; auto.
  apply in_afs_1; auto.
Qed.

(* af is a specialization of afs on the full subset *)

Fact af_afs_True_eq X R : @af X R <-> afs (fun _ => True) R.
Proof.
  split.
  induction 1.
  constructor 1; auto.
  constructor 2; auto.
  induction 1.
  constructor 1; auto.
  constructor 2; auto.
Qed.

Section afs_mono.

  Variables (X : Type) (P Q : X -> Prop) (R S : X -> X -> Prop).
  
  Fact afs_mono : Q inc1 P -> R inc2 S -> afs P R -> afs Q S.
  Proof.
    intros H1 H2 H3.
    revert S H2.
    induction H3 as [ R HR | R HR IHR ]; intros S H2.
    constructor 1; intros; auto.
    constructor 2; intros a Ha.
    apply (IHR a); auto.
    intros; firstorder.
  Qed.

End afs_mono.

Section afs_relmap.

  Variables (X Y : Type) (f : X -> Y -> Prop) (P : X -> Prop) (Q : Y -> Prop)
            (R : X -> X -> Prop) (S : Y -> Y -> Prop)
            (Hf : forall y, Q y -> exists x, P x /\ f x y)
            (HRS : forall x1 x2 y1 y2, P x1 -> P x2 -> Q y1 -> Q y2 -> f x1 y1 -> f x2 y2 -> R x1 x2 -> S y1 y2).

  Fact afs_relmap : afs P R -> afs Q S.
  Proof.
    intros H; revert S HRS.
    induction H as [ R HR | R HR IH ]; intros S H.

    constructor 1.
    intros y1 y2 Hy1 Hy2.
    destruct (Hf Hy1) as (x1 & H1 & H2).
    destruct (Hf Hy2) as (x2 & H3 & H4).
    apply H with (7 := HR _ _ H1 H3); auto.
    
    constructor 2; intros y0 Hy0.
    destruct (Hf Hy0) as (x0 & Hx0 & Hx0').
    apply IH with (1 := Hx0).
    intros x1 x2 y1 y2 Hx1 Hx2 H1 H2 H3 H4.
    generalize (H _ _ _ _ Hx1 Hx2 H1 H2 H3 H4).
    generalize (H _ _ _ _ Hx0 Hx1 Hy0 H1 Hx0' H3).
    unfold lift_rel; tauto.
  Qed.
    
End afs_relmap.

Section afs_fmap_gen.

  Variables (X Y : Type) (f : X -> Y) (P : X -> Prop) (Q : Y -> Prop) 
            (R : X -> X -> Prop) (S : Y -> Y -> Prop)
            (Hf : forall y, Q y -> exists x, P x /\ y = f x)
            (HRS : forall x y, P x -> P y -> Q (f x) -> Q (f y) -> R x y -> S (f x) (f y)).

  Fact afs_fmap_gen : afs P R -> afs Q S.
  Proof.
    apply afs_relmap with (f := fun x y => y = f x); auto.
    intros; subst; auto.
  Qed.

End afs_fmap_gen.

Section afs_sum.

  Variable (X Y : Type) (P : X -> Prop) (Q : Y -> Prop) (R : X -> X -> Prop) (S : Y -> Y -> Prop).

  Hypothesis HX : afs P R.
  Hypothesis HY : afs Q S.

  Let f : sig P + sig Q -> sig (pred_sum P Q).
  Proof. 
    intros [ (x & Hx) | (y & Hy) ].
    exists (inl x); auto.
    exists (inr y); auto.
  Defined.

  Let f_surj : forall y, exists x, y = f x.
  Proof. 
    intros ( [ x | y ] & H ); simpl in H.
    exists (inl (exist _ _ H)); auto.
    exists (inr (exist _ _ H)); auto.
  Qed.

  Fact afs_sum_both : afs (pred_sum P Q) (sum_both R S).
  Proof.
    rewrite afs_af_rel_restr_eq in HX.
    rewrite afs_af_rel_restr_eq in HY.
    rewrite afs_af_rel_restr_eq.
    generalize (af_sum_both HX HY).
    apply af_fmap_gen with (1 := f_surj).
    
    intros [ (x1 & Hx1) | (y1 & Hy1) ]
           [ (x2 & Hx2) | (y2 & Hy2) ]; cbv; auto.
  Qed.

End afs_sum.

Section afs_product.

  Variable (X Y : Type) (P : X -> Prop) (Q : Y -> Prop) (R : X -> X -> Prop) (S : Y -> Y -> Prop).

  Hypothesis HX : afs P R.
  Hypothesis HY : afs Q S.

  Let f : sig P * sig Q -> sig (pred_prod P Q).
  Proof. 
    intros [ (x & Hx) (y & Hy) ].
    exists (x,y); red; auto.
  Defined.

  Let f_surj : forall y, exists x, y = f x.
  Proof. 
    intros ( (x,y) & H1 & H2).
    exists (exist _ _ H1, exist _ _ H2); auto.
  Qed.

  Fact afs_product : afs (pred_prod P Q) (rel_prod R S).
  Proof.
    rewrite afs_af_rel_restr_eq in HX.
    rewrite afs_af_rel_restr_eq in HY.
    rewrite afs_af_rel_restr_eq.
    generalize (af_product HX HY).
    apply af_fmap_gen with (1 := f_surj).
    
    intros [ (x1 & Hx1) (y1 & Hy1) ]
           [ (x2 & Hx2) (y2 & Hy2) ]; cbv; auto.
  Qed.

End afs_product.

Section afs_idx_rel.
  
  Variables (I X : Type) (P : X -> Prop).

  Fact afs_idx_rel R (li : list I) : (forall i, In i li -> afs P (R i))
                         -> afs (fun l => Forall P l /\ length l = length li) (idx_rel R li).
  Proof.
    induction li as [ | i li IHli ]; intros H.

    apply in_afs_0.
    intros [ | ] [ | ] (Hx1 & Hx2) (Hy1 & Hy2).
    constructor.
    discriminate Hy2.
    discriminate Hx2.
    discriminate Hx2.
    
    spec_all IHli.
    intros; apply H; right; auto.
    specialize (H i); spec_all H.
    left; auto.

    generalize (afs_product H IHli); intros H2; spec_all H2; auto.
    revert H2.
    apply afs_relmap with (f := fun c m => fst c::snd c = m); auto.

    intros [ | y ly ] (H1 & H2).
    discriminate H2.
    injection H2; clear H2; intros H2.
    apply Forall_cons_inv in H1; destruct H1.
    exists (y,ly); cbv; auto.
    
    intros (x1 & lx1) (x2 & lx2) 
           [ | y1 ly1 ] [ | y2 ly2 ]; simpl;
    intros (H1 & H2 & H3) (H4 & H5 & H6) (H7 & H8) (H9 & H10); 
      try discriminate 1; try discriminate 2.
    intros G1 G2; injection G1; injection G2; clear G1 G2; intros ? ? ? ?; subst.
    intros (? & ?); constructor; auto.
  Qed.

End afs_idx_rel.

Section afs_union.

  Variable (X : Type) (P Q : X -> Prop) (R S : X -> X -> Prop).
  
  Hypothesis HR : afs P R.
  Hypothesis HS : afs Q S.
  
  Fact afs_union : afs (P cup1 Q) (R cup2 S).
  Proof.
    generalize (afs_sum_both HR HS).
    apply afs_relmap with (f := fun x y => y = match x with inl x => x | inr x => x end).
    intros x [ Hx | Hx ].
    exists (inl x); auto.
    exists (inr x); auto.
    intros [|] [|]; simpl; intros; subst; tauto.
  Qed.

End afs_union.

Section afs_sig_not.

  Variables (X : Type) (R : X -> X -> Prop) (x : X).
  
  Hypothesis Rdec : forall y, R x y \/ ~ R x y.

  Fact afs_rlift_afs_eq P : afs P (R rlift x) <-> afs (P cap1 (fun y => ~ R x y)) R.
  Proof.
    split.
 
    apply afs_relmap with (f := fun a b => b = a /\ ~ R x a).   
    intros y (H1 & H2); exists y; auto.
    unfold lift_rel; intros x1 x2 y1 y2 _ _ _ _ (? & H) (? & _); subst; tauto.

    (* we use afs_union with P = (P cap ~ R x) cup (P cap R x) *)
    
    intros H0.
    assert (afs (P cap1 R x) (R rlift x)) as H1.
      constructor 1.
      unfold lift_rel.
      firstorder.
    generalize (afs_union H0 H1).

    apply afs_relmap with (f := fun a b => b = a).
    intros y Hy; exists y; split; auto.
    specialize (Rdec y); tauto.
    unfold lift_rel.
    intros; subst; tauto.
  Qed.
  
  Fact af_rlift_afs_eq : af (R rlift x) <-> afs (fun y => ~ R x y) R.
  Proof.
    rewrite af_afs_True_eq, afs_rlift_afs_eq.
    split; apply afs_mono; auto.
    intro; tauto.
  Qed.

End afs_sig_not.

(* conditions under which a polymorphic operator like subword or subtree 
   which preserves almost full relations transfers easily from af _ -> af (op _)
   to afs _ _ -> afs _ (op _)
*)

Section afs_operator.

  Variable (f : Type -> Type) (op_1 : forall X, (X -> Prop) -> (f X -> Prop))
                              (op_2 : forall X, relation X -> relation (f X)).
  
  Hypothesis Hop : forall X R, af R -> af (@op_2 X R).

  Variable h : forall X (P : X -> Prop), f (sig P) -> sig (op_1 P).

  Hypothesis Hh1 : forall X P y, exists x, @proj1_sig _ (op_1 P) y = proj1_sig (@h X P x).
  Hypothesis Hh2 : forall X P R a b, op_2 (R <# P #>) a b -> op_2 R (proj1_sig (@h X P a)) (proj1_sig (h b)).

  Theorem afs_operator X P R : @afs X P R -> afs (op_1 P) (op_2 R).
  Proof.
    intros H.
    rewrite afs_af_rel_restr_eq in H.
    apply Hop in H.
    rewrite afs_af_rel_restr_eq.

    revert H; apply af_relmap with (f := fun x y => proj1_sig y = proj1_sig (@h X P x)); auto.
    intros x1 x2 (y1 & Hy1) (y2 & Hy2); simpl; intros ? ?; subst y1 y2.
    apply Hh2.
  Qed.

End afs_operator.







