(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Permutation.
Require Import Relations.
Require Import Wf.
Require Import Wellfounded.
Require Import Transitive_Closure.

Require Import tacs.

Require Import reverse.
Require Import wf_upto.
Require Import rel_utils.
Require Import list_utils.
Require Import lex.

Require Import af.
Require Import afs.

Set Implicit Arguments.

(* the two "induction" principles afs_easier_ind afs_facile_ind
   derive from the notion "well-founded upto a projection"
   defined in wf_upto.

   Basically, with a "well-founded upto f" relation R (where f
   is a map that "forgets" some information), you can prove
   a predicate P by induction using R provided that P 
   expresses no property about the information that is forgotten.

   Definition well_founded_upto f R := 
          forall Q, (forall a, (forall b, R b a -> Q (f b)) -> Q (f a)) 
                  -> forall a, Q (f a).

   In that case, R is not necessarily well-founded (it can even
   contain x such that R x x !!) 

   This is potentially an *important notion* to study for itself
   It autorizes to reason by case induction on lexicographic
   combinations of inductive predicates

*)

Section af_witness.

  Inductive af_witness : Set := af_empty | af_full | af_rlift.
  
  Definition lt_afw w1 w2 :=
    match w1, w2 with
      | af_empty, af_empty => False
      | af_empty, _        => True
      | af_full , af_rlift => True
      | _       , _        => False
    end.

  Definition lt_afw_wf : well_founded lt_afw.
  Proof.
    assert (Acc lt_afw af_empty).
      constructor; intros [ | | ]; simpl; try tauto.
    assert (Acc lt_afw af_full).
      constructor; intros [ | | ]; simpl; try tauto.
      
    intros [ | | ]; auto.
    constructor; intros [ | | ]; simpl; try tauto.
  Qed.

End af_witness.

Section af_subrel_wf_upto.

  Variable (X : Type).

  Definition af_subrel := (af_witness * ((X -> Prop) * relation X))%type.

  Definition afsr_correct (c : af_subrel) := 
    match c with
      | (af_empty,(P,_)) => forall x, ~ P x                    (* the type is empty    *) 
      | (af_full ,(P,R)) => forall x y, P x -> P y -> R x y    (* the relation is full *)
      | (af_rlift,(P,R)) => afs P R                            (* the relation is af   *)
    end.

  Definition afs_correct w P R := forall i : nat, afsr_correct (w i,(P i,R i)).

  Fact afsr_correct_afs w P R : afsr_correct (w,(P,R)) -> afs P R.
  Proof.
    destruct w; simpl; auto.
    intros H; constructor 1; intros x ? Hx; contradict Hx; auto.
    constructor 1; auto.
  Qed.
  
  Fact afs_correct_afs w P R : afs_correct w P R -> forall i, afs (P i) (R i).
  Proof.
    intros H i; generalize (H i); apply afsr_correct_afs.
  Qed.
  
  Definition lt_afsr (c1 c2 : af_subrel) :=
    match c1         , c2 with
      | (w1,(P1,R1)) , (w2,(P2,R2)) => lt_afw w1 w2 
                                    \/ w1 = w2 
                                    /\ w1 = af_rlift 
                                    /\ P1 = P2 
                                    /\ exists p, P1 p 
                                              /\ R1 = (R2 rlift p)
    end.

  Fact lt_afsr_empty_full PR1 PR2 : lt_afsr (af_empty,PR1) (af_full,PR2).
  Proof. destruct PR1; destruct PR2; left; cbv; auto. Qed.
  
  Fact lt_afsr_rlift (P : _ -> Prop) R m : P m -> lt_afsr (af_rlift, (P,R rlift m)) (af_rlift, (P,R)).
  Proof. right; repeat split; auto; exists m; auto. Qed. 

  Section lt_afsr_wf.

    Let lt_afsr_empty c1 PR : lt_afsr c1 (af_empty,PR) -> False.
    Proof.
      destruct PR.
      destruct c1 as ([ | | ] & P1 & R1); simpl.
      intros [ [] | (_ & H & _) ]; discriminate H.
      intros [ [] | (_ & H & _) ]; discriminate H.
      intros [ [] | (H & _ & _) ]; discriminate H.
    Qed.
    
    Let lt_afsr_full w1 PR1 PR2 : lt_afsr (w1,PR1) (af_full,PR2) -> w1 = af_empty.
    Proof.
      destruct w1; simpl; destruct PR1; destruct PR2.
      unfold lt_afw; simpl; tauto.
      intros [ [] | (_ & H & _) ]; discriminate H.
      intros [ [] | (H & _ & _) ]; discriminate H.
    Qed.
        
    Let lt_afsr_rlift P1 R1 P2 R2 : lt_afsr (af_rlift,(P1,R1)) (af_rlift,(P2,R2)) <-> P1 = P2 /\ exists p, P1 p /\ R1 = (R2 rlift p).
    Proof.
      simpl.
      split; unfold lt_afw; simpl; tauto.
    Qed.
    
    Variable Q : ((X -> Prop) * relation X) -> Prop.
    
    Hypothesis HQ : forall v c, afsr_correct (v,c) -> (forall w d, afsr_correct (w,d) -> lt_afsr (w,d) (v,c) -> Q d) -> Q c.

    Let HQ0 PR : afsr_correct (af_empty,PR) -> Q PR.
    Proof.
      intros H.
      apply HQ with (v := af_empty); auto.
      intros w d H1 H2.
      exfalso; revert H2.
      apply lt_afsr_empty.
    Qed.
      
    Let HQ1 PR : afsr_correct (af_full,PR) -> Q PR.
    Proof.
      simpl; intros H.
      apply HQ with (v := af_full); auto.
      intros w1 PR1 H1 H2.
      apply lt_afsr_full in H2.
      subst w1; auto.
    Qed.
      
    Fact lt_afsr_wf_upto_rec : forall w c, afsr_correct (w,c) -> Q c.
    Proof.
      intros w (P,R); simpl.
      revert P R; 
      induction w as [ [ | | ] IHw ] using (well_founded_induction lt_afw_wf); intros P R; auto.
      
      simpl.
      induction 1 as [ R HR | R HR IHR ] using afs_ind; auto.
      
      apply (HQ af_rlift (P,R)).
      red; auto.
      constructor 2; auto.
      intros [ | | ] (P1,R1) H1 H2; auto.
      rewrite lt_afsr_rlift in H2.
      destruct H2 as (H2 & p & H3 & H4). 
      subst P R1; auto.
    Qed.
      
  End lt_afsr_wf.

  (* lt_afsr is a well-founded relation UPTO forgetting the witness !!! *)
  
  Theorem lt_afsr_upto_wf : well_founded_upto (fun c => snd c) afsr_correct lt_afsr. 
  Proof.
    intros Q HQ.
    intros (w,c) Ha; simpl; revert w c Ha.
    apply lt_afsr_wf_upto_rec.
    intros v c Hc H.
    apply (HQ (v,c) Hc).
    intros (w,d) Hd; simpl.
    apply H; auto.
  Qed.

End af_subrel_wf_upto.

Section easier_direct.

  Variables (X : Type) (k : nat) 
            (Q : (nat -> X -> Prop) -> (nat -> relation X) -> Prop) 
            (Qext : ext2 Q).

  Implicit Type P : nat -> X -> Prop.

  Definition afs_easier w1 P1 R1 w2 P2 R2 := 
    exists i, i < k /\ lt_afsr (w1 i,(P1 i,R1 i)) (w2 i,(P2 i,R2 i))
                    /\ forall j, i < j -> w1 j = w2 j /\ P1 j = P2 j /\ R1 j = R2 j.

  Hypothesis IH : forall w P R, afs_correct w P R 
                            -> (forall w' P' R', afs_correct w' P' R' 
                                              -> afs_easier w' P' R' w P R
                                              -> Q P' R') 
                            -> Q P R.

  Let X1 := af_subrel X.
  Let P1 (_ : nat) : X1 -> Prop := @afsr_correct X.
  Let R1 (_ : nat) : relation X1 := @lt_afsr X.
  
  Let lex_R1_wf : well_founded_upto_ext ext (fun s i => snd (s i)) (fun s => forall i, P1 i (s i)) (lex_nat_seq k R1).
  Proof.
    apply wf_upto_lex_nat_seq.
    intro; apply lt_afsr_upto_wf.
  Qed.

  Let K s := Q (fun i => fst (s i)) (fun i => snd (s i)).

 Theorem afs_easier_ind P R : (forall i, afs (P i) (R i)) -> Q P R.
  Proof.
    intros H.
    set (s i := (af_rlift,(P i,R i))).
    assert (forall i, afsr_correct (s i)) as Hs.
      intros; simpl; auto.
    change (K (fun i => snd (s i))).
    revert Hs.
    
    apply lex_R1_wf.
    
    intros s1 s2 H12; unfold K; apply Qext;
    intro; rewrite H12; auto.
    
    intros a Ha IHa.
    unfold K.
    apply IH with (w := fun i => fst (a i)).
    intros i.
    generalize (Ha i); destruct (a i) as (?,(?,?)); auto.
    
    intros w' P' R' H2 (i & H3 & H4 & H5).
    apply IHa with (1 := H2).

    exists i.
    split; auto.
    split.
    intros j Hj; generalize (H5 j Hj).
    destruct (a j) as (?,(?,?)); simpl. 
    intros (? & ? & ?); repeat f_equal; auto.
    replace_with H4.
    destruct (a i) as (?,(?,?)); auto.
  Qed.

End easier_direct.

Section facile_rev. 

  Variable (U : Type).
  
  Let X := af_subrel U.
  Let Y := (((U -> Prop) * relation U))%type.

  Let f : X -> Y := @snd _ _.
  Let P : X -> Prop := @afsr_correct _.
  
  Let R : relation X := @lt_afsr _.
  
  Let HR : well_founded_upto f P R.
  Proof.
    apply lt_afsr_upto_wf.
  Qed.

  Definition afs_facile_stump := ((X * nat) * (nat -> X))%type.
  Notation T := afs_facile_stump.
        
  Let f1 (c : X * nat) : Y * nat := (snd (fst c), snd c). 
  Let P1 (c1 : X * nat) := P (fst c1).
  Let R1 (c1 c2 : X * nat) := R (fst c1) (fst c2).
  
  Let R1_wf : well_founded_upto f1 P1 R1.
  Proof.
    intros Q HQ (a,n); unfold P1, f1; simpl.
    intros Ha; revert a Ha n.
    apply (HR (fun x => forall n, Q (x,n))).
    intros a Ha IHa n.
    apply (HQ (a,n)); auto.
    intros (b,m); unfold P1, R1; simpl.
    intros; apply IHa; auto.
  Qed.

  Let R2 (c : X * nat) := @lex_nat_seq_rev _ (snd c) (fun _ => R).  

  Let R2_wf c : well_founded_upto_ext ext (fun s i => f (s i)) (fun s => forall i, P (s i)) (R2 c).
  Proof.
    apply wf_upto_lex_nat_seq_rev.
    intros; auto.
  Qed.
  
  Let R12_wf := wf_upto_lex_dep_prod_ext _ R1_wf R2_wf.

  Definition afs_facile_rev (t1 t2 : T) := 
       R (fst (fst t1)) (fst (fst t2)) 
    \/ fst t1 = fst t2 
    /\ exists i, i < snd (fst t1)
              /\ R (snd t1 i) (snd t2 i)
              /\ forall j, j < i -> snd t1 j = snd t2 j.

  Let ext' (Q : (Y * nat * (nat -> Y)) -> Prop) := forall c, ext (fun s => Q (c,s)).
  Let f' (c : X * nat * (nat -> X)) : Y * nat * (nat -> Y) := (snd (fst (fst c)), snd (fst c), fun i => snd (snd c i)).
  Let P' (c : X * nat * (nat -> X)) := P (fst (fst c)) /\ forall i, P (snd c i).

  Let afs_facile_rev_wf : well_founded_upto_ext ext' f' P' afs_facile_rev.
  Proof.
    intros Q Qext HQ.
    apply R12_wf; auto.
    
    intros a (H1 & H2) H3.
    apply HQ.
    unfold P'; auto.
    intros b (H4 & H5) [ H6 | (H6 & H7) ].

    apply H3; auto.
    destruct a; destruct b.
    left; auto.
    
    destruct a as (a,sa).
    destruct b as (b,sb).
    apply H3; auto.
    right.
    simpl in H6; split; auto.
    destruct H7 as (i & H7 & H8 & H9).
    exists i; auto.
  Qed.

  Definition afs_fs_forget (s : T) : Y * nat * (nat -> Y) := (snd (fst (fst s)), snd (fst s), fun i => snd (snd s i)). 
  
  Definition afs_fs_correct (s : T) := afsr_correct (fst (fst s)) 
                                   /\ forall i, afsr_correct (snd s i). 

  Theorem afs_facile_rev_ind (Q : (Y * nat * (nat -> Y)) -> Prop) :
          (forall y n, ext (fun s => Q (y,n,s)))
       -> (forall a, afs_fs_correct a -> (forall b, afs_fs_correct b -> afs_facile_rev b a -> Q (afs_fs_forget b)) ->  Q (afs_fs_forget a))
       -> (forall a, afs_fs_correct a                                                                              ->  Q (afs_fs_forget a)).
  Proof.
    intros Qext HQ.
    apply afs_facile_rev_wf.
    unfold ext'; intros (y,n) ? ?; apply Qext.
    intros ((x,n),s) (H1 & H2) H3.
    apply HQ.
    red; auto.
    intros ((x',m),s') (H4 & H5) H6.
    apply H3; auto.
    red; auto.
  Qed.

End facile_rev.

Section facile_reverse.

  Variables (X : Type).
  
  Implicit Type (k : nat) (w : nat -> af_witness) (P : nat -> X -> Prop) (R : nat -> relation X).
  
  Variable (Q : nat -> (nat -> X -> Prop) -> (nat -> relation X) -> Prop). 
  Hypothesis (Qext : forall k, ext2 (Q k)).
  
  Definition afs_facile_reverse k1 w1 P1 R1 k2 w2 P2 R2 :=
         lt_afsr (w1 0,(P1 0,R1 0)) (w2 0,(P2 0,R2 0))
      \/ k1 = k2 
      /\ exists i, i < k1 
                /\ lt_afsr (w1 (S i),(P1 (S i),R1 (S i))) (w2 (S i),(P2 (S i),R2 (S i)))
                /\ forall j, j <= i -> w1 j = w2 j /\ P1 j = P2 j /\ R1 j = R2 j.

  Hypothesis HQ : forall k w P R, afs_correct w P R 
                                        -> (forall k' w' P' R', afs_correct w' P' R' 
                                                            -> afs_facile_reverse k' w' P' R' k w P R
                                                            -> Q k' P' R')
                                        -> Q k P R.

  Let Q' (u : ((X -> Prop) * relation X) * nat * (nat -> ((X -> Prop) * relation X))) := 
    forall k P R,   k = snd (fst u)
                 -> (P 0, R 0) = fst (fst u)
                 -> (forall i, (P (S i), R (S i)) = snd u i)
                 -> Q k P R.              

  Let afs_facile_rev_rec : forall s, afs_fs_correct s -> Q' (afs_fs_forget s).
  Proof.
    apply afs_facile_rev_ind.
    
    unfold Q'.
    intros pr n s t Hst H k P R H1 H2 H3.
    apply H; auto.
    intros i; rewrite Hst; auto.
    
    intros (((wk,(Pk,Rk)),k0),s) (H1 & H2) H3 k P R H4 H5 H6.
    simpl in H4; subst k0.
    simpl fst in H1.
    simpl snd in H2.
    unfold afs_fs_forget in H5.
    simpl in H5.
    simpl snd in H6.
    
    apply HQ with (w := fun i => match i with 0 => wk | S i => fst (s i) end).
  
    intros [ | i ].
    rewrite H5; auto.
    rewrite H6, <- surjective_pairing; auto.
    
    intros k' w' P' R' H7 H8.
    
    apply (H3 ((w' 0,(P' 0,R' 0)),k',(fun i => (w' (S i),(P' (S i),R' (S i)))))); auto.
    unfold afs_fs_correct; split; simpl fst.
    apply H7.
    intros; apply H7.
    
    destruct H8 as [ H8 | (H8 & i & H9 & H10 & H11) ].   
    
    left; simpl fst; rewrite <- H5; auto.
    
    subst k'; right.
    split.
    
    simpl.
    rewrite <- H5.
    specialize (H11 0); spec_all H11.
    omega.
    simpl in H11.
    destruct H11 as (? & ? & ?); repeat f_equal; auto.
    
    simpl fst; simpl snd.
    exists i; repeat split; auto.
    rewrite surjective_pairing, <- H6; auto.
    intros j Hj.
    specialize (H11 (S j)).
    spec_all H11.
    omega.
    simpl in H11.
    destruct H11 as (? & ? & ?).
    rewrite surjective_pairing, <- H6.
    repeat f_equal; auto.
  Qed.
  
  Theorem afs_facile_reverse_ind k P R : (forall i, afs (P i) (R i)) -> Q k P R.
  Proof.
    intros HPR.
    apply afs_facile_rev_rec with
      (s := (af_rlift,(P 0, R 0),k,fun i => (af_rlift, (P (S i), R (S i))))); 
      split; simpl; auto.
  Qed.

End facile_reverse.  

Section facile_direct.

  (* I really need a good facile induction principle otherwise 
     it creates too many subcases in the proofs *)

  Variables (X : Type).
  
  Implicit Type (k : nat) (w : nat -> af_witness) (P : nat -> X -> Prop) (R : nat -> relation X).
  
  Variable (Q : nat -> (nat -> X -> Prop) -> (nat -> relation X) -> Prop). 
  Hypothesis (Qext : forall k, ext2 (Q k)).
 
  Definition afs_facile k1 w1 P1 R1 k2 w2 P2 R2 :=
         lt_afsr (w1 k1,(P1 k1,R1 k1)) (w2 k2,(P2 k2,R2 k2))
      \/ k1 = k2 
      /\ exists i, i < k1 
                /\ lt_afsr (w1 i,(P1 i,R1 i)) (w2 i,(P2 i,R2 i))
                /\ forall j, i < j <= k1 -> w1 j = w2 j /\ P1 j = P2 j /\ R1 j = R2 j.

  Hypothesis HQ : forall k w P R,  afs_correct w P R 
                               -> (forall k' w' P' R', afs_correct w' P' R' 
                                                   -> afs_facile k' w' P' R' k w P R
                                                   -> Q k' P' R')
                               ->  Q k P R.

  Theorem afs_facile_ind k P R : (forall i, afs (P i) (R i)) -> Q k P R.
  Proof.
    intros H.
    apply Qext with (r := rev (S k) (rev (S k) P)) (f := rev (S k) (rev (S k) R));
      try (intros; rewrite rev_inv; auto; fail).
    apply afs_facile_reverse_ind with (Q := fun k P R => Q k (rev (S k) P) (rev (S k) R)).
    2: unfold rev; intro; auto.

    assert (forall i, revk (S i) i = 0) as E0.
      unfold revk.
      intros i; destruct (le_lt_dec (S i) i); omega.
    assert (forall i, revk (S i) 0 = i) as E1.
      unfold revk.
      intros; destruct (le_lt_dec (S i) 0); omega.
    assert (forall q i, i < q -> revk (S q) (S (revk q i)) = i) as E2.
      intros q i Hi.
      unfold revk.
      destruct (le_lt_dec q i); try omega.
      destruct (le_lt_dec (S q) (S (q - S i))); omega.
    assert (forall q i, i < q -> revk (S q) i = S (revk q i)) as E3.
      intros q i ?; unfold revk.
      destruct (le_lt_dec q i); try omega.
      destruct (le_lt_dec (S q) i); omega.      
    
    clear k P R H.
    intros k w P R H1 IH1.
    apply HQ with (w := rev (S k) w).
    unfold rev; intro; apply H1.
    intros k' w' P' R' H2 H3. 
    apply Qext with (r := rev (S k') (rev (S k') P')) (f := rev (S k') (rev (S k') R'));
      try (intros; rewrite rev_inv; auto; fail).
    apply IH1 with (w' := rev (S k') w').
    unfold rev; intro; apply H2.

    destruct H3 as [ H3 | (H3 & i & H4 & H5 & H6) ].
   
    left; revert H3; unfold rev; rewrite E0, E1; auto.
    
    subst k'; right; split; auto.
    exists (revk k i).
    split; auto.
    apply revk_prop; auto.    
    split.    
    unfold rev.
    rewrite E2; auto.
    rewrite <- E3; auto.
    intros j Hj; unfold rev.
    specialize (H6 (revk (S k) j)).
    spec_all H6.
    split.
    revert Hj.
    unfold revk.
    destruct (le_lt_dec k i); 
    destruct (le_lt_dec (S k) j); intro; omega.
    revert Hj.
    unfold revk.
    destruct (le_lt_dec k i); 
    destruct (le_lt_dec (S k) j); intro; omega.
    unfold rev in H6.
    rewrite revk_inv in H6.
    auto.
  Qed.

End facile_direct.

