(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Permutation.
Require Import Relations.
Require Import Wf.
Require Import Transitive_Closure.

Require Import rel_utils.

Set Implicit Arguments.

Section rels.

  Fixpoint rel (l : list Type) := 
    match l with 
      | nil  => Prop
      | X::l => X -> rel l
    end.

  Fixpoint rel_incl l : rel l -> rel l -> Prop :=
    match l with 
      | nil  => fun R S => R -> S
      | X::l => fun R S => forall x : X, rel_incl l (R x) (S x)
    end.

  Implicit Arguments rel_incl [ l ].

  Infix "inc" := rel_incl (at level 70, no associativity).

  Infix "~r" := (fun R S => R inc S /\ S inc R) (at level 70, no associativity).

  Fact rel_incl_refl l R : @rel_incl l R R.
  Proof.
    induction l; simpl; auto.
  Qed.
  
  Fact rel_incl_trans l R S T : R inc S -> S inc T -> @rel_incl l R T.
  Proof.
    induction l; simpl; auto.
    intros ? ? x; apply IHl with (S := S x); auto.
  Qed.

  Definition op_mono l m (f : rel l -> rel m) := forall R S, R inc S -> f R inc f S.
  Definition seq_inc l (f : nat -> rel l) := forall n, f n inc f (S n).

  Implicit Arguments op_mono [ l m ].
  Implicit Arguments seq_inc [ l ].

  Hint Resolve rel_incl_refl.

  Fact seq_inc_any l K : @seq_inc l K -> forall n m, n <= m -> K n inc K m.
    Proof.
      intros HK.
      induction 1 as [ | m _ IHm ]; auto.
      apply rel_incl_trans with (1 := IHm); auto.
    Qed.

  Fixpoint lub A l : (A -> rel l) -> rel l :=
    match l with
      | nil  => fun f => exists a, f a
      | X::l => fun f => fun x => lub l (fun a => f a x)
    end.
  
  Definition op_w_cont l m f := forall K, seq_inc K -> f (lub l K) inc lub m (fun n => f (K n)).
  
  Implicit Arguments op_w_cont [ l m ].

  Fact op_mon_w_cont_eq l m f K : op_mono f -> op_w_cont f -> seq_inc K -> f (lub l K) ~r lub m (fun n => f (K n)).
  Proof.
    intros H1 H2 H3; split; auto.
    revert l f K H1 H2 H3.
    induction m as [ | X m IHm ]; simpl; intros l f K Hf1 Hf2 HK.

    intros [ n Hn ]; revert Hn; apply Hf1.
    clear f Hf1 Hf2 HK. 
    induction l; simpl.
    exists n; auto.
    intros x.
    apply IHl with (K := fun n => K n x).
    
    intros x.
    apply (IHm l (fun R => f R x)); auto.
    intros R S H; apply (Hf1 _ _ H x).
    intros K' HK'; apply (Hf2 K' HK').
  Qed.

End rels.

Infix "inc" := (@rel_incl _) (at level 70, no associativity).

Infix "~r" := (fun R S => R inc S /\ S inc R) (at level 70, no associativity).

Notation R1 := (_::nil) (only parsing).
Notation R2 := (_::_::nil) (only parsing).
Notation R3 := (_::_::_::nil) (only parsing).

Section rel_props.

  Variable X : Type.

  Implicit Types (R S : relation X) (K : nat -> relation X).

  Section defs. 

    Variable (Y : Type).
 
    Variable f : relation X -> relation Y.
    
    Definition op_monotonic := forall R S, R inc2 S -> f R inc2 f S.
    Definition seq_increasing K := forall n, K n inc2 K (S n). 
    Definition op_w_continuous := forall K, seq_increasing K -> f (fun x y => exists n, K n x y) inc2 fun x y => exists n, f (K n) x y.

    Fact seq_inc_all K : seq_increasing K -> forall n m, n <= m -> K n inc2 K m.
    Proof.
      intros HK.
      induction 1; auto.
    Qed.

    Hypothesis H1 : op_monotonic.
    Hypothesis H2 : op_w_continuous.

    Fact op_mono_w_cont_eq K a b : seq_increasing K -> (f (fun x y => exists n, K n x y) a b <-> exists n, f (K n) a b).
    Proof.
      intros H; split.
      intros; apply H2; auto.
      intros (n & Hn); revert Hn; apply H1.
      exists n; auto.
    Qed.

  End defs.

  Fact clos_trans_mono : op_monotonic (clos_trans X).
  Proof.
    intros R S H x y.
    induction 1 as [ | ? y ].
    constructor 1; auto.
    constructor 2 with y; auto.
  Qed.
  
  Fact clos_trans_w_cont : op_w_continuous (clos_trans X).
  Proof.
    intros K HK x y.
    induction 1 as [ x y (n & Hxy) | x y z H1 (n & H2) H3 (m & H4)  ].
    exists n; constructor 1; auto.
    exists (n+m).
    constructor 2 with y.
    revert H2; apply clos_trans_mono, seq_inc_all with (1 := HK), le_plus_l.
    revert H4; apply clos_trans_mono, seq_inc_all with (1 := HK), le_plus_r.
  Qed.

End rel_props.

Section op_comp.

  Variable (X Y Z : Type) (f : relation X -> relation Y) (g : relation Y -> relation Z).
  
  Hypothesis Hf1 : op_monotonic f.
  Hypothesis Hf2 : op_w_continuous f.
  
  Hypothesis Hg1 : op_monotonic g.
  Hypothesis Hg2 : op_w_continuous g.
  
  Fact op_comp_monotonic : op_monotonic (fun R => g (f R)).
  Proof.
    intros ? ? H; apply Hg1, Hf1, H.
  Qed. 

  Fact op_comp_w_continuous : op_w_continuous (fun R => g (f R)).
  Proof.
    intros K HK x y H.
    rewrite <- op_mono_w_cont_eq; auto.
    revert H; apply Hg1.
    clear x y; intros x y.
    apply Hf2; auto.
    intros n; apply Hf1, HK.
  Qed.

End op_comp.

Section list_sup.

  Variables (X : Type) 
            (P : nat -> X -> Prop) (P_mono : forall n x, P n x -> P (S n) x)
            (R : nat -> relation X) (R_mono : forall n, R n inc2 R (S n)).

  Let P_monotonic n m : n <= m -> forall x, P n x -> P m x.
  Proof.
    induction 1; intros; auto.
  Qed.

  Let R_monotonic n m : n <= m -> R n inc2 R m.
  Proof.
    induction 1; intros; auto.
  Qed.

  Fact list_Forall_sup ll : (forall x, In x ll -> exists n, P n x) -> exists n, forall x, In x ll -> P n x.
  Proof.
    induction ll as [ | x ll IH ]; intros H.
    exists 0; intros ? [].
    destruct (H x) as (n1 & H1).
    left; auto.
    destruct IH as (n2 & H2).
    intros; apply H; right; auto.
    exists (n1+n2); intros y [ Hy | Hy ].
    subst; revert H1; apply P_monotonic, le_plus_l.
    generalize (H2 _ Hy); apply P_monotonic, le_plus_r.
  Qed.

  Fact list_Forall2_mono Y Z (S T : Y -> Z -> Prop) : S inc2 T -> Forall2 S inc2 Forall2 T.
  Proof.
    intros ? ? ?; induction 1; constructor; auto.
  Qed.

  Fact list_Forall2_sup ll mm : Forall2 (fun x y => exists n, R n x y) ll mm -> exists n, Forall2 (R n) ll mm.
  Proof.
    induction 1 as [ | x y ll mm (n1 & H1) _ (n2 & IH2) ].
    exists 0; auto.
    exists (n1+n2); constructor.
    revert H1; apply R_monotonic, le_plus_l.
    revert IH2; apply list_Forall2_mono, R_monotonic, le_plus_r.
  Qed.

End list_sup.

Section lfp.

  Variables (X : Type) (f : relation X -> relation X) 
                       (Hf1 : op_monotonic f)
                       (Hf2 : op_w_continuous f).

  Let Rn := fix loop n := 
    match n with 
      | 0   => fun _ _ => False
      | S n => f (loop n)
    end.

  Let Rn_inc n : Rn n inc2 Rn (S n).
  Proof.
    induction n; simpl.
    intros _ _ [].
    apply Hf1; auto.
  Qed.

  Let Rn_monotonic n m : n <= m -> Rn n inc2 Rn m.
  Proof. induction 1; auto. Qed.
  
  Definition lfp := fun x y => exists n, Rn n x y.

  Section lfp_induction.

    Variable P : relation X -> Prop.

    Hypothesis P_0 : P (fun _ _ => False).
    Hypothesis P_S : forall S, P S -> P (f S).
    Hypothesis P_w_cont : forall K, seq_increasing K -> (forall n, P (K n)) -> P (fun x y => exists n, K n x y).

    Theorem lfp_induction : P lfp.
    Proof.
      apply P_w_cont; auto.
      induction n; simpl; auto.
    Qed.

  End lfp_induction.

  Section lfp_least.
 
    Variable R : relation X.
    Hypothesis R_closed : forall x y, f R x y -> R x y.
    
    Theorem lfp_least : lfp inc2 R.
    Proof.
      apply lfp_induction.
      intros ? ? [].
      intros S HS ? ? ?; apply R_closed, (Hf1 _ HS); auto.
      intros K _ HK x y (n & Hn); revert Hn; apply HK.
    Qed.

  End lfp_least.

  Lemma lfp_post_fp : f lfp inc2 lfp.
  Proof.
    intros x y H.
    unfold lfp in H.
    rewrite op_mono_w_cont_eq in H; auto.
    destruct H as (n & ?).
    exists (S n); auto.
  Qed.

  Corollary lfp_pre_fp : lfp inc2 f lfp.
  Proof.
    apply lfp_least, Hf1, lfp_post_fp.
  Qed.

  Hint Immediate lfp_post_fp lfp_pre_fp.
  
  Theorem lfp_fixpoint_eq x y : lfp x y <-> f lfp x y.
  Proof.
    split; auto.
  Qed.

  Section lfp_ind.
  
    Variable R : relation X.
    Hypothesis R_closed : forall x y,  f (lfp cap2 R) x y -> R x y.

    Theorem lfp_ind : lfp inc2 R.
    Proof.
      assert (lfp inc2 lfp cap2 R) as H.
        apply lfp_least.
        intros x y Hxy; split; auto.
        rewrite lfp_fixpoint_eq.
        revert Hxy; apply Hf1; tauto.
      intros ? ? Hx; generalize (H _ _ Hx); tauto.
    Qed.

  End lfp_ind.

  Section lfp_refl.

    Hypothesis Hf : @eq _ inc2 f (fun _ _ => False).
    
    Fact lfp_refl : reflexive _ lfp.
    Proof.
      exists 1; apply Hf; auto.
    Qed.

  End lfp_refl.

  Section lfp_trans.

    Hypothesis Hf : forall R S, f R o f S inc2 f (R o S).

    Let lft_trans_rec n : Rn n o lfp inc2 lfp.
    Proof.
      induction n as [ | n IHn ]; simpl.
      intros ? ? (? & [] & _).
      intros x z (y & H1 & H2).
      rewrite lfp_fixpoint_eq in H2.
      rewrite lfp_fixpoint_eq.
      apply (Hf1 _ IHn).
      apply (@Hf (Rn n) lfp).
      exists y; auto.
    Qed.      

    Fact lfp_trans : transitive _ lfp.
    Proof.
      intros ? y ? (n & ?) ?.
      apply lft_trans_rec with n.
      exists y; auto.
    Qed.
 
  End lfp_trans.

  Variable R : relation X.

  Implicit Type S : relation X.

  Hypothesis HR : forall S, (forall x u v, R u v -> S x u -> S x v)
                         -> (forall x u v, R u v -> f S x u -> f S x v).

  Fact lfp_has_prop x u v : R u v -> lfp x u -> lfp x v.
  Proof.
    revert x u v.
    apply lfp_induction; auto.
    intros K HK H x u v H1 (n & H2); exists n; auto.
    apply H with (1 := H1); auto.
  Qed.

End lfp.

Section lfp_props.

  Variable (X : Type).
  
  Fact lfp_mono (f g : _ -> relation X) :
                 op_monotonic f 
              -> op_monotonic g
              -> op_w_continuous g 
              -> (forall R, f R inc2 g R) 
              -> lfp f inc2 lfp g.
  Proof.
    intros H1 H2 H3 H4.
    apply lfp_least; auto.
    intros x y H.
    apply H4 in H.
    revert H.
    apply lfp_post_fp; auto.
  Qed.

End lfp_props.


