(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Set Implicit Arguments.

Definition finite_repr X (P : X -> Prop) ll := forall x, In x ll <-> P x.
Definition finite_inverse X Y (f : X -> Y) y ll := forall x, In x ll <-> y = f x.

Definition finite X (P : X -> Prop) := exists ll, forall x, In x ll <-> P x.

Definition finite_t X (P : X -> Prop) := { ll | forall x, In x ll <-> P x }.

Fact finite_inhabited X P : @finite X P <-> inhabited (finite_t P).
Proof.
  split.
  intros (l & ?); exists; exists l; auto.
  intros [ (l & ?) ]; exists l; auto.
Qed.

Fact finite_t_finite X P : @finite_t X P -> finite P.
Proof. intros [ x Hx ]; exists x; intro; rewrite Hx; split; auto. Qed.

Section finite_nat.

  Fact finite_t_lt n : finite_t (fun x => x < n).
  Proof.
    exists (list_n n).
    intros; symmetry; apply list_n_prop.
  Qed.
  
  Fact finite_lt n : finite (fun x => x < n).
  Proof.
    apply finite_t_finite, finite_t_lt.
  Qed.
 
End finite_nat.

Fact finite_t_eq X (P Q : X -> Prop) : P ~eq1 Q -> finite_t P -> finite_t Q.
Proof.
  intros H1; intros (ll & H2); exists ll; intros x; split; intros H3.
  apply H1, H2; auto.
  apply H2, H1; auto.
Qed.

Local Ltac solve H := repeat rewrite finite_inhabited; inhabited_tac; apply H; auto.
    
Fact finite_eq X (P Q : X -> Prop) : P ~eq1 Q -> finite P -> finite Q.
Proof. intro; solve finite_t_eq. Qed.
  
Fact finite_t_img (X Y : Type) (f : X -> Y) x : finite_t (fun y => y = f x).
Proof.
  exists ((f x)::nil).
  intro; simpl; split; auto.
  intros [ | []]; auto.
Qed.

Fact finite_img (X Y : Type) (f : X -> Y) x : finite (fun y => y = f x).
Proof. solve finite_t_img. Qed. 

Fact finite_t_cup (X : Type) (A B : X -> Prop) : finite_t A -> finite_t B -> finite_t (A cup1 B).
Proof.
  intros (lA & H1) (lB & H2); exists (lA++lB).
  intros x; rewrite <- H1, <- H2.
  split.
  apply in_app_or.
  apply in_or_app.
Qed.

Fact finite_cup (X : Type) (A B : X -> Prop) : finite A -> finite B -> finite (A cup1 B).
Proof. solve finite_t_cup. Qed.

Section finite_comp.

  Variable (X Y Z : Type) (R : X -> Y -> Prop) (S : Y -> Z -> Prop).
  
  Variable (x : X).  

  Section finite.

    Hypothesis HR : finite (R x).
    Hypothesis HQ : forall y, R x y -> finite (S y).
    
    Fact finite_comp : finite ((R o S) x).
    Proof.
      destruct HR as (ly & Hly).
      assert (HQ' : forall y, In y ly -> finite (S y)).
        intro; rewrite Hly; auto.
      apply sig_invert in HQ'.
      destruct HQ' as (mm & Hmm).
      exists (flat_map (fun x => x) mm).
      
      intros z; rewrite in_flat_map; split.
      
      intros (lz & H1 & H2).
      destruct (Forall2_In_inv_right Hmm _ H1) as (y & H3 & H4).
      exists y; rewrite <- H4, <- Hly; auto.
      
      intros (y & H1 & H2).
      rewrite <- Hly in H1.
      destruct (Forall2_In_inv_left Hmm _ H1) as (kk & H3 & H4).
      exists kk; rewrite H4; auto.     
    Qed.
    
  End finite.
  
  Section finite_t.

    Hypothesis HR : finite_t (R x).
    Hypothesis HQ : forall y, R x y -> finite_t (S y).
    
    Fact finite_t_comp : finite_t ((R o S) x).
    Proof.
      destruct HR as (ly & Hly).
      assert (HQ' : forall y, In_t y ly -> finite_t (S y)).
        intros y Hy.
        apply In_t_In in Hy.
        rewrite Hly in Hy; auto.
      apply sig_t_invert in HQ'.
      destruct HQ' as (mm & Hmm).
      exists (flat_map (fun x => x) mm).
      
      intros z; rewrite in_flat_map; split.
      
      intros (lz & H1 & H2).
      destruct (Forall2_In_inv_right Hmm _ H1) as (y & H3 & H4).
      exists y; rewrite <- H4, <- Hly; auto.
      
      intros (y & H1 & H2).
      rewrite <- Hly in H1.
      destruct (Forall2_In_inv_left Hmm _ H1) as (kk & H3 & H4).
      exists kk; rewrite H4; auto.    
    Qed.
    
  End finite_t.
    
End finite_comp.

Section finite_product.

  Variable (X Y : Type) (P : X -> Prop) (Q : Y -> Prop).
  
  Fact finite_t_product : finite_t P -> finite_t Q -> finite_t (pred_prod P Q).
  Proof.
    intros (l & Hl) (m & Hm).
    exists (list_prod pair l m).
    intros (x,y); rewrite list_prod_spec; split.
    intros (a & b & H1 & H2 & H3); injection H3; intros ? ?; subst.
    red; rewrite <- Hl, <- Hm; auto.
    unfold pred_prod; rewrite <- Hl, <- Hm; simpl.
    intros []; exists x, y; auto.
  Qed.

  Fact finite_product : finite P -> finite Q -> finite (pred_prod P Q).
  Proof. solve finite_t_product. Qed.

End finite_product.

Section finite_inverse.

   Variable (X Y : Type).
   Implicit Type (f : X -> Y).
   
   Fact finite_inverse_map_t f ll (fi_f : forall x, In_t x ll -> list X) :
                                    (forall x Hx, finite_inverse f x (fi_f x Hx))
                                  -> finite_inverse (map f) ll (list_fan (list_In_t_map _ fi_f)).
   Proof.
     intros Hf fl; split.
     intros Hfl.
     apply list_fan_spec in Hfl.
     revert fl Hfl;
     induction ll as [ | x ll IHll ]; intros [ | y fl ] Hfl.
     auto.
     inversion Hfl.
     inversion Hfl.
     simpl in Hfl.
     apply Forall2_cons_inv in Hfl.
     destruct Hfl as [ H1 H2 ].
     simpl; f_equal.
     apply Hf in H1; auto.
     apply IHll with (2 := H2).
     intros; apply Hf.
     intros; subst. 
     apply list_fan_spec.
     induction fl; simpl; constructor; auto.
     apply Hf; auto.
   Qed.

End finite_inverse.

Section finite_rel_inv.

  Variable (X Y : Type) (R : X -> Y -> Prop).
  
  Fact finite_rel_inv ll : (forall x, In x ll -> finite (R x))
                         -> finite (Forall2 R ll).
  Proof.
    intros H.
    apply sig_invert in H.
    destruct H as (mm & Hmm).
    exists (list_fan mm).
    intros x.
    rewrite list_fan_spec.
    split.

    intros H; rewrite Forall2_exchg in H.
    generalize (Forall2_compose Hmm H); apply Forall2_impl.
    intros a b (c & H1 & H2); rewrite <- H1; auto.
    
    intros H; rewrite Forall2_exchg in H.
    generalize (Forall2_compose H Hmm); apply Forall2_impl.
    intros a b (c & H1 & H2); rewrite H2; auto.
  Qed.
  
End finite_rel_inv.

Notation finite_Forall2 := finite_rel_inv.

Section finite_t_rel_inv.

  Variable (X Y : Type) (R : X -> Y -> Prop).
  
  Fact finite_t_rel_inv ll : (forall x, In_t x ll -> finite_t (R x))
                         -> finite_t (Forall2 R ll).
  Proof.
    intros H.
    apply sig_t_invert in H.
    destruct H as (mm & Hmm).
    exists (list_fan mm).
    intros x.
    rewrite list_fan_spec.
    split.

    intros H; rewrite Forall2_exchg in H.
    generalize (Forall2_compose Hmm H); apply Forall2_impl.
    intros a b (c & H1 & H2); rewrite <- H1; auto.
    
    intros H; rewrite Forall2_exchg in H.
    generalize (Forall2_compose H Hmm); apply Forall2_impl.
    intros a b (c & H1 & H2); rewrite H2; auto.
  Qed.

End finite_t_rel_inv.

Section list_rel_invert.

  Variables (X Y : Type) (R : X -> Y -> Prop) (ly : list Y).
  Hypothesis HQ : forall y, In_t y ly -> finite_t (fun x => R x y).
  
  Definition rel_list_invert : finite_t (fun lx => Forall2 R lx ly).
  Proof.
    destruct finite_t_rel_inv with (ll := ly) (R := fun x y => R y x) as (mm & Hmm).
    intros y Hy; destruct (HQ _ Hy) as (lx & Hlx).
    exists lx; intro; rewrite Hlx; split; auto.
    exists mm; intro; rewrite Hmm.
    rewrite Forall2_exchg; split; auto.
  Qed.

End list_rel_invert.

Section list_choice.

  Variable (X Y : Type) (R : X -> Y -> Prop).

  Fact list_choice ly : (forall y, In y ly -> exists x, R x y)
                      -> exists lx, Forall2 R lx ly.
  Proof.
    induction ly as [ | y ly IH ]; simpl; intros Hly.
    exists nil; constructor.
    destruct (Hly _ (or_introl eq_refl)) as (x & Hx).
    destruct IH as (lx & Hlx).
    intros; apply Hly; auto.
    exists (x::lx); constructor; auto.
  Qed.  

  Fact list_choice_t ly : (forall y, In_t y ly -> { x | R x y })
                       -> { lx | Forall2 R lx ly }.
  Proof.
    induction ly as [ | y ly IH ]; simpl; intros Hly.
    exists nil; constructor.
    destruct (Hly _ (inl eq_refl)) as (x & Hx).
    destruct IH as (lx & Hlx).
    intros; apply Hly; auto.
    exists (x::lx); constructor; auto.
  Qed.

End list_choice.  

Section list_rel_invert'.

  Variables (X Y : Type) (R : X -> Y -> Prop) (P : X -> Prop) (ly : list Y).
  Hypothesis HQ : forall y, In_t y ly -> finite_t (fun x => P x /\ R x y).
  
  Definition rel_list_invert' : finite_t (fun lx => Forall P lx /\ Forall2 R lx ly).
  Proof.
    set (R' x y := P x /\ R x y).
    destruct (rel_list_invert R' ly) as (ll & Hll).
    apply HQ.
    exists ll.
    intros lx; rewrite Hll.
    unfold R'.
    rewrite Forall2_conj.
    rewrite Forall2_left_Forall.
    
    split; intros H;
    repeat split; try tauto.
    apply proj2 in H; revert H.
    apply Forall2_length.
  Qed.
 
End list_rel_invert'.

Section finite_utils.
  
  Variable X Y : Type.

  Fact finite_cap_dec P Q : finite P -> (forall x : X, Q x \/ ~ Q x) -> finite (P cap1 Q).
  Proof.
    intros (ll & Hll) HQ.
    revert P Hll.
    induction ll as [ | x ll IH ]; intros P Hll.

    exists nil.
    intros x; split.
    intros [].
    intros (? & _); apply Hll; auto. 
    
    destruct (IH (fun x => In x ll)) as (mm & Hmm).
    split; auto.
    destruct (HQ x) as [ Hx | Hx ].

    exists (x::mm).
    intros z; split; rewrite <- Hll.
    intros [ H1 | H1 ]; subst; simpl; auto.
    rewrite Hmm in H1; tauto.
    intros ([ H1 | H1 ] & H2).
    left; auto.
    right; auto.
    apply Hmm; auto.
    
    exists mm.
    intros z; split; rewrite <- Hll.
    intros H1; simpl; auto.
    rewrite Hmm in H1; tauto.
    intros ([ H1 | H1 ] & H2).
    subst; tauto.
    apply Hmm; auto.
  Qed.

  Fact finite_t_cap_dec P Q : finite_t P 
                           -> (forall x : X, { Q x } + { ~ Q x }) 
                           -> finite_t (P cap1 Q).
  Proof.
    intros (ll & Hll) HQ.
    revert P Hll.
    induction ll as [ | x ll IH ]; intros P Hll.

    exists nil.
    intros x; split.
    intros [].
    intros (? & _); apply Hll; auto.
    
    destruct (IH (fun x => In x ll)) as (mm & Hmm).
    split; auto.
    destruct (HQ x) as [ Hx | Hx ].

    exists (x::mm).
    intros z; split.
    intros [ H1 | H1 ]; subst; simpl; auto.
    rewrite <- Hll; simpl; tauto.
    rewrite Hmm in H1.
    rewrite <- Hll; simpl; tauto.
    rewrite <- Hll.
    intros ([ H1 | H1 ] & H2).
    left; auto.
    right; auto.
    apply Hmm; auto.
    
    exists mm.
    intros z; split; rewrite <- Hll.
    intros H1; simpl; auto.
    rewrite Hmm in H1; tauto.
    intros ([ H1 | H1 ] & H2).
    subst; tauto.
    apply Hmm; auto. 
  Qed.

  Fact finite_list_inv (f : X -> Y) Q ll : (forall y, In y ll -> finite (fun x => Q x /\ f x = y))
                                         -> finite (fun mm => ll = map f mm /\ Forall Q mm).
  Proof.
    intros H.
    apply finite_rel_inv in H.
    revert H.
    apply finite_eq.
    split; intros x;
    rewrite Forall2_exchg, Forall2_analysis; 
    intros [ ]; split; auto.
  Qed.

  Variable P : X -> Y -> Prop.

  Fact finite_list_repr ll : (forall x, In x ll -> finite (P x)) -> exists mm, Forall2 (fun x ly => finite_repr (P x) ly) ll mm.
  Proof.
    intros H.
    apply sig_invert in H.
    destruct H as (mm & Hmm).
    exists mm.
    revert Hmm; apply Forall2_impl.
    unfold finite_repr; auto.
  Qed.
    
End finite_utils.

Section finite_list_decide.

  Variable (X Y : Type) (P : X -> Y -> Prop) (Q : list Y -> Prop) (R : Y -> Prop).
  
  Fact finite_list_decide ll : (forall x, In x ll -> finite (P x))
                            -> (forall mm, Forall2 P ll mm -> Q mm \/ exists y, In y mm /\ R y)
                            -> (exists mm, Forall2 P ll mm /\ Q mm) 
                            \/ (exists x, In x ll /\ forall y, P x y -> R y).
  Proof.
    intros HP HPQR.
    destruct (finite_list_repr P ll HP) as (mm & Hmm).
    clear HP.
    destruct (@hig_combi_principle _ mm Q R) as [ H | H ].

    intros p H; apply HPQR; revert p H.
    clear HPQR.
    induction Hmm as [ | x ly ll mm H1 H2 H3 ]; intros p H.
    inversion_clear H; constructor.
    apply Forall2_cons_inv_r in H.
    destruct H as (y & pp & G1 & G2 & G3).
    subst p.
    constructor; auto.
    apply H1; auto.
    
    clear HPQR.    
    destruct H as (kk & H1 & H2).
    left; exists kk; split; auto.
    clear H2.
    revert kk H1.
    induction Hmm as [ | x ly ll mm H1 H2 H3 ]; intros kk H0.
    inversion_clear H0; constructor.
    apply Forall2_cons_inv_r in H0.
    destruct H0 as (y & pp & G1 & G2 & G3).
    subst kk.
    constructor; auto.
    apply H1; auto.
    
    clear HPQR.    
    destruct H as (ly & H1 & H2).
    right.
    destruct (Forall2_In_inv_right Hmm _ H1) as (x & H3 & H4).
    exists x; split; auto.
    intros y Hy; apply H2, H4; auto.
  Qed.

End finite_list_decide. 

Section finite_decide.

  Variable (X : Type) (Q R S : X -> Prop).
  
  Hypothesis HQ1 : finite Q.
  Hypothesis HQ2 : forall x, Q x -> R x \/ S x.
  
  Fact finite_decide : (forall x, Q x -> R x) \/ (exists x, Q x /\ S x).
  Proof.
    destruct HQ1 as (ll & Hll).
    destruct (list_decide_ind ll (fun x _ => S x) (fun x _ => R x)) as [ (x & H1 & H2) | H ].
    intros x Hx; apply In_t_In in Hx; apply Hll in Hx; specialize (HQ2 Hx); tauto.
    apply In_t_In in H1; right; exists x; split; auto; apply Hll; auto.
    left; intros x Hx; apply Hll in Hx; revert Hx; apply In_t_P; auto.
  Qed.

End finite_decide.
    
Section finite_prod_decide.

  Variable (X Y : Type) (P : X -> Prop) (Q : X -> Y -> Prop).
  
  Variable R S : X -> Y -> Prop.

  Hypothesis HP : finite P.
  Hypothesis HQ : forall x, P x -> finite (Q x).

  Fact finite_prod_decide :  (forall x, P x -> forall y, Q x y -> R x y \/ S x y)
                          -> (forall x, P x -> exists y, Q x y /\ R x y)
                          \/ (exists x, P x /\ forall y, Q x y -> S x y).
  Proof.
    intros HRS.
    set (R' x := exists y, Q x y /\ R x y).
    set (S' x := forall y, Q x y -> S x y).
    destruct (@finite_decide _ P R' S' HP) as [ H | H ].    
    unfold R', S'.
    intros x Hx.
    destruct (@finite_decide _ (Q x) (S x) (R x) (HQ Hx)); auto.
    intros y Hy; destruct (HRS x Hx y Hy); tauto.
    left; auto.
    right; auto.
  Qed.

End finite_prod_decide.

Section finite_combi.

  Variable (X Y : Type) (A : X -> Y -> Prop) (P : list Y -> Prop) (B : X -> Y -> Prop) (lx : list X).

  Hypothesis HA : forall x, In x lx -> finite (A x).
  Hypothesis HPB : forall ly, Forall2 A lx ly -> P ly \/ Exists2 B lx ly.
  
  Theorem finite_combi : (exists ly, Forall2 A lx ly /\ P ly)
                      \/ (exists x, In x lx /\ A x inc1 B x).
  Proof.
    revert P HA HPB.
    induction lx as [ | x ll IH ]; intros P HA HPB.

    left; exists nil; split.
    constructor.
    destruct (@HPB nil); auto.
    apply Exists2_nil_inv in H.
    contradiction H.
   
    set (Q ly := Forall2 A (x::ll) ly).
    set (R ly := Exists2 B (x::ll) ly).
    set (S ly := P ly).    
    destruct (@finite_decide _ Q R S) as [ HPB' | H ].    
    apply finite_rel_inv; auto.
    intros a Ha; destruct (HPB Ha); tauto.
    2: left; auto.    
    unfold Q, R, S in HPB' |- *; clear Q R S.    
  
    set (Q y pp := Forall2 A (x::ll) (y::pp)).
    set (R y (_ : list Y) := B x y).
    set (S (_ : Y) pp := Exists2 B ll pp).
    destruct (@finite_prod_decide _ _ (A x) Q R S) as [ H | H ].

    apply HA; left; auto.
    unfold Q.
    intros. 
    apply finite_eq with (P := fun pp => Forall2 A ll pp).
    2: apply finite_rel_inv; intros; apply HA; right; auto.
    split; intros pp Hpp.
    constructor; auto.
    apply Forall2_cons_inv in Hpp; tauto.
    
    unfold Q, R, S.
    intros y Hy pp Hpp.
    apply HPB' in Hpp.
    apply Exists2_cons_inv in Hpp; tauto.
    
    Focus 2.
    unfold Q, S in H.
    destruct H as (y & H1 & H2).
    destruct (IH (fun ly => P (y::ly))) as [ (ly & H3 & H4) | (z & H3 & H4) ].
    intros; apply HA; right; auto.
    right; apply H2; constructor; auto.
    left; exists (y::ly); split; auto.
    right; exists z; split; auto; right; auto.
        
    unfold Q, R in H.
    clear Q R S.
    right; exists x; split.
    left; auto.
    intros y Hy.
    destruct (H _ Hy) as (? & _ & ?); auto.
  Qed.

End finite_combi.

Section product_analysis.
  
  Variable (X : Type) (P Q R : X -> Prop).
  
  Hypothesis (HQ : finite Q) (HR : finite R).

  Fact product_analysis :  (forall c, pred_prod Q R c -> P (fst c) \/ P (snd c)) 
                         -> Q inc1 P \/ R inc1 P.
  Proof.
    intros H0.
    set (A b := match b with true => Q | false => R end).
    destruct (@finite_combi _ _ A (fun _ => False) (fun _ => P) (true::false::nil)) as [ H | H ].
    
    intros []; simpl; auto.

    intros [ | a [ | b [ | ? ? ] ] ]; intros H;
      try (repeat (apply Forall2_cons_inv, proj2 in H); inversion H; fail).
    apply Forall2_cons_inv in H; destruct H as (H1 & H).
    apply Forall2_cons_inv in H; destruct H as (H2 & _).
    simpl in H1, H2.
    right.
    specialize (H0 (a,b) (conj H1 H2)).
    destruct H0.
    constructor 1; auto.
    constructor 2; constructor 1; auto.
    
    destruct H as (_ & _ & []).
    
    destruct H as ( [] & _ & H ).
    left; auto.
    right; auto.
  Qed.

End product_analysis.

Section finite_analysis.

  Variable (X Y : Type) (A : X -> Y -> Prop) (P : Y -> Prop).
  
  Fact finite_analysis ll : (forall x, In x ll -> finite (A x))
                         -> (forall pp, Forall2 A ll pp -> exists y, In y pp /\ P y)
                         -> exists x, In x ll /\ forall y, A x y -> P y.
  Proof.
    intros H1 H2.
    destruct (@finite_combi _ _ A (fun _ => False) (fun _ y => P y) ll) as [ H | H ]; auto.
    intros pp Hpp; right.
    apply Exists2_exists_right.
    split; revert Hpp; auto. 
    apply Forall2_length.
    destruct H as (? & _ & []).
  Qed.

End finite_analysis.

Section finite_t_build.

  Variable (X : Type).

  Implicit Type (Q : X -> Prop).
  
  Let local_find Q l :  (forall x, In x l -> { Q x } + { ~ Q x })
              -> (exists x, In x l /\ Q x)
              -> { x | In x l /\ Q x }.
  Proof.
    induction l as [ | x l IH ]; intros H0.
    intros H; exfalso; destruct H as (? & [] & _).
    destruct (H0 x) as [ H | H ].
    left; auto.
    exists x; simpl; auto.
    intros H3.
    destruct IH as (u & H1 & H2).
    intros; apply H0; right; auto.
    destruct H3 as (u & [ H3 | H3 ] & H4).
    contradict H; subst; auto.
    exists u; auto.
    exists u; simpl; auto.
  Qed.

  Variables (P Q : X -> Prop) (HP : finite_t P).
  Hypothesis HQ  : forall x, P x -> { Q x } + { ~ Q x }.
  Hypothesis HPQ : exists x, P x /\ Q x.
  
  Fact finite_t_build : { x | P x /\ Q x }.
  Proof.
    destruct HP as (ll & Hll).
    destruct (@local_find Q ll) as (x & H1 & H2).
    intros x Hx; rewrite Hll in Hx; auto.
    destruct HPQ as (x & Hx).
    rewrite <- Hll in Hx.
    exists x; auto.
    exists x; rewrite <- Hll; auto.
  Qed.

End finite_t_build.

Section finite_not.

  Variable (X : Type) (P Q : X -> Prop).
  
  Hypothesis HQ : forall x, P x -> Q x \/ ~ Q x.

  Fact finite_not :  finite P 
                  -> ~ (forall x, P x -> Q x) 
                  -> exists x, P x /\ ~ Q x.
  Proof.
    intros (l & Hl) H.
    assert (~ forall x, In x l -> Q x) as H1.
      contradict H; intros x; rewrite <- Hl; auto.
    clear H; rename H1 into H.
    assert (forall x, In_t x l -> ~ Q x \/ Q x) as HQ'.
      rewrite In_In_t; intros x; rewrite Hl; specialize (@HQ x); tauto.
    clear HQ; rename HQ' into HQ.
    
    apply list_decide_ind in HQ.
    destruct HQ as [ (z & H1 & H2) | HQ' ].
    apply In_t_In in H1.
    exists z; rewrite <- Hl; auto.
    rewrite In_In_t in HQ'.
    contradict H; auto.
  Qed.

End finite_not.

Section finite_not_build.

  Variable (X : Type) (P Q : X -> Prop) (HP : finite_t P).
  Hypothesis HQ1  : forall x, P x -> { Q x } + { ~ Q x }.
  Hypothesis HQ2 : ~ forall x, P x -> Q x.

  Fact finite_t_not_build : { x | P x /\ ~ Q x }.
  Proof.
    apply finite_t_build; auto.
    intros x Hx; destruct (HQ1 Hx); tauto.
    apply finite_not; auto.
    intros x Hx; destruct (HQ1 Hx); tauto.
    apply finite_t_finite; auto.
  Qed.

End finite_not_build.


Fact finite_nat_decide n (P : nat -> Prop) :  (forall i, i < n -> { P i } + { ~ P i })
                                           -> { i | i < n /\ P i } + { forall i, i < n -> ~ P i }.
Proof.
  
  induction n as [ | n IHn ]; intros HP.
  right; intros; omega.
  destruct (HP n) as [ Hn | Hn ].
  omega.
  left; exists n; split; auto; omega.
  destruct IHn as [ (i & H1 & H2) | IHn ].
  intros; apply HP; omega.
  left; exists i; split; auto; omega.
  right.
  intros.
  destruct (le_lt_dec n i); auto.
  cutrewrite (i = n); auto; omega.
Defined.  


    
    
