(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.

Require Import wf_utils.
Require Import rel_utils.
Require Import list_utils.

Require Import sublist.
Require Import good_base.

Require Import bar.

Set Implicit Arguments.

Section Bar_Good.

  Variables (A : Type) (R : A -> A -> Prop).

  Infix "<<" := R (at level 70).

  (* every infinite sequence contains a Good prefix *)

  (* BarWToGoodInit *)

  Theorem bar_good_nil_seq : bar (good R) nil -> forall f, exists m, good R (pfx_rev f m).
  Proof.
    apply bar_inv.
  Defined.

  Definition bar_good_nil_all : bar (good R) nil -> forall ll, bar (good R) ll. 
  Proof.
    apply bar_nil_all.
    intros ? ?; apply in_good_1.
  Defined.

  Fact bar_good_app ll mm : bar (good R) mm -> bar (good R) (ll++mm).
  Proof.
    apply bar_app.
    intros; apply in_good_1; auto.
  Qed.

  Fact sublist_bar_good ll mm : ll <sl mm -> bar (good R) ll -> bar (good R) mm.
  Proof.
    intros H1 H2; revert H2 mm H1.
    induction 1 as [ ll Hll | ll Hll IH]; intros mm Hmm.
    constructor 1; revert Hll; apply sublist_good; auto.
    constructor 2; intros a.
    apply (IH a); constructor 2; auto.
  Qed.

  Fact bar_good_refl a : bar (good R) nil -> a << a.
  Proof.
    intros H.
    apply ctxt_nil.
    revert H.
    generalize (@nil A).
    induction 1 as [ ll Hll | ll Hll IH ] using bar_ind.
    left; auto.
    specialize (IH a).
    apply ctxt_cons in IH; tauto.
  Qed.

  Section bar_lift_rel.
  
    Let lemma1 ll a : bar (good (R rlift a)) ll -> bar (good R) (ll++a::nil).
    Proof.
      induction 1 using bar_ind.
      apply in_bar_0; apply good_lift_rel; auto.
      apply in_bar_1; auto.
    Qed.

    Let lemma2 a ll : good R (ll++a::nil) -> bar (good (lift_rel R a)) ll.
    Proof.
      intros H.
      apply good_app_inv in H.
      destruct H as [ H | [ H | (x & b & H1 & H2 & H3) ] ].
      
      constructor 1.
      revert H; apply good_inc.
      intros ? ?; red; tauto.
      
      apply good_sg_inv in H; destruct H.
      
      destruct H2 as [ H2 | [] ]; subst b.
      constructor 2; intros u.
      constructor 1.
      constructor 1 with x; auto.
      red; auto.
    Qed.
  
    Let lemma3 mm : bar (good R) mm -> forall a ll, mm = ll++a::nil -> bar (good (R rlift a)) ll.
    Proof.
      induction 1 as [ mm Hmm | mm Hmm IH ] using bar_ind; intros a ll ?; subst.

      apply lemma2; auto.
      
      constructor 2; intros b.
      apply IH with b; auto.
    Qed.

    Theorem bar_lift_rel a ll : bar (good (R rlift a)) ll <-> bar (good R) (ll++a::nil).
    Proof.
      split; auto.
      intros ?; apply lemma3 with (2 := eq_refl); auto.
    Qed.

  End bar_lift_rel. 

  Fact good_bar_ctxt w1 w2 : good R (w1++w2) -> bar (good (ctxt R w2)) w1.
  Proof.
    intros H.
    apply good_app_inv in H.
    destruct H as [ H | [ H | (a & b & H1 & H2 & H3) ] ].
    
    (* w1 is good *)
    
    constructor 1.
    revert H; apply good_inc.
    intros ? ? ?; red; tauto.
    
    (* w2 is good *)
    constructor 2; intros u.
    constructor 2; intros v.
    constructor 1. 
    constructor 1 with u; left; auto.
    
    (* a in w1, b in w2 and b << a *)
    constructor 2; intros u.
    constructor 1.
    constructor 1 with a; auto.
    red.
    right; left.
    exists b; auto.
  Qed.

  Fact bar_ctxt w1 w2 : bar (good R) (w1++w2) <-> bar (good (ctxt R w2)) w1.
  Proof.
    split.

    set (ll := w1++w2).
    intros H.
    generalize (eq_refl ll).
    unfold ll at 2.
    revert H.
    generalize ll; clear ll; intros ll.
    intros H; revert H w1 w2.
    induction 1 as [ ll Hll | ll Hll IHll ] using bar_ind; intros w1 w2 H.  
    subst.
    apply good_bar_ctxt; auto.
    constructor 2; intros a.
    apply IHll with a; subst; auto.
  
    induction 1 as [ w1 Hw1 | w1 Hw1 IH ] using bar_ind.
    
    induction Hw1 as [ w1 a b H1 H2 | ]; simpl.
    destruct H2 as [ H2 | [ H2 | H2 ] ].
    constructor 1; constructor 2; apply good_app_left; auto.
    constructor 1; constructor 2; apply exists_good_app.
    destruct H2 as (c & H2 & H3); exists b, c; auto.
    constructor 1; constructor 1 with b; auto; apply in_or_app; tauto.
    apply bar_good_app with (ll := a::nil); auto.
  
    constructor 2; auto.
  Qed.

End Bar_Good.

Section another_characterization.

  (* this one is from M. Seisenberger *)

  Variable (X : Type) (R : X -> X -> Prop).

  (* which works for DECIDABLE R *)
  
  Hypothesis R_dec : forall x y, R x y \/ ~ R x y.

  Let bad_ext w := fun b => forall a, In a w -> ~ R a b.

  Let R_ext w1 w2 := exists b, bad_ext w2 b /\ w1 = b::w2.

  Theorem Acc_bar : Acc (R_ext restr (bad R)) nil -> bar (good R) nil.
  Proof.
    generalize (@nil X) (in_bad_0 R).
    apply Acc_r_rect.
    intros w' H1 IH H2.
    constructor 2; intros b.
    destruct (list_dec_ind (fun a => R a b) w') as [ (a & H3 & H4) | H ]. (* R must be decidable here *)
    intros; auto.
    constructor 1; constructor 1 with a; auto.
    apply IH.
    constructor 2; auto.
    exists b; auto.
  Qed.
  
  Theorem bar_Acc : bar (good R) nil -> Acc (R_ext restr (bad R)) nil.
  Proof.
    generalize (@nil X).
    induction 1 as [ w Hw | w Hw IH ].
    constructor.
    intros ? (_ & _ & ?); exfalso. 
    apply good_bad_False with (1 := Hw); auto.
    constructor.
    intros w' ((b & H1 & H2) & H3 & H4); subst; apply IH.
  Qed.

End another_characterization.