(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Min.
Require Import List.
Require Import Relations.
Require Import Wf.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import list_sig.
Require Import list_fan.

Require Import finite.

Require Import good_base.
Require Import af_t.
Require Import bar_t.
Require Import good_bar_t.
Require Import af_bar_t.
Require Import fan_t.
Require Import afs_t.

Require Import af.
Require Import bar.
Require Import af_bar.
Require Import good_bar.
Require Import afs.
Require Import fan.

Require Import subword.

Set Implicit Arguments.

Section higman_ev.

  Variable (X : Type).

  Let hig_ev (c : list X * list X) := snd c ++ fst c.

  Notation analysis := (fun l m => hig_ev m = l).  

  Let analysis_nil : analysis nil ~eq1 fun a => a = (nil,nil).
  Proof.
    split.
    intros ([ | ],[ | ]); 
      unfold hig_ev; simpl; auto; discriminate 1.
    intros ? ?; subst; auto.
  Qed.
  
  Let analysis_cons x l : analysis (x::l) ~eq1 fun a => (exists a', analysis l a' /\ (fst a = fst a' 
                                                                                  /\ snd a = x::snd a'))
                                                      \/ (fst a = x::l /\ snd a = nil).
  Proof.
    split; unfold hig_ev.

    intros (y,[ | x' m ]); simpl.    
    destruct y as [ | x' l' ]; simpl.
    discriminate 1.
    intros H; injection H; clear H; intros ? ?; subst x' l'; right; auto.
    intros H; injection H; clear H; intros ? ?; subst x' l; left; exists (y,m); auto.
    
    intros (y,m) [ (m' & H1 & H2 & H3) | (H1 & H2) ]; simpl in *; subst; simpl; auto.
  Qed.
  
  Let analysis_finite l : finite_t (analysis l).
  Proof.
    induction l as [ | x l IHl ].
   
    apply finite_t_eq with (1 := eq1_sym _ _ analysis_nil).
    exists ((nil,nil)::nil). 
    intros ( ? & ? ); simpl.
    split; auto; intros [ | [] ]; auto.
    
    apply finite_t_eq with (1 := eq1_sym _ _ (analysis_cons x l)).    
    apply finite_t_cup.
    apply finite_t_comp with (R := analysis); auto.
    intros (y,m) _; simpl.
    exists ((y,x::m)::nil).
    intros (y',m'); simpl; split.
    intros [ H | [] ]; injection H; intros; subst; auto.
    intros (? & ?); subst; auto.
    
    exists ((x::l,nil)::nil).
    intros (y',m'); simpl; split.
    intros [ H | [] ]; injection H; intros; subst; auto.
    intros (? & ?); subst; auto.
  Qed.

  Let hig_ev_inv l := proj1_sig (analysis_finite l).
  Let hig_ev_inv_spec l a : In a (hig_ev_inv l) <-> analysis l a.
  Proof. apply (proj2_sig (analysis_finite l)). Qed.
  
  Variables (a : X) (ll : list X) (R : X -> X -> Prop).
  
  Let hig_rel0 (y1 y2 : list X) :=
    match y1, y2 with
      | nil    , nil     => True
      | x1::m1 , x2::m2  => R x1 x2 /\ (subword R rlift ll) m1 m2
      | _          , _           => False
    end. 

  Let hig_rel0_af_t : af_t R -> af_t (subword R rlift ll) -> af_t hig_rel0.
  Proof.
    intros H1 H2.
    generalize (af_t_sum_both (af_t_True unit) (af_t_prod H1 H2)).
    apply af_t_relmap with (f := fun x l => match x with inl _ => nil | inr (y,m) => y::m end = l).
    intros [ | x m ].
    exists (inl tt); auto.
    exists (inr (x,m)); auto.
    intros [ [] | (?,?) ] [ [] | (?,?) ] ? ? ? ?; subst; cbv; auto.
  Qed.

  Let hig_rel0_af : af R -> af (subword R rlift ll) -> af hig_rel0.
  Proof.
    intros H1 H2. 
    generalize (af_sum_both (af_True unit) (af_product H1 H2)).
    apply af_relmap with (f := fun x l => match x with inl _ => nil | inr (y,m) => y::m end = l).
    intros [ | x m ].
    exists (inl tt); auto.
    exists (inr (x,m)); auto.
    intros [ [] | (?,?) ] [ [] | (?,?) ] ? ? ? ?; subst; cbv; auto.
  Qed.
    
  Definition hig_rel := rel_prod hig_rel0 (subword (R rlift a)).
  
  Fact hig_rel_af_t : af_t R 
                   -> af_t (subword R rlift ll)
                   -> af_t (subword (R rlift a))
                   -> af_t hig_rel.
  Proof.
    intros; apply af_t_prod; auto.
  Qed.
  
  Fact hig_rel_af :   af R 
                   -> af (subword R rlift ll)
                   -> af (subword (R rlift a))
                   -> af hig_rel.
  Proof.
    intros; apply af_product; auto.
  Qed.
  
  Let hig_ev_disapointing (c : list X*list X) :=
    match c with
      | (_,x::_) => R a x
      | (y::m,_) => subword R ll m
      | _        => False
    end.

  Let hig_ev_has_disap u := exists v w, hig_ev_disapointing (fst u,w) /\ u = (fst u,v++w).
 
  Let hig_ev_disap_prop m : analysis m inc1 hig_ev_has_disap -> subword R (a::ll) m.
  Proof.
    induction m as [ | x m IH ]; intros H.
    
    specialize (H (nil,nil)). 
    simpl in H; spec_all H; auto.
    destruct H as (v & w & H1 & H2); simpl in H1, H2.
    destruct v; try discriminate H2.
    destruct w; try discriminate H2.
    destruct H1.

    set (U := hig_ev_has_disap).
    set (V u := hig_ev_disapointing (fst u,x::snd u)).
    
    generalize (@finite_decide _ (analysis m) U V); intros H1; spec_all H1.
    apply finite_t_finite, analysis_finite.
    intros (y,k) H2.
    specialize (H (y,x::k)); spec_all H. 
    unfold hig_ev; simpl.
    subst m; auto.
    destruct H as ( [ | z v ] & w & H3 & H4 ).
    simpl in H4; injection H4; clear H4; intro; subst.
    right; unfold V; auto.
    simpl in H4; injection H4; clear H4; intros ? ?.
    subst z k.
    left; exists v, w; simpl; auto.
    
    destruct H1 as [ H1 | H1 ].
    constructor 2; auto.
    
    unfold V in H1.
    destruct H1 as ((y,k) & H1 & H2 ).
    simpl in H2.
    assert (R a x) as H3.
      destruct y; auto.
    clear H2.
    
    specialize (H (x::m,nil)); spec_all H.
    apply analysis_cons; right; auto.
    destruct H as (v & w & H4 & H5).
    simpl in H5.
    destruct v; try discriminate H5.
    destruct w; try discriminate H5.
    simpl in H4.
    constructor 3; auto.
  Qed.
  
  Let hig_ev_subword x y : hig_rel x y 
                        -> subword R (hig_ev x) (hig_ev y)  
                        \/ hig_ev_has_disap x.
  Proof.
    destruct x as (y1,m1).
    destruct y as (y2,m2).
    unfold hig_rel.
    intros (H1 & H2); simpl in H1, H2.
    induction H2 as [ | m1 m2 x _ H3 | m1 m2 x y H2 _ H3 ].

    destruct y1 as [ | x1 l1 ];
    destruct y2 as [ | x2 l2 ]; try (destruct H1; fail); simpl in H1.
    left; constructor 1.
    destruct H1 as (H1 & [ H2 | H2 ]).
    left; constructor 3; auto.
    right; exists nil, nil; simpl; auto.
    
    destruct H3 as [ H3 | H3 ].
    left; constructor 2; auto.
    right; auto.
    
    destruct H3 as [ H3 | (v & w & H3 & H4) ].
    destruct H2 as [ H2 | H2 ].
    left; constructor 3; auto.
    right; exists nil, (x::m1); split; auto.
    simpl; destruct y1; auto.
    right; exists (x::v), w.
    split; auto.
    simpl in H4 |- *.
    injection H4; intro; subst; auto.
  Qed.
  
  Let hig_ev_good_or_disap l : 
                   good hig_rel l
                -> good (subword R) (map hig_ev l)
                \/ exists x, In x l /\ hig_ev_has_disap x.
  Proof.
    induction 1 as [ l x1 x2 H1 H2 
                   | l x1 x2 [ H1 | (x & H1 & H2) ] ].
                   
    apply hig_ev_subword in H2.
    destruct H2 as [ H2 | H2 ].
    left; simpl; constructor 1 with (hig_ev x2); auto.
    rewrite in_map_iff; exists x2; auto.
    right; exists x2; simpl; tauto.
    
    left; simpl; constructor 2; auto.
    
    right; exists x; simpl; tauto.
  Qed.
    
  Hypothesis HST : af hig_rel.
  Hypothesis HST_t : af_t hig_rel.

  Let HST' : bar (good hig_rel) nil.
  Proof.
    revert HST; apply af_eq_bar_good.
  Qed.
  
  Let HST_t' : bar_t (good hig_rel) nil.
  Proof.
    revert HST_t; apply af_t_bar_t.
  Qed.

  Notation U := (fun ws => Forall (good hig_rel) (list_fan (map hig_ev_inv ws))).

  Let HU : bar U nil.
  Proof.
    apply bar_map_inv with (f := hig_ev_inv) 
                             (Q := fun ll => Forall (good hig_rel) (list_fan ll)); auto.
    simpl; apply fan_on_list; auto.
    constructor; auto.
  Qed.
  
  Let HU_t : bar_t U nil.
  Proof.
    apply bar_t_map_inv with (f := hig_ev_inv) 
                             (Q := fun ll => Forall (good hig_rel) (list_fan ll)); auto.
    simpl; apply fan_t_on_list; auto.
    constructor; auto.
  Qed.
  
  Let U_to_good ws : U ws -> good (subword R) (ws ++ (a::ll) :: nil).
  Proof.
    set (Q l := good (subword R) (map hig_ev l)).
    set (B a := hig_ev_has_disap a).
    
    intros H0.
    destruct (@hig_combi_principle _ (map hig_ev_inv ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
    
    (* condition of hig_combi_principle *)
    
    intros l Hl.
    apply list_fan_spec in Hl.
    rewrite Forall_forall in H0.
    apply H0 in Hl; revert Hl.
    apply hig_ev_good_or_disap; auto.
    
    (* there is a good sequence p that evaluates to ws *)
      
    unfold Q in Q2.
    apply good_app_right.
    replace_with Q2.
    symmetry.
    apply Forall2_eq.
    revert Q1.
    rewrite Forall2_map_right, Forall2_map_left.
    apply Forall2_impl.
    intros ? ?; apply hig_ev_inv_spec.
    
    (* there is a fan in which every element contains a dispointing subtree *)
    
    unfold B in Q2.
    rewrite in_map_iff in Q1.
    destruct Q1 as (x & ? & Q1); subst p.
    apply in_split in Q1.
    destruct Q1 as (l1 & l2 & Q1).
    subst ws.
    rewrite <- app_assoc; simpl.
    apply good_app_left.
    constructor 1 with (a::ll).
    apply in_or_app; right; left; auto.
    apply hig_ev_disap_prop.
    intros ?; rewrite <- hig_ev_inv_spec; auto.
  Qed.
  
  Let bar_U_bar_good ws : bar U ws -> bar (good (subword R rlift (a::ll))) ws.
  Proof.
    induction 1 using bar_ind.
    apply bar_lift_rel, in_bar_0; auto.
    apply in_bar_1; auto.
  Qed.  
  
  Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (subword R rlift (a::ll))) ws.
  Proof.
    induction 1 using bar_t_rect.
    apply bar_t_lift_rel_rev, in_bar_t0; auto.
    apply in_bar_t1; auto.
  Qed.
  
  Theorem hig_ev_af : af (subword R rlift (a::ll)).
  Proof.
    apply af_eq_bar_good; auto.
  Qed. 
   
  Theorem hig_ev_af_t : af_t (subword R rlift (a::ll)).
  Proof.
    apply bar_t_af_t; auto.
  Qed. 

End higman_ev.

Section higman_lemma.

  Variable X : Type.

  Implicit Type R : X -> X -> Prop.

  Let subword_full_length R : full R -> forall l m, length l <= length m -> subword R l m.
  Proof.
    intros HR.
    induction l as [ | x l IH ]; intros [ | y m ].
    constructor.
    intros; apply subword_nil_true.
    simpl; omega.
    simpl; intros H; apply le_S_n in H.
    constructor 3; auto.
  Qed.

  Lemma higman_list_af R : @af X R -> af (subword R).
  Proof.
    induction 1 as [ R HR | R HR IHR ].

    generalize leq_af.
    apply af_relmap with (f := fun n l => n = length l).
    intros l; exists (length l); auto.
    intros; subst; auto.
    
    apply in_af_1.
    intros ll; induction ll as  [ | a ll IHll ].
    
    apply in_af_0.
    intros x ?; right.
    apply subword_nil_true.
    
    apply hig_ev_af.
    apply hig_rel_af; auto.
    apply in_af_1; auto.
  Qed.

  Lemma higman_list_af_t R : @af_t X R -> af_t (subword R).
  Proof.
    induction 1 as [ R HR | R HR IHR ] using af_t_rect.

    generalize le_af_t.
    apply af_t_relmap with (f := fun n l => n = length l).
    intros l; exists (length l); auto.
    intros; subst; auto.
    
    apply in_af_t1.
    intros ll; induction ll as  [ | a ll IHll ].
    
    apply in_af_t0.
    intros x ?; right.
    apply subword_nil_true.
    
    apply hig_ev_af_t.
    apply hig_rel_af_t; auto.
    apply in_af_t1; auto.
  Qed.

End higman_lemma.

Section hig_lemma_afs_t.

  Variables (X : Type) 
            (P : X -> Prop) 
            (Pdec : forall x, { P x } + { ~ P x })
            (R : X -> X -> Prop).

  Fact higman_list_afs (HR : afs P R) : afs (Forall P) (subword R).
  Proof. 
    apply afs_af_rel_restr_eq in HR.
    apply higman_list_af in HR.
    apply afs_af_rel_restr_eq.
    
    revert HR.
    apply af_relmap with (f := fun x y => map (@proj1_sig _ _) x = proj1_sig y).
    intros (ll & Hll); simpl.
    rewrite Forall_forall in Hll.
    exists (list_Forall_sig _ _ Hll).
    apply list_sig_Forall_eq.
    intros l1 l2 (l3 & Hl3) (l4 & Hl4); simpl.
    intros; subst.
    revert H1.
    rewrite subword_Restr.      
    unfold rel_restr; simpl; auto.
  Qed.            
                         
  Fact higman_list_afs_t (HR : afs_t P R) : afs_t (Forall P) (subword R).
  Proof.
    apply afs_t_impl_af_t in HR.
    apply higman_list_af_t in HR.
    apply af_t_impl_afs_t.
    intros; apply Forall_dec; intros; apply Pdec.
    
    revert HR.
    apply af_t_relmap with (f := fun x y => map (@proj1_sig _ _) x = proj1_sig y).
    intros (ll & Hll); simpl.
    rewrite Forall_forall in Hll.
    exists (list_Forall_sig _ _ Hll).
    apply list_sig_Forall_eq.
    intros l1 l2 (l3 & Hl3) (l4 & Hl4); simpl.
    intros; subst.
    revert H1.
    rewrite subword_Restr.      
    unfold rel_restr; simpl; auto.
  Qed.

End hig_lemma_afs_t.
