(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.

Require Import tacs.

Require Import rel_utils.

Require Import af.
Require Import afs.

Require Import af_t.
Require Import afs_t.

Require Import tree.
Require Import tree_embeddings.

Require Import higman_ktree_af.
Require Import higman_ktree_af_t.

Set Implicit Arguments.

Section higman_tree_with_sig.

  Variable (X : Type) (ar : X -> nat) (k : nat) (R : X -> X -> Prop).
  
  Variable Har : forall x, ar x < k.

  Section afs_t.
  
    Hypothesis HR : forall n, n < k -> afs_t (fun x => ar x = n) R.
    
    Theorem higman_stree_afs_t : afs_t (stree ar) (emb_tree_product R).
    Proof.
      set (P' n (x : X) := n < k /\ ar x = n).
      set (R' (_ : nat) := R).
      
      generalize (@higman_wfptree _ k P' R').
      intros Hig.
      spec_all Hig.
      
      intros n x; unfold P'.
      destruct (lt_dec n k); 
      destruct (eq_nat_dec (ar x) n); 
      tauto.
      
      intros ? ? ? (? & _); omega.
      intros n Hn; generalize (HR Hn).
      apply afs_t_incl; auto.
      intros ? (_ & ?); auto.
      
      generalize Hig.
      apply afs_t_incl.
      apply tree_fall_inc.
      intros x ll H; split; auto.
      rewrite <- H; apply Har.
      unfold R'.
      apply emb_tree_product_higman_eq.
    Qed.
    
  End afs_t.

  Section af_t.

    Hypothesis HR : forall n, n < k -> af_t (R <# fun x => ar x = n #>).
  
    Theorem higman_stree_af_t : af_t (emb_tree_product R <# stree ar #>).
    Proof.
      apply afs_t_impl_af_t.
      apply higman_stree_afs_t.
      intros n Hn.
      apply af_t_impl_afs_t; auto.
      intro; apply eq_nat_dec.
    Defined.

  End af_t.

End higman_tree_with_sig.
  
Section higman_btree_af_t.

  Variable (X : Type) (k : nat) (R : X -> X -> Prop) (HR : af_t R).
  
  Theorem higman_btree_afs_t : afs_t (btree k) (emb_tree_product R).
  Proof.
    set (P' n (_ : X) := n < k).
    set (R' (_ : nat) := R).
    
    generalize (@higman_wfptree _ k P' R').
    intros Hig.
    spec Hig H1.
    intros; apply lt_dec.
    spec Hig H2.
    unfold P'; intros; omega.
    spec Hig H3.
    intros; apply af_t_any_afs_t; auto.
    
    revert Hig.
    apply afs_t_inc.
    apply btree_dec.
    unfold R'; intros; subst; auto.
    apply emb_tree_product_higman_eq; auto.
  Qed.
  
  Theorem higman_btree_af_t : af_t (emb_tree_product R <# btree k #>).
  Proof.
    apply afs_t_impl_af_t, higman_btree_afs_t.
  Qed.

End higman_btree_af_t.

Section higman_btree_af.

  Variable (X : Type) (k : nat) (R : X -> X -> Prop) (HR : af R).
  
  Theorem higman_btree_afs : afs (btree k) (emb_tree_product R).
  Proof.
    set (P' n (_ : X) := n < k).
    set (R' (_ : nat) := R).
    
    generalize (@higman_wfptree_afs _ k P' R').
    intros Hig.
    spec_all Hig.
    intros; apply lt_dec.
    unfold P'; intros; omega.
    intros; apply af_any_afs; auto.
    
    revert Hig.
    apply afs_inc.
    unfold R'; intros; subst; auto.
    apply emb_tree_product_higman_eq; auto.
  Qed.
  
  Theorem higman_btree_af : af (emb_tree_product R <# btree k #>).
  Proof.
    apply afs_af_rel_restr_eq, higman_btree_afs.
  Qed.

End higman_btree_af.

Check higman_btree_af.
Print Assumptions higman_btree_af.


Check higman_btree_af_t.
Print Assumptions higman_btree_af_t.
