(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Min.
Require Import List.
Require Import Relations.
Require Import Wf.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import subword.

Require Import af.
Require Import afs.
Require Import afs_lex.

Require Import tree.
Require Import tree_embeddings.

Require Import tree_kruskal_fix.
Require Import tree_kruskal_ev1.
Require Import tree_kruskal_ev2.
Require Import tree_kruskal_ev3.
Require Import tree_kruskal_ev4.

Require Import higman_ktree_af.

Set Implicit Arguments.

Section kruskal_ktree_rec.

  Variable X : Type.
  Notation T := (ktree_fix X).
  
  Let Q k (P : nat -> T -> Prop) R :=  (forall n, ~ P n in_ktf_u)
                  -> (forall n x, { P n x } + { ~ P n x }) 
                  -> (forall n, k <= n -> P k = P n)
                  ->  afs (wfptree P) (emb_tree_upto k R).

  Theorem kruskal_ktree_af_rec : forall k P R, (forall n, afs (P n) (R n)) -> Q k P R.
  Proof.
  
(* the outermost induction is based on lexicographic
   product for the notion of well-foundedness upto a projection 

*)
    apply afs_facile_ind; unfold Q; clear Q.

    (* first we need to show that Q is extensional *)
    
    intros k P1 P2 R1 R2 HP HR H1 H2 H3 H4.
    spec_all H1; auto.
    intro; rewrite HP; auto.
    intro; rewrite HP; auto.
    intros; rewrite HP; auto.
    rewrite HP; auto.
    revert H1.
    apply afs_mono.
    apply tree_fall_inc.
    intros ? ?; rewrite HP; auto.
    apply emb_tree_upto_inc.
    intro; rewrite HR; auto.
    
    (* then we can proof the result by well-founded upto induction *)

    intros k w P R H_PR IHs HP1 HP2 HP3.
    
    assert (forall i, afs (P i) (R i)) as HPR.
      intros i.
      generalize (H_PR i).
      apply afsr_correct_afs.   
    assert (forall t, { wfptree P t } + { ~ wfptree P t }) as HP4.
      apply wfptree_dec, HP2.
  
(* now starts the nested induction over the well-formed tree *)

    apply in_afs_1; auto.
    induction 1 as [ m lt H_m_lt IHt ] using wfptree_ind.
    
    case_eq (length lt).

(* case where lt is empty, ie we have a leaf *)
    
    intros Hl_lt.
    destruct lt.
    2: discriminate Hl_lt.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in H_m_lt.
    destruct H_m_lt as [ Hm1 _ ]; simpl in Hm1.
    
    generalize (H_PR 0); intros HPR0.
    case_eq (w 0); intros Hw0; rewrite Hw0 in HPR0; simpl in HPR0.

  (* if P 0 is empty *)
  
    contradict Hm1; auto.
    
  (* if R 0 is full on P 0 *)
   
    apply in_afs_0.
    intros x y Hx Hy; right.
    clear y Hy.
    destruct (tree_leaf x) as (l & Hl).
    unfold wfptree in Hx.
    rewrite tree_fall_sub_tree in Hx.    
    generalize Hl.     
    apply emb_sub_tree_upto.
    apply in_emb_tree_upto_nil; auto.
    apply HPR0; auto.
    apply Hx with (1 := Hl).
    
  (* if R rlift x is secured by f x *)
  
    specialize (@IHs k w P (fun i => match i with 0 => R 0 rlift m | S i => R (S i) end)).
    spec_all IHs; auto.
    
    intros [ | i ]; auto.
    rewrite Hw0; simpl; apply afs_inv; auto.
    
    destruct k as [ | k ].
    left; right; repeat split; auto; exists m; auto.
    right; split; auto; exists 0; split.
    omega.
    split; auto.
    right; repeat split; auto; exists m; auto.
    intros [ | j ] (H1 & H2); try omega; auto.
    
    (* now we can use the induction hypothesis *)
  
    revert IHs; apply afs_inc; auto.
    intros t1 t2.
    induction 1 as [ t1 t2 x ll Hll H IH 
                   | x y ll mm Hll Ha Hb IH
                   | x y ll mm Hll Ha Hb IH 
                   ] using emb_tree_upto_ind.

    destruct IH as [ IH | IH ].
    left; apply in_emb_tree_upto_0 with t2; auto.
    right; auto.
    
    apply Forall2_decide_ind in IH.
    destruct IH as [ IH | IH ].
    
    destruct ll; simpl in Ha.
    destruct Ha as [ Ha | Ha ].
    inversion_clear IH.
    left; apply in_emb_tree_upto_nil; auto.
    right; apply in_emb_tree_upto_nil; auto.
    left; apply in_emb_tree_upto_1; auto.
    
    rewrite Exists2_exists_left in IH.
    destruct IH as (IH & t & G4 & G5).
    right; apply in_emb_tree_upto_0 with t; auto.
    
    apply subword_cup in IH.
    destruct IH as [ IH | IH ].
    destruct k.
    destruct Ha as [ Ha | Ha ].
    left; apply in_emb_tree_upto_2; auto.
    right; apply in_emb_tree_upto_2; auto; apply subword_nil_true.
    left; apply in_emb_tree_upto_2; auto.
    destruct IH as (u & _ & G1 & _ & G2).
    right; apply in_emb_tree_upto_0 with u; auto.
       
(* case where lt is NOT empty, ie NOT a leaf *)
    
    intros i Hi.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in Hm0.
    destruct Hm0 as [ Hm1 Hm2 ].
    rewrite Hi in Hm1.    

    case (lt_eq_lt_dec k (S i)); [ intros [ H_k_i | H_k_i ] | intros H_k_i ].
        
(* case where S i = length lt and k <= i 
   in that case nts_seq s (S i) must be None 
*)
    
    apply le_S_n in H_k_i.
    generalize (H_PR k); intros HRk.
    
    case_eq (w k); intros Hwk; rewrite Hwk in HRk; simpl in HRk.

  (* wk = af_empty contradicts P (S i) m *)
  
    contradict Hm1; rewrite <- HP3; try omega; auto.
    
  (* wk = af_full, then R k is full on P k *)
    
    set (P' := @ktf_prop2 X k P i).
    assert (G0 : afs (P' i) (ktf_rel2_i k i lt R)).
      apply afs_ktf_rel2_i; auto.     
    set (R' := @ktf_rel2 _ k i lt R).
  
    generalize (@higman_wfptree_afs _ (S i) P' R'); intros Hig; spec_all Hig.
    
    apply ktf_prop2_dec; auto.
    intros n ? Hn; unfold P'; rewrite ktf_prop2_gt; auto.
    intros n Hn.
    unfold P', R'.
    destruct (ktf_pos2 k i n) as [[[(?&?)|]|H]|]; try omega.

    rewrite ktf_prop2_lt; try omega.
    rewrite ktf_rel2_eq_0; auto.
   
    rewrite ktf_prop2_lt; try omega.
    rewrite ktf_rel2_eq_1; auto.
    rewrite <- HP3; try omega; auto.
 
    rewrite H, ktf_rel2_eq_2; auto.

    (* now Hig is proof that emb_ntree R' is af on wfptree P' *)
    
    rewrite afs_af_rel_restr_eq in Hig.
    rewrite afs_af_rel_restr_eq.
    revert Hig; apply ktf_ev2_af with (H_m_lt := H_m_lt); auto.

  (* wk = af_rlift, then R k rlift x is secured by f x *)
    
    set (P' := @ktf_prop1 X k P i).
    assert (G0 : afs (P' i) (ktf_rel1_i k i lt R)).
      apply afs_ktf_rel1_i; auto.
    assert (Hki : k <= S i).
      omega.
    set (R' := @ktf_rel1 _ k i m lt R).
    set (w' j := match ktf_pos1 k i j with
                   | inleft (inleft (inleft (left _)))   => w j
                   | inleft (inleft (inleft (right _)))  => af_rlift
                   | inleft (inleft (inright _))         => af_rlift
                   | inleft (inright _)                  => af_rlift
                   | inright _                           => w j
                 end).
    specialize (IHs (S i) w' P' R').
    spec_all IHs; auto.
    
    unfold w', P', R'.
    intros j.
    destruct (ktf_pos1 k i j) as [[[[|]|]|]|]; try omega; auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_0; try omega. 
    auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_1; try omega.
    rewrite <- Hwk, <- HP3; auto; omega.
    subst j.
    rewrite ktf_rel1_eq_2; try omega.
    apply G0.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; try omega.
    rewrite <- HP3; try omega.
    simpl; apply afs_inv; auto.
    rewrite <- HP3 in Hm1; auto; omega.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_4; try omega.
    auto.
    
    left.
    unfold w', P', R'.
    destruct (ktf_pos1 k i (S i)) as [[[[|]|]|]|]; try omega; auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; try omega.
    rewrite <- HP3; auto.
    right.
    repeat split; auto; exists m; split; auto.
    rewrite <- HP3 in Hm1; auto.
    
    apply ktf_prop1_u; auto.
    apply ktf_prop1_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop1_neq; try omega).
    rewrite <- (HP3 n); try omega.
    rewrite <- HP3; try omega.
    auto.
 
    (* now IHs is proof that emb_tree_upto (S i) R' is af on wfptree P' *)
   
    rewrite afs_af_rel_restr_eq in IHs.
    rewrite afs_af_rel_restr_eq.
    revert IHs; apply ktf_ev1_af with (H_m_lt := H_m_lt); auto.

(* case k = S i *)

    generalize (H_PR k); intros HRk.
    rewrite <- H_k_i in Hm1.

    case_eq (w k); intros Hwk; rewrite Hwk in HRk; simpl in HRk.

  (* wk = af_empty, is not possible because m belongs to P (S i) *)
  
    contradict Hm1; auto.
    
  (* wk = af_full, then R (S i) is full on P (S i) *)

    set (P' := @ktf_prop2 X k P i).
    assert (G0 : afs (P' i) (ktf_rel2_i k i lt R)).
      apply afs_ktf_rel2_i; auto.
      omega.
    set (R' := @ktf_rel2 _ k i lt R).

    (* we can use Higman theorem in that case *)
  
    generalize (@higman_wfptree_afs _ (S i) P' R'); intros Hig; spec_all Hig.

    apply (@ktf_prop2_dec _ k); auto; omega.
    intros n ? Hn; unfold P'; rewrite (ktf_prop2_gt (k := k)); auto; omega.
    intros n Hn.
    unfold P', R'.
    destruct (ktf_pos2 k i n) as [[[(?&?)|]|H]|]; try omega.

    rewrite (ktf_prop2_lt (k := k)); try omega.
    rewrite ktf_rel2_eq_0; auto; try omega.
    subst n; rewrite ktf_rel2_eq_2; try omega.
    apply G0.

    (* now Hig is proof that emb_ntree R' is af on wfptree P' *)
    
    rewrite afs_af_rel_restr_eq in Hig.
    rewrite afs_af_rel_restr_eq.
    revert Hig; apply ktf_ev2_af with (H_m_lt := H_m_lt); auto. 
    omega.
    
  (* wk = af_rlift, then R (S i) rlift x is secured by f x *)
      
    set (P' := @ktf_prop1 X k P i).
    assert (G0 : afs (P' i) (ktf_rel1_i k i lt R)).
      apply afs_ktf_rel1_i; auto; try omega.
    assert (Hki : k <= S i).
      omega.
    set (R' := @ktf_rel1 _ k i m lt R). 
    set (w' j := match ktf_pos1 k i j with
                   | inleft (inleft (inleft (left _)))   => w j
                   | inleft (inleft (inleft (right _)))  => af_rlift
                   | inleft (inleft (inright _))         => af_rlift
                   | inleft (inright _)                  => af_rlift
                   | inright _                           => w j
                 end).
    specialize (IHs (S i) w' P' R').
    spec_all IHs; auto.
    
    unfold w', P', R'.
    intros j.
    destruct (ktf_pos1 k i j) as [[[[|]|]|]|]; try omega; auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_0; try omega. 
    auto.
    subst j.
    rewrite ktf_rel1_eq_2; try omega.
    apply G0.
    subst j.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; try omega.
    rewrite <- HP3; try omega.
    simpl; apply afs_inv; auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_4; try omega.
    auto.
    
    left.
    unfold w', P', R'.
    destruct (ktf_pos1 k i (S i)) as [[[[|]|]|]|]; try omega; auto.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; try omega.
    rewrite <- HP3; auto.
    right.
    repeat split; auto; exists m; split; auto.
    
    apply ktf_prop1_u; auto.
    apply ktf_prop1_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop1_neq; try omega).
    rewrite <- (HP3 n); try omega.
    rewrite <- HP3; try omega.
    auto.
 
    (* now IHs is proof that emb_tree_upto (S i) R' is af on wfptree P' *)
   
    rewrite afs_af_rel_restr_eq in IHs.
    rewrite afs_af_rel_restr_eq.
    revert IHs; apply ktf_ev1_af with (H_m_lt := H_m_lt); auto.

(* case S i < k *)

    generalize (H_PR (S i)); intros HRSi.
    
    case_eq (w (S i)); intros Hwk; rewrite Hwk in HRSi; simpl in HRSi.

  (* wSi = af_empty, is not possible because m belongs to P (S i) *)
  
    contradict Hm1; auto.

  (* wSi = af_full, then R (S i) is full on P (S i) *)
    
    set (P' := @ktf_prop4 X _ P _ H_k_i).
    assert (G0 : afs (P' i) (ktf_rel4_i k i lt R)).
      apply afs_ktf_rel4_i; auto.
    set (R' := @ktf_rel4 _ _ _ H_k_i lt R).
    set (w' j := match (ktf_pos4 H_k_i j) with
                   | inleft (inleft (left _))   => w j
                   | inleft (inleft (right _))  => af_rlift
                   | inleft (inright _)         => af_empty 
                   | inright _                  => w j
                 end).
    specialize (IHs k w' P' R').
    spec_all IHs; auto.

    unfold w', P', R'.
    intros j.
    destruct (ktf_pos4 H_k_i j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_0; try omega. 
    auto.
    subst j.
    rewrite ktf_rel4_eq_1; try omega.
    apply G0.
    subst j.
    rewrite ktf_prop4_eqS; try omega.
    simpl; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_3; try omega.
    auto.
    
    right; split; auto.
    exists (S i); split; auto.
    unfold w', P', R'; split.
    destruct (ktf_pos4 H_k_i (S i)) as [[[|]|]|]; try omega; auto.
    rewrite Hwk.
    left; red; auto.
    intros j Hj.
    destruct (ktf_pos4 H_k_i j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_3; try omega.
    auto.
    
    apply ktf_prop4_u; auto.
    apply ktf_prop4_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop4_neq; try omega).
    rewrite <- (HP3 n); auto.
    
    (* now IHs is proof that emb_tree_upto k R' is af on wfptree P' *)
   
    rewrite afs_af_rel_restr_eq in IHs.
    rewrite afs_af_rel_restr_eq.
    revert IHs; apply ktf_ev4_af with (H_m_lt := H_m_lt); auto.

  (* wSi = af_rlift, then R (S i) rlift x is secured by f x *)
    
    set (P' := @ktf_prop3 X P i).
    assert (G0 : afs (P' i) (ktf_rel3_i k i lt R)).
      apply afs_ktf_rel3_i; auto. 
    set (R' := @ktf_rel3 _ _ _ H_k_i m lt R).
    set (w' j := match (ktf_pos3 H_k_i j) with
                   | inleft (inleft (left _))   => w j
                   | inleft (inleft (right _))  => af_rlift
                   | inleft (inright _)         => af_rlift
                   | inright _                  => w j
                 end).
    specialize (IHs k w' P' R').
    spec_all IHs; auto.

    unfold w', P', R'.
    intros j.
    destruct (ktf_pos3 H_k_i j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_0; try omega. 
    auto.
    subst j.
    rewrite ktf_rel3_eq_1; try omega.
    apply G0.
    subst j.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_2; try omega.
    simpl; apply afs_inv; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_3; try omega.
    auto.
    
    right; split; auto.
    exists (S i); split; auto.
    unfold w', P', R'; split.
    destruct (ktf_pos3 H_k_i (S i)) as [[[|]|]|]; try omega; auto.
    rewrite Hwk.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_2; try omega.
    right; repeat split; auto; exists m; auto.
    intros j Hj.
    destruct (ktf_pos3 H_k_i j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_3; try omega.
    auto.
  
    apply ktf_prop3_u; auto.
    apply ktf_prop3_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop3_neq; try omega).
    rewrite <- (HP3 n); auto.
                 
    (* now IHs is proof that emb_tree_upto k R' is af on wfptree P' *)
   
    rewrite afs_af_rel_restr_eq in IHs.
    rewrite afs_af_rel_restr_eq.
    revert IHs; apply ktf_ev3_af with (H_m_lt := H_m_lt); auto.
  Qed.
  
End kruskal_ktree_rec.

Section kruskal_wfptree.
  
  Variable (X : Type) (k : nat) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop).
  Hypothesis HP1 : forall n x, { P n x } + { ~ P n x }. 
  Hypothesis HP2 : forall n, k <= n -> P k = P n.
  Hypothesis HR  : forall n, n <= k -> afs (P n) (R n).
  
  Theorem kruskal_wfptree_afs : afs (wfptree P) (emb_tree_upto k R).
  Proof.

    set (P' n (t : ktree_fix X) := 
      match t with
        | in_ktf_0 x => P n x 
        | _          => False
      end).

    assert (forall n x, { P' n x } + { ~ P' n x }) as Pdec'.
      intros n []; unfold P'; try tauto; auto.

    set (R' n (t1 t2 : ktree_fix X) :=
      match t1, t2 with
        | in_ktf_0 u, in_ktf_0 v => 
        match le_lt_dec n k with
          | left  _  => R n u v
          | right _  => R k u v
        end
        | _         , _          => False
      end). 

    assert (HR' : forall n, afs (P' n) (R' n)).
      intros n; unfold P', R'.
      destruct (le_lt_dec n k) as [ Hn | Hn ].
      
      generalize (HR Hn).
      apply afs_relmap with (f := fun x y => y = in_ktf_0 x); auto.
      unfold P'; intros [ | x | | ] H; try (destruct H; fail). 
      exists x; auto.
      unfold P'; intros x1 x2 [ ] [ ] H1 H2 H3 H4 H7 H8; try tauto.
      injection H7; injection H8; intros ? ?; subst; auto.
      
      rewrite <- HP2; try omega.
      specialize (@HR k); spec_all HR; auto.
      revert HR.
      apply afs_relmap with (f := fun x y => y = in_ktf_0 x); auto.
      unfold P'; intros [ | x | | ] H; try (destruct H; fail). 
      exists x; auto.
      unfold P'; intros x1 x2 [ ] [ ] H1 H2 H3 H4 H7 H8; try tauto.
      injection H7; injection H8; intros ? ?; subst; auto.
      
    generalize (@kruskal_ktree_af_rec _ k P' R'); intros Kru.
    spec_all Kru; auto.
    intros n Hn; unfold P'; rewrite <- (HP2 Hn); auto.

    clear HR'.
    revert Kru.
    apply afs_relmap with (f := fun x y => x = tree_map (@in_ktf_0 _) y).

    intros y Hy; exists (tree_map (@in_ktf_0 _) y); split; auto.
    revert Hy.
    induction 1 as [ a ll Hll IH ] using wfptree_ind.
    rewrite tree_map_fix.
    revert Hll; do 2 rewrite wfptree_fix.
    intros (G1 & G2); split.
    rewrite map_length.
    red; auto.
    intros t Ht.
    apply in_map_iff in Ht.
    destruct Ht as (u & ? & Ht); subst t; auto.
    
    intros x1 x2 y1 y2 _ _ _ _ G1 G2 H.
    revert H y1 y2 G1 G2.    
    induction 1 as [ t1 t2 x ll Hll H IH 
                   | x y ll mm Hll Ha Hb IH
                   | x y ll mm Hll Ha Hb IH 
                   ] using emb_tree_upto_ind.
                   
    intros y1 [ b lb ] ? H1; subst t1.
    rewrite tree_map_fix in H1.
    injection H1; clear H1; intros ? ?; subst.
    rewrite in_map_iff in Hll.
    destruct Hll as (u & ? & Hll); subst t2.
    apply in_emb_tree_upto_0 with u; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5.
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in Hll.
    rewrite map_length in Ha.
    red in Ha.
    destruct (le_lt_dec (length kk) k); try omega.
    apply in_emb_tree_upto_1; auto.
    rewrite Forall2_map_both in IH.
    revert IH.
    apply Forall2_impl.
    intros ? ? H; apply H; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5;
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in Hll.
    red in Ha.
    destruct (le_lt_dec k k); try omega.
    apply in_emb_tree_upto_2; auto.
    apply subword_map_both in IH.
    revert IH.
    apply subword_inc.
    intros ? ? H; apply H; auto.
  Qed.

End kruskal_wfptree.
