(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.

Require Omega.

Require Import tacs.

Set Implicit Arguments.

Definition app_split X (l1 : list X) : forall l2 r1 r2, l1++r1 = l2++r2
                                      -> { m | l2 = l1++m /\ r1 = m++r2 }
                                       + { m | l1 = l2++m /\ r2 = m++r1 }.
Proof.
  induction l1 as [ | x l1 Hl1 ].
  left; exists l2; auto.
  intros [ | y l2 ] r1 r2 H.
  right; exists (x::l1); auto.
  simpl in H; injection H; clear H; intros H ?; subst.
  apply Hl1 in H.
  destruct H as [ (m & H1 & H2) | (m & H1 & H2) ].
  left; exists m; subst; auto.
  right; exists m; subst; auto.
Qed.

Fact list_split_first_half U (ll : list U) x : x <= length ll -> { l : _ & { r | ll = l++r /\ length l = x } }.
Proof.
  revert ll; induction x as [ | x IHx ]; intros [ | u ll ] Hx.
  exists nil, nil; simpl; auto.
  exists nil, (u::ll); auto.
  simpl in Hx; omega.
  destruct (IHx ll) as (l & r & H1 & H2).
  simpl in Hx; omega.
  exists (u::l), r; simpl; split; f_equal; auto.
Qed.
    
Fact list_split_second_half U (ll : list U) x : x <= length ll -> { l : _ & { r | ll = l++r /\ length r = x } }.
Proof.
  intros Hx.
  destruct list_split_first_half with (ll := ll) (x := length ll - x)
    as (l & r & H1 & H2).
  omega.
  exists l, r; split; auto.
  apply f_equal with (f := @length _) in H1.
  rewrite app_length in H1.
  omega.
Qed.  

(* prefix of an infinite sequence, in reverse order *)
   
Section prefix.
 
  Variable X : Type.

  Fixpoint pfx_rev f n : list X := 
    match n with 
      | 0   => nil
      | S n => (f n)::(pfx_rev f n)
    end.

  Fact pfx_rev_plus f a b : pfx_rev f (a+b) = pfx_rev (fun n => f (b+n)) a ++ pfx_rev f b.
  Proof.
    induction a; simpl; auto.
    f_equal; auto.
    f_equal; apply plus_comm.
  Qed.
  
  Fact pfx_rev_S f a : pfx_rev f (S a) = pfx_rev (fun n => f (S n)) a ++ f 0 :: nil.
  Proof.
    generalize (pfx_rev_plus f a 1); intros H.
    replace_with H; f_equal; rewrite plus_comm; auto.
  Qed.  

  Fact pfx_rev_In f n x : In x (pfx_rev f n) <-> exists i, i < n /\ x = f i.
  Proof.
    split.

    induction n as [ | n IHn ]; simpl.
    intros [].
    intros [ H | H ].
    subst; exists n; auto.
    destruct (IHn H) as (i & H1 & H2).
    exists i; split; auto.
    
    intros (i & H1 & H2); subst x.
    revert i H1.
    induction n as [ | n IH ]; intros i Hi; try omega; simpl.
    destruct (le_lt_dec n i).
    left; f_equal; omega.
    right; apply IH; auto.
  Qed.

  Fact pfx_length f n : length (pfx_rev f n) = n.
  Proof.
    induction n; simpl; auto.
  Qed.

  Fact pfx_rev_eq f n l x r : pfx_rev f n = l++x::r -> f (length r) = x.
  Proof.
    revert l x r; induction n as [ | n IH ]; intros l x r; simpl.
    destruct l; discriminate 1.
    destruct l as [ | y l ]; simpl;
    intros H; injection H; clear H; intros H1 H2.
    replace_with H2; f_equal.
    subst r; apply pfx_length.
    apply IH with (1 := H1).
  Qed.

End prefix.

Definition list_prefix X (ll : list X) n : n <= length ll -> { l : _ & { r | ll = l++r /\ length l = n } }.
Proof.
  revert n; induction ll as [ | x ll IH ].
  intros [ | n ] Hn; exists nil, nil; split; auto.
  exfalso; simpl in Hn; omega.
  intros [ | n ] Hn.
  exists nil, (x::ll); auto.
  simpl in Hn.
  destruct (IH n) as (l & r & H1 & H2).
  omega.
  subst.
  exists (x::l), r; auto.
Qed.

Definition list_suffix X (ll : list X) n : n <= length ll -> { l : _ & { r | ll = l++r /\ length r = n } }.
Proof.
  intros H.
  destruct (@list_prefix _ ll (length ll - n)) as (l & r & H1 & H2).
  omega.
  exists l, r; repeat split; auto.
  apply f_equal with (f := @length _) in H1.
  rewrite app_length in H1.
  omega.
Qed.

Fact list_prefix_eq X (l r l' r' : list X) : l++r = l'++r' -> length l = length l' -> l = l' /\ r = r'.
Proof.
  revert l'; induction l as [ | x l IHl ]; intros [ | y l' ] H1 H2; auto; simpl in H2; try discriminate H2.
  simpl in H1; injection H1; clear H1; intros H1 H3; subst.
  destruct IHl with (1 := H1).
  omega.
  subst; auto.
Qed.

Fact list_suffix_eq X (l r l' r' : list X) : l++r = l'++r' -> length r = length r' -> l = l' /\ r = r'.
Proof.
  intros H1 H2; apply list_prefix_eq; auto.
  apply f_equal with (f := @length _) in H1.
  do 2 rewrite app_length in H1.
  omega.
Qed.

Fact list_suffix_split X (l r l' r' : list X) : l++r = l'++r' -> length r <= length r' -> { m | l = l'++m /\ r' = m++r }.
Proof.
  revert r l' r'; induction l as [ | x l IHl ]; intros r [ | y l' ] r' H1 H2; simpl in H1.
  exists nil; auto.
  apply f_equal with (f := @length _) in H1.
  simpl in H1.
  rewrite app_length in H1.
  exfalso; omega.
  exists (x::l); auto.
  injection H1; clear H1; intros H1 ?; subst y.
  destruct IHl with (1 := H1) as (m & H3 & H4); auto.
  exists m; subst; auto.
Qed.
