(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import continuity.

Require Import list_utils.
Require Import sublist.

Set Implicit Arguments.

Section subword.

  Variable (X Y : Type) (R : X -> Y -> Prop).

  Infix "<<" := R (at level 70).

  Reserved Notation "x '<*' y" (at level 70).

  Inductive subword : list X -> list Y -> Prop :=
    | in_sw_0 : nil <* nil
    | in_sw_1 : forall ll mm b,             ll <* mm ->    ll <* b::mm
    | in_sw_2 : forall ll mm a b, a << b -> ll <* mm -> a::ll <* b::mm
  where "x <* y" := (subword x y).

  (* subword inversion is the basis of subword decision *)

  Fact subword_inv w1 w2 : w1 <* w2 -> (w1 = nil)
                                    \/ (exists w3, In w3 (tails w2) /\ w1 <* w3)
                                    \/ (exists a b v1 v2, w1 = a::v1 /\ w2 = b::v2 /\ a << b /\ v1 <* v2).
  Proof.
    induction 1 as [ | w1 w2 b H1 [ IH1 | [ (w3 & H2 & H3) | (a & c & v1 & v2 & H2 & H3 & H4 & H5) ] ] 
                     | w1 w2 a b H0 H1 [ IH2 | [ (w3 & H2 & H3) | (c & d & v1 & v2 & H2 & H3 & H4 & H5) ] ] ].
    left; auto.
    left; auto.
    right; left; exists w3; simpl; split; auto.
    subst; right; left; exists (c::v2); split; auto; left; auto.
    subst; right; right; exists a, b, nil, w2; auto.
    right; right; exists a, b, w1, w2; auto.
    subst; right; right; exists a, b, (c::v1), (d::v2); auto.
  Qed.

  Fact subword_cons_inv a w1 w2 : a::w1 <* w2 -> exists l b r, w2 = l++b::r /\ a << b /\ w1 <* r.
  Proof.
    intros H.
    generalize (eq_refl (a::w1)).
    revert H.
    set (w := a::w1).
    unfold w at 3.
    generalize w; clear w.
    intros w H; revert H a w1.
    induction 1 as [ | w1 w2 b H1 IH1 
                     | w1 w2 a b H1 H2 IH2 ]; intros c w H.
    discriminate H.
    subst.
    destruct (IH1 _ _ (eq_refl _)) as (l & d & r & H2 & H3 & H4).
    subst; exists (b::l), d, r; auto.
    injection H; clear H; intros; subst c w.
    exists nil, b, w2; auto.
  Qed.

  Fact subword_length w1 w2 : w1 <* w2 -> length w1 <= length w2.
  Proof.
    induction 1; simpl; omega.
  Qed.

  Fact subword_incl w1 w2 : w1 <* w2 -> forall a, In a w1 -> exists b, In b w2 /\ a << b.
  Proof.
    induction 1 as [ | w1 w2 b H IH | w1 w2 a b H1 H2 IH2 ].

    intros ? [].

    intros a Ha.
    destruct (IH _ Ha) as (c & H1 & H2).
    exists c; split; auto; right; auto.
    
    intros c [ Hc | Hc ].
    subst c; exists b; split; auto; left; auto.
    destruct (IH2 _ Hc) as (d & H3 & H4).
    exists d; split; auto; right; auto.
  Qed.
    
  Fact subword_app w1 w2 l : w1 <* w2 -> w1 <* l++w2.
  Proof. 
    induction l; simpl; intros; auto; constructor; auto. 
  Qed.

  Fact subword_app_inv w1 w2 w : w1++w2 <* w -> exists w3 w4, w = w3++w4 /\ w1 <* w3 /\ w2 <* w4.
  Proof.
    revert w.
    induction w1 as [ | a w1 IH ]; intros w; simpl.
    exists nil, w; repeat split; auto.
    constructor 1.
    intros H.
    apply subword_cons_inv in H.
    destruct H as (l & b & r & H1 & H2 & H3); subst.
    destruct (IH _ H3) as (w3 & w4 & H4 & H5 & H6); subst.
    exists (l++b::w3), w4; repeat split; auto.
    rewrite app_ass; auto.
    apply subword_app.
    constructor 3; auto.
  Qed.
    
  Fact subword_nil_true w : nil <* w.
  Proof.
    rewrite app_nil_end.
    apply subword_app.
    constructor.
  Qed.

  Fact subword_app_comp w1 w2 w3 w4 : w1 <* w2 -> w3 <* w4 -> w1++w3 <* w2++w4.
  Proof.
    intros H1 H2.
    induction H1; auto.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact not_subword_cons a w1 w2 : ~ a::w1 <* w2 -> forall l b r, w2 = l++b::r -> w1 <* r -> ~ a << b.
  Proof.
    intros H l b r H1 H2 H3.
    apply H; subst.
    apply subword_app_comp with (w1 := nil).
    apply subword_nil_true.
    constructor 3; auto.
  Qed.
 
  Fact subword_middle_inv l a r w : l++a::r <* w -> exists l' b r', w = l'++b::r' /\ l <* l' /\ a << b /\ r <* r'.
  Proof.
    intros H.
    apply subword_app_inv in H.
    destruct H as (l' & w1 & H1 & H2 & H3); subst.
    apply subword_cons_inv in H3.
    destruct H3 as (w2 & b & r' & H3 & H4 & H5); subst.
    exists (l'++w2), b, r'; repeat split; auto.
    rewrite app_ass; auto.
    rewrite (app_nil_end l).
    apply subword_app_comp; auto.
    apply subword_nil_true.
  Qed.

  Fact subword_nil_false a w : ~ a::w <* nil.
  Proof. intros H; inversion H. Qed.

  Fact subword_nil_inv a : a <* nil -> a = nil.
  Proof. intros H; destruct a; auto; apply subword_nil_false in H; elim H. Qed.

  Fact subword_tails w w1 w2 : In w1 (tails w2) -> w <* w1 -> w <* w2.
  Proof.
    revert w w1.
    induction w2 as [ | b w2 IH ]; simpl; intros w w1.
    intros [].
    intros [ H | H ]; subst.
    apply in_sw_1.
    intros H0; apply IH in H0; auto.
    apply in_sw_1; auto.
  Qed.

  Fact subword_Forall2 w1 w2 : Forall2 R w1 w2 -> w1 <* w2.
  Proof.
    induction 1.
    constructor 1.
    constructor 3; auto.
  Qed.

  Fact subword_Forall2_length l m : length l = length m 
                                 -> subword l m
                                 -> Forall2 R l m.
  Proof.
    intros H1 H2; revert H2 H1.
    induction 1.
    constructor.
    intros H.
    apply subword_length in H2.
    simpl in H; omega.
    injection 1; constructor; auto.
  Qed.

  Fact Forall2_subword_lsum l m : Forall2 subword l m -> lsum (map (@length _) l) <= lsum (map (@length _) m).
  Proof.
    induction 1 as [ | x y l m H1 H2 H3 ]; simpl; auto.
    apply subword_length in H1; omega.
  Qed.

  Fact lsum_map_length_eq_0 U l : lsum (map (@length U) l) = 0 
                             -> l = map (fun _ => nil) l.
  Proof.
    induction l as [ | [ | x k ] l IH ]; simpl; auto.
    intros; f_equal; auto.
    omega.
  Qed.

  Fact Forall2_subword_Forall2 l m : lsum (map (@length _) l) = 1
                                  -> lsum (map (@length _) m) = 1
                                  -> Forall2 subword l m
                                  -> Forall2 (Forall2 R) l m.
  Proof.
    intros H1 H2 H3; revert H3 H1 H2.
    induction 1 as [ | x y l m H1 H2 H3 ]; simpl; intros H4 H5.
    constructor.
    destruct x as [ | x [ | ] ]; 
    destruct y as [ | y [ | ] ]; simpl in H4, H5; try omega;
    
    simpl in H4, H5.
    constructor; auto.
    apply Forall2_subword_lsum in H2; omega.
    apply subword_length in H1; simpl in H1; omega.
    constructor; auto.
    constructor; auto.
    apply subword_Forall2_length in H1; simpl; auto.
    apply Forall2_cons_inv, proj1 in H1; auto.
    apply subword_cons_inv in H1.
    injection H4; clear H4; intros H4.
    injection H5; clear H5; intros H5.
    apply lsum_map_length_eq_0 in H4.
    apply lsum_map_length_eq_0 in H5.
    rewrite H4, H5.
    clear H1 H3 H4 H5.
    induction H2; simpl; auto.
  Qed.   

  Fact subword_sublist_eq ll nn : ll <* nn <-> exists mm, Forall2 R ll mm /\ mm <sl nn.
  Proof.
    split.
    
    induction 1 as [ | w1 w2 b H (mm & H1 & H2) | w1 w2 a b H1 H2 (mm & H3 & H4) ].
    exists nil; split; constructor.
    exists mm; split; auto; constructor 3; auto.
    exists (b::mm); split; auto; constructor; auto.
    
    intros (mm & H1 & H2).
    revert ll H1.
    induction H2 as [ nn | a mm nn H1 H2 | a mm nn ]; intros ll Hll.
    inversion_clear Hll; apply subword_nil_true.
    apply Forall2_cons_inv_r in Hll.
    destruct Hll as (b & ll' & H3 & H4 & H5); subst.
    constructor 3; auto.
    constructor 2; auto.
  Qed.

  Fact sublist_subword_trans w1 w2 w3 : w1 <sl w2 -> w2 <* w3 -> w1 <* w3.
  Proof.
    intros H1 H2; revert w1 H1.
    induction H2; intros w1 H1.
    apply sublist_nil_inv in H1; subst; constructor.
    constructor 2; auto.
    destruct w1 as [ | x w1 ].
    apply subword_nil_true.
    apply sublist_cons_inv in H1.
    destruct H1 as [ (k & H1 & H3) | (v & k & H1 & H3) ];
    injection H1; clear H1; intros ? ?; subst.

    constructor 3; auto.
    constructor 2; auto.
  Qed.
  
  Fact subword_sublist_trans w1 w2 w3 : w1 <* w2 -> w2 <sl w3 -> w1 <* w3.
  Proof.
    do 2 rewrite subword_sublist_eq.
    intros (l & H1 & H2) H3.
    exists l; split; auto.
    apply sl_trans with (1 := H2); auto.
  Qed.

  Section subword_dec.
    
    Hypothesis Rdec : forall x y, { x << y } + { ~ x << y }.
  
    Hint Resolve subword_nil_true subword_nil_false.

    Let sw_dec_rec n : forall w1 w2, length w2 < n -> (forall x y, In_t x w1 -> In_t y w2 -> { x << y } + { ~ x << y })
                                                   -> ( w1 <* w2 ) + ( ~ w1 <* w2 ).
    Proof.
      induction n as [ | n IHn ].
      intros ? ? ? ?; exfalso; omega.

      intros [ | a w1 ] w2 H0 H2; simpl in H0.
      left; auto.
      
      destruct w2 as [ | b w2]; simpl in H2.
      right; auto.
      
      simpl in H0.
      destruct (list_dec_rect (fun w => a::w1 <* w) (fun w => ~ a::w1 <* w) (tails w2)) as [ (w3 & H3 & H4) | H3 ].
      
      intros ? H; apply IHn.
      apply In_t_In, tails_length in H; omega.
      intros ? ? ? ?; apply H2; auto.
      right; apply tail_In_t with (2 := H); auto.
      
      apply In_t_In in H3.
      left; constructor 2; apply subword_tails with (1 := H3); auto.
      
      destruct (IHn (a::w1) w2) as [ H4 | H4 ]; try omega.
      intros ? ? ? ?; apply H2; auto.
      left; constructor 2; auto.
      
      assert (forall z, In_t z (tails (b::w2)) -> ~ a::w1 <* z) as H1.
        intros z [ H | H ].
        subst; auto.
        apply H3; auto.
      assert (forall z, In z (tails (b::w2)) -> ~ a::w1 <* z) as H1'.
        intro; apply In_t_P; auto.
      clear H3 H4 H1.
     
      destruct (IHn w1 w2) as [ H6 | H6 ]; try omega.
      intros ? ? ? ?; apply H2; auto.
      destruct (H2 a b) as [ H5 | H5 ]; auto.
      left; constructor 3; auto.
      
      right; intros H.
      destruct (subword_inv H) as [ H8 | [ H8 | H8 ] ].
      discriminate H8.
      destruct H8 as (z & [ H8 | H8 ] & H9).
      subst; revert H9; apply H1'; left; auto.
      revert H9; apply H1'; right; auto.
      destruct H8 as (c & d & v1 & v2 & H8 & H9 & H10 & H11).
      injection H8; injection H9; clear H8 H9; intros; subst.
      apply H5; auto.
      
      right; intros H.
      destruct (subword_inv H) as [ H8 | [ H8 | H8 ] ].
      discriminate H8.
      destruct H8 as (z & [ H8 | H8 ] & H9).
      subst; revert H9; apply H1'; left; auto.
      revert H9; apply H1'; right; auto.
      destruct H8 as (c & d & v1 & v2 & H8 & H9 & H10 & H11).
      injection H8; injection H9; clear H8 H9; intros; subst.
      apply H6; auto.
    Qed.
    
    Definition subword_dec w1 w2 : (forall x y, In_t x w1 -> In_t y w2 -> { x << y } + { ~ x << y })
                                                                    ->  { w1 <* w2 } + { ~ w1 <* w2 }.
    Proof.
      intros H; destruct sw_dec_rec with (n := S (length w2)) (2 := H); auto.
    Qed.

  End subword_dec.

End subword.

Section subword_same_type.

  Variables (X : Type) (R : X -> X -> Prop).
  
  Infix "<<" := R (at level 70).
  Infix "<*" := (subword R) (at level 70).

  Section subword_cons_inv_dec.
 
    Hypothesis Rdec : forall x y, { x << y } + { ~ x << y }.
    
    Fact subword_cons_inv_dec a w1 w2 : a::w1 <* w2 -> { l : _ & { b : _  & { r | w2 = l++b::r /\ a << b /\ w1 <* r } } }.
    Proof.
      set (P (l : list X) x r := a << x /\ w1 <* r). 
      destruct (@list_find_split _ P w2) as [ (l & x & r & H1 & H2 & H3) | H1 ]; unfold P in * |- *; clear P.
      
      intros l x r _; destruct (Rdec a x); destruct (subword_dec R w1 r); try (intros; auto; fail); tauto.
      
      exists l, x, r; auto.
      
      intros H; exfalso.
      apply subword_cons_inv in H.
      destruct H as (l & x & r & H2 & H3 & H4).
      apply H1 with (1 := H2); auto.
    Qed.

  End subword_cons_inv_dec.

End subword_same_type.

Section subword_map_both.

  Variables (X Y : Type) (f : X -> Y) (R : Y -> Y -> Prop).

  Let sw_map_both_rec u v : subword R u v -> forall l m, u = map f l -> v = map f m -> subword (fun x y => R (f x) (f y)) l m.
  Proof.
    induction 1.
    intros [ | ] [ | ]; try discriminate 1; try discriminate 2; constructor.
    intros l [ | ? ? ]; try discriminate 2; intros H1 H2.
    injection H2; clear H2; intros; subst; auto; constructor 2; auto.
    intros [ | ? ? ] [ | ? ? ]; try discriminate 1; try discriminate 2; intros H1 H2.
    injection H1; injection H2; clear H1 H2; intros; subst; auto; constructor 3; auto.
  Qed.

  Fact subword_map_both  l m : subword R (map f l) (map f m) <-> subword (fun x y => R (f x) (f y)) l m.
  Proof.
    split.
    intro H; apply sw_map_both_rec with (1 := H); auto.
    induction 1; simpl.
    constructor 1.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

End subword_map_both.

Section subword_props.
  
  Variable (X : Type).

  Fact sublist_eq_subword ll mm : ll <sl mm <-> subword (@eq X) ll mm.
  Proof.
    split.
    
    induction 1.
    apply subword_nil_true.
    constructor 3; auto.
    constructor 2; auto.
    
    induction 1.
    constructor 1.
    constructor 3; auto.
    subst; constructor 2; auto.
  Qed.

  Fact subword_inc (R S : X -> X -> Prop) : R inc2 S -> subword R inc2 subword S.
  Proof.
    intros H l m.
    induction 1.
    constructor 1.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact subword_comp (R S : X -> X -> Prop) : subword R o subword S inc2 subword (R o S).
  Proof.
    intros u w (v & H1 & H2).
    revert H1 w H2.
    induction 1 as [ | u v b H1 IH | u v a b H1 H2 IH ].
    intros ? _; apply subword_nil_true.
    intros w Hw; apply subword_cons_inv in Hw;
    destruct Hw as (l & b' & r & H2 & H3 & H4); subst.
    apply subword_app; constructor 2; auto.
    intros w Hw; apply subword_cons_inv in Hw;
    destruct Hw as (l & b' & r & H3 & H4 & H5); subst.
    apply subword_app; constructor 3; auto.
    exists b; auto.
  Qed.
    
  Fact sw_monotonic : op_monotonic (@subword X X).
  Proof.
    intros ? ?; apply subword_inc.
  Qed.

  Fact sw_w_continuous : op_w_continuous (@subword X X).
  Proof.
    intros K HK.
    induction 1 as [ | ll mm a H (n & IH) | ll mm a b (n & Hn) H (m & IH) ].
    exists 0; constructor 1.
    exists n; constructor 2; auto.
    exists (n+m).
    constructor 3.
    revert Hn; apply seq_inc_all; auto with arith.
    revert IH; apply sw_monotonic, seq_inc_all; auto with arith.
 Qed.

End subword_props.

Section subword_refl_trans.

  Variable (X : Type) (R : X -> X -> Prop).

  Fact subword_refl w : Forall (fun x => R x x) w -> subword R w w.
  Proof.
    rewrite Forall_forall; intros H.
    induction w; [ constructor 1 | constructor 3 ]; auto.
    apply H; left; auto.
    apply IHw; intros; apply H; right; auto.      
  Qed.

  Hypothesis R_trans : transitive _ R.

  Fact subword_trans : transitive _ (subword R). 
  Proof.
    intros u v w H1 H2.
    apply subword_inc with (2 := subword_comp (ex_intro _ v (conj H1 H2))).
    intros x z (y & H3 & ?); apply R_trans with (1 := H3); auto.
  Qed.

End subword_refl_trans.

Fact subword_cup X Y (R S : X -> Y -> Prop) l m : subword (R cup2 S) l m 
                                             -> subword R l m 
                                             \/ exists x y, In x l /\ In y m /\ S x y.
Proof.
  induction 1 as [ | l m x H1 H2 | l m x y H1 H2 H3 ].
  left; constructor.
  destruct H2 as [ H2 | (u & v & H2 & H3 & H4) ].
  left; constructor 2; auto.
  right; exists u, v; simpl; tauto.
  destruct H1 as [ H1 | H1 ].
  destruct H3 as [ H3 | (u & v & H5 & H3 & H4) ].
  left; constructor 3; auto.
  right; exists u, v; simpl; tauto.
  right; exists x, y; simpl; tauto.
Qed.

Section subword_Restr.

  Variables (X : Type) (P : X -> Prop) (R : X -> X -> Prop).

  Let sw_Restr_rec l1 l2 : subword R l1 l2 
                         -> forall m1 m2, 
                            map (@proj1_sig _ _) m1 = l1
                         -> map (@proj1_sig _ _) m2 = l2
                         -> subword (R <# P #>) m1 m2.
  Proof.
    induction 1 as [ | l1 l2 a | l1 l2 a b ].
    intros [ | ] [ | ]; try discriminate 1; try discriminate 2.
    constructor 1.
    intros m1 [ | (x2 & Hx2) m2 ]; try discriminate 2.
    intros H1 H2; injection H2; clear H2; intros H2 ?; subst x2.
    constructor 2; auto.
    intros [ | (x1 & Hx1) m1 ] [ | (x2 & Hx2) m2 ]; try discriminate 2; try discriminate 1.
    intros H1 H2; injection H1; injection H2; clear H1 H2; intros ? ? ? ?; subst x1 x2.
    constructor 3; auto.
  Qed.   
  
  Fact subword_Restr l1 l2 : subword (R <# P #>) l1 l2 
                         <-> subword R (map (@proj1_sig _ _) l1) (map (@proj1_sig _ _) l2).
  Proof.
    split.

    do 2 rewrite subword_sublist_eq.
    intros (mm & H1 & H2).
    exists (map (@proj1_sig _ _) mm).
    split.
    apply Forall2_map_both; auto.
    apply sublist_map; auto.
    
    intros H; apply sw_Restr_rec with (1 := H); auto.
  Qed.
    
End subword_Restr.
 