(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import sublist.
Require Import subword.

Require Import continuity.

Require Import tree.

Set Implicit Arguments.

Section emb_tree_upto.

  Variables (X : Type) (k : nat) (R : nat -> X -> X -> Prop). 
  
  Reserved Notation "x '<et' y" (at level 70, no associativity).

  (** This is the "embeds" definition of page 243-244 of W. Veldman 
   **
   ** U=<x|u1,...,un> neatly embeds into V=<y|v1,...,vp> iff
   **    a/ U neatly embeds into vi for some i in {1,...,p}
   ** or b/ R n x y and n = p and ui neatly embeds into vi for all i in {1,...,n=p}
   **)

  (* Definition of the tree embedding as a least fixpoint using w-cpo fixpoint theory *)

  Let embut_fop (S : relation (tree X)) t1 t2 := 
      (exists t a ll, In t ll /\ S t1 t /\ t2 = in_tree a ll )
   \/ (exists a b ll mm, length ll < k  /\ R (length ll) a b /\ Forall2 S ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm)
   \/ (exists a b ll mm, k <= length ll /\ R  k          a b /\ subword S ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm).

  (* this corresponds to the following "Inductive" definition that is not able to generate
     powerfull enough induction principles 
 
  Inductive emb_tree_upto : tree X -> tree X -> Prop :=
      | in_embut_0 : forall s t x ll, In t ll -> s <et t -> s <et in_tree x ll
      | in_embut_1 : forall x y ll mm, length ll < k -> R (length ll) x y -> Forall2 emb_ntree ll mm -> in_tree x ll <et in_tree y mm
      | in_emnut_2 : forall x y ll mm, k <= length ll -> R k x y -> subword emb_tree_upto ll mm -> in_tree x ll <et in_tree y mm
    where "x <et y" := (emb_tree_upto x y).
    
  *)
    
  Let embut_fop_monotonic : op_monotonic embut_fop.
  Proof.
    intros S T H x y [ (t & a & ll & H1 & H2 & ?)
                   | [ (a & b & ll & mm & H1 & H2 & H3 & ? & ?) 
                     | (a & b & ll & mm & H1 & H2 & H3 & ? & ?) 
                   ] ]; subst.
    left; exists t, a, ll; auto.
    right; left; exists a, b, ll, mm; repeat split; auto.
    revert H3; apply Forall2_impl; auto.
    right; right; exists a, b, ll, mm; repeat split; auto.
    revert H3; apply subword_inc; auto.
  Qed.

  Let embut_fop_w_continuous : op_w_continuous embut_fop.
  Proof.
    intros K HK x y  [ (t & a & ll & H1 & H2 & ?)
                   | [ (a & b & ll & mm & H1 & H2 & H3 & ? & ?) 
                     | (a & b & ll & mm & H1 & H2 & H3 & ? & ?) 
                   ] ]; subst.

    destruct H2 as (n & H2); exists n; left; exists t, a, ll; auto.

    apply Forall2_w_continuous in H3; auto.
    destruct H3 as (n & Hn).
    exists n; right; left; exists a, b, ll, mm; auto.
    
    apply sw_w_continuous in H3; auto. 
    destruct H3 as (n & Hn).
    exists n; right; right; exists a, b, ll, mm; auto.
  Qed.

  Definition emb_tree_upto := lfp embut_fop.

  Infix "<t" := emb_tree_upto (at level 70).

  Let embut_fix t1 t2 : t1 <t t2 <-> (exists t a ll, In t ll /\ t1 <t t /\ t2 = in_tree a ll )
                                  \/ (exists a b ll mm, length ll < k  /\ R (length ll) a b /\ Forall2 emb_tree_upto ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm)
                                  \/ (exists a b ll mm, k <= length ll /\ R  k          a b /\ subword emb_tree_upto ll mm /\ t1 = in_tree a ll /\ t2 = in_tree b mm).
  Proof.
    apply lfp_fixpoint_eq; auto.
  Qed.

  Fact emb_tree_upto_inv x ll y mm : in_tree x ll <t in_tree y mm <-> (exists t, In t mm /\ in_tree x ll <t t)
                                                                  \/ (length ll < k  /\ R (length ll) x y /\ Forall2 emb_tree_upto ll mm)
                                                                  \/ (k <= length ll /\ R k           x y /\ subword emb_tree_upto ll mm).
  Proof.
    rewrite embut_fix; split.

    intros [ (t & a & ll' & H1 & H2 & H3)
         | [ (a & b & ll' & mm' & H1 & H2 & H3 & H4 & H5) 
           | (a & b & ll' & mm' & H1 & H2 & H3 & H4 & H5) 
           ] ]; subst.
    injection H3; intros; subst; left; exists t; auto.
    injection H4; injection H5; intros; subst; right; left; auto.
    injection H4; injection H5; intros; subst; right; right; auto.
    
    intros [ (t & H1 & H2) 
         | [ (H1 & H2 & H3)
           | (H1 & H2 & H3) ] ].
    left; exists t, y, mm; auto.
    right; left; exists x, y, ll, mm; auto.
    right; right; exists x, y, ll, mm; auto.
  Qed. 
    
  (* Let us show that emb_tree_upto is indeed the least relation closed under the
     intended inductive rules *)

  Section emb_tree_upto_closed.

    (* These come in place of the constructors for inductive rules *)
  
    Fact in_emb_tree_upto_0 t t' a ll : In t' ll -> t <t t' -> t <t in_tree a ll.
    Proof.
      rewrite (embut_fix _ (in_tree _ _)).
      left; exists t', a, ll; tauto.
    Qed.

    Fact in_emb_tree_upto_1 a b ll mm : length ll < k -> R (length ll) a b -> Forall2 emb_tree_upto ll mm -> in_tree a ll <t in_tree b mm.
    Proof.
      rewrite (embut_fix _ (in_tree _ _)).
      right; left; exists a, b, ll, mm; tauto.
    Qed.
    
    Fact in_emb_tree_upto_2 a b ll mm : k <= length ll -> R k a b -> subword emb_tree_upto ll mm -> in_tree a ll <t in_tree b mm.
    Proof.
      rewrite (embut_fix _ (in_tree _ _)).
      right; right; exists a, b, ll, mm; tauto.
    Qed.

  End emb_tree_upto_closed.

  (* Hence we have the 3 constructors. Let us provide a general and powerful induction principle emb_tree_upto_ind *)

  Section emb_tree_upto_ind.

    Variable S : tree X -> tree X -> Prop.

    Infix "<<" := S (at level 70).

    Hypothesis S_sub0 : forall t t' a ll, In t' ll -> t <t t' -> t << t' -> t << in_tree a ll.
    Hypothesis S_sub1 : forall a b ll mm, length ll < k  -> R (length ll) a b -> Forall2 emb_tree_upto ll mm -> Forall2 S ll mm -> in_tree a ll << in_tree b mm.
    Hypothesis S_sub2 : forall a b ll mm, k <= length ll -> R k           a b -> subword emb_tree_upto ll mm -> subword S ll mm -> in_tree a ll << in_tree b mm.

    Fact emb_tree_upto_ind : forall t1 t2, t1 <t t2 -> S t1 t2.
    Proof.
      apply lfp_ind; auto.     
      intros x y [ (t & a & ll' & H1 & (H2 & H3) & H4)
               | [ (a & b & ll' & mm' & H1 & H2 & H3 & H4 & H5) 
                 | (a & b & ll' & mm' & H1 & H2 & H3 & H4 & H5) 
                 ] ]; subst.
      apply S_sub0 with (1 := H1); auto.
      apply S_sub1 with (1 := H1); auto; revert H3; apply Forall2_impl; tauto.
      apply S_sub2 with (1 := H1); auto; revert H3; apply sw_monotonic; tauto.
    Qed.
     
  End emb_tree_upto_ind.

  (* we specialize it to well-formed trees *)

  Section emb_wfptree_upto_ind.

    Variable P : nat -> X -> Prop.
    Variable S : tree X -> tree X -> Prop.

    Infix "<<" := S (at level 70).
  
    Hypothesis S_sub0 : forall t t' a ll, wfptree P t 
                                       -> wfptree P (in_tree a ll)
                                       -> In t' ll 
                                       -> t <t t' 
                                       -> t << t' 
                                       -> t << in_tree a ll.
  
    Hypothesis S_sub1 : forall a b ll mm, wfptree P (in_tree a ll)
                                       -> wfptree P (in_tree b mm)
                                       -> length ll < k  
                                       -> R (length ll) a b 
                                       -> Forall2 emb_tree_upto ll mm 
                                       -> Forall2 S ll mm 
                                       -> in_tree a ll << in_tree b mm.

    Hypothesis S_sub2 : forall a b ll mm, wfptree P (in_tree a ll)
                                       -> wfptree P (in_tree b mm)
                                       -> k <= length ll 
                                       -> R k a b 
                                       -> subword emb_tree_upto ll mm 
                                       -> subword S ll mm 
                                       -> in_tree a ll << in_tree b mm.

    Theorem emb_wfptree_upto_ind t1 t2 : wfptree P t1 -> wfptree P t2 -> t1 <t t2 -> S t1 t2.
    Proof.
      intros H1 H2 H; revert H H1 H2.
      induction 1 as [ s t x ll Hll H IH 
                     | x y ll mm Hll Ha Hb IH
                     | x y ll mm Hll Ha Hb IH 
                     ] using emb_tree_upto_ind; intros H1 H2.
                   
      generalize H2; rewrite wfptree_fix; intros (? & ?). 
      apply S_sub0 with t; auto.
      
      apply S_sub1; auto.
      apply wfptree_fix, proj2, Forall_forall in H1.
      apply wfptree_fix, proj2, Forall_forall in H2.
      apply Forall2_Forall_left with (1 := IH) in H1.
      apply Forall2_Forall_right with (1 := H1); auto.
      
      apply S_sub2; auto.
      rewrite subword_sublist_eq in IH.
      destruct IH as (pp & H3 & H4).
      apply wfptree_fix, proj2, Forall_forall in H1.
      apply wfptree_fix, proj2 in H2.
      assert (forall t, In t pp -> wfptree P t) as H5.
        intros t Ht; apply H2; revert Ht.
        apply sl_In; auto.
      rewrite <- Forall_forall in H5.
      apply Forall2_Forall_left with (1 := H3) in H1.
      apply Forall2_Forall_right with (1 := H1) in H5.
      rewrite subword_sublist_eq; exists pp; auto.
    Qed.

  End emb_wfptree_upto_ind.

  Fact in_emb_tree_upto_nil x y : R 0 x y -> in_tree x nil <t in_tree y nil.
  Proof.
    intros H.
    case_eq k.
    intros Hk.
    apply in_emb_tree_upto_2.
    rewrite Hk; auto.
    rewrite Hk; auto.
    constructor.
    intros j Hj.
    apply in_emb_tree_upto_1.
    simpl; omega.
    simpl; auto.
    constructor.
  Qed.

  Section emb_tree_upto_refl_trans.

    Hypothesis R_refl : forall n, n <= k -> reflexive _ (R n).
    
    Fact emb_tree_upto_refl : reflexive _ emb_tree_upto.
    Proof.
      intros t; induction t as [ a ll IH ].
      rewrite <- Forall_forall in IH.
      destruct (le_lt_dec k (length ll)).

      apply in_emb_tree_upto_2; auto.
      apply R_refl; auto.
      apply subword_refl; auto.
      
      apply in_emb_tree_upto_1; auto.
      apply R_refl; omega.
      rewrite Forall2_Forall; auto.
    Qed.
  
  End emb_tree_upto_refl_trans.

  Section emb_tree_upto_dec.
  
    Hypothesis R_dec : forall n x y, n <= k -> { R n x y } + { ~ R n x y }.

    Definition emb_tree_upto_dec s t : { s <t t } + { ~ s <t t }.
    Proof.
      revert s; induction t as [ y l2 IH ]; intros t1.
      
      destruct (list_dec_rect (fun z => t1 <t z) (fun z => ~ t1 <t z) l2)
        as [ (t2 & H1 & H2) | Ht1 ].
        
      intros x Hx; specialize (IH x Hx t1); tauto.
      
      apply In_t_In in H1.
      left; revert H2; apply in_emb_tree_upto_0; auto.
      
      rewrite In_In_t in Ht1.
      destruct t1 as [ x l1 ].
      destruct (le_lt_dec k (length l1)) as [ Hl1 | Hl1 ].
      
      destruct (R_dec x y (le_refl _)) as [ H1 | H1 ].
      
      destruct (subword_dec emb_tree_upto l1 l2) as [ H2 | H2 ].
      intros t1 t2 H3 H4; apply IH; auto.
      
      left; apply in_emb_tree_upto_2; auto.

      right; contradict H2.
      apply emb_tree_upto_inv in H2.
      destruct H2 as [ ( t & H2 & H3) | [ (H2 & H3 & H4) | (H2 & H3 & H4) ] ]; auto. 
      apply Ht1 in H3; tauto.
      apply subword_Forall2; auto.
      
      right; contradict H1.
      apply emb_tree_upto_inv in H1.
      destruct H1 as [ ( t & H2 & H3) | [ (H2 & H3 & H4) | (H2 & H3 & H4) ] ]; auto.
      apply Ht1 in H3; tauto.
      omega.
      
      assert (Hl1' : length l1 <= k). omega.
      destruct (R_dec x y Hl1') as [ H1 | H1 ].
      
      destruct (Forall2_dec emb_tree_upto l1 l2) as [ H2 | C ]; auto.
      
      left; apply in_emb_tree_upto_1; auto.
      
      right; contradict C.
      apply emb_tree_upto_inv in C.
      destruct C as [ (t & H2 & H3) | [ (H2 & H3 & H4) | (H2 & H3 & H4) ] ]; auto.
      apply Ht1 in H3; tauto.
      omega.
      
      right; contradict H1.
      apply emb_tree_upto_inv in H1.
      destruct H1 as [ (t & H2 & H3) | [ (H2 & H3 & H4) | (H2 & H3 & H4) ] ]; auto.
      apply Ht1 in H3; tauto.
      omega.
    Qed.

  End emb_tree_upto_dec.

  Fact emb_sub_tree_upto r s t : r <t s -> s <st t -> r <t t.
  Proof.
    intros Hrs.
    induction 1 as [ | ? ? ? ? H ]; auto.
    apply in_emb_tree_upto_0 with (1 := H); auto.
  Qed.
  
  Fact sub_emb_tree_upto r s t : r <st s -> s <t t -> r <t t.
  Proof.
    intros H0.
    induction 1 as [ s t x ll Hll H IH 
                   | x y ll mm Hll Ha Hb IH
                   | x y ll mm Hll Ha Hb IH 
                   ] using emb_tree_upto_ind.

    apply in_emb_tree_upto_0 with (1 := Hll); auto.

    apply sub_tree_inv in H0.
    destruct H0 as [ H0 | H0 ].
    subst; apply in_emb_tree_upto_1; auto.    
    apply ssub_tree_inv_right in H0.
    destruct H0 as (s & H3 & z & kk & H4 & H5).
    injection H5; clear H5; intros ? ?; subst z kk.
    apply in_split in H4; destruct H4 as (u & v & H4); subst ll.
    apply Forall2_app_inv_l in IH.
    destruct IH as (c & d & _ & IH2 & ?); subst.
    apply Forall2_cons_inv_l in IH2.
    destruct IH2 as (t & e & IH2 & ? & _); subst.
    apply in_emb_tree_upto_0 with t; auto.
    apply in_or_app; right; simpl; auto.
    
    apply sub_tree_inv in H0.
    destruct H0 as [ H0 | H0 ].
    subst; apply in_emb_tree_upto_2; auto.    
    apply ssub_tree_inv_right in H0.
    destruct H0 as (s & H3 & z & kk & H4 & H5).
    injection H5; clear H5; intros ? ?; subst z kk.
    apply in_split in H4; destruct H4 as (u & v & H4); subst ll.
    apply subword_app_inv in IH.
    destruct IH as (c & d & ? & _ & IH2); subst.
    apply subword_cons_inv in IH2.
    destruct IH2 as (l & e & q & ? & IH2 & _); subst.
    apply in_emb_tree_upto_0 with e; auto.
    do 2 (apply in_or_app; right); simpl; auto.
  Qed.
    
  Fact sub_emb_tree_upto_leaf_inv x y t : R 0 x y -> in_tree y nil <st t -> in_tree x nil <t t.
  Proof.
    revert x.
    induction t as [ x2 ll IH ]; intros x1 Hx1.
    intros H.
    apply sub_in_tree_inv in H.
    destruct H as [ H | (? & H1 & H2) ].
    
    injection H; clear H; intros; subst.
    apply in_emb_tree_upto_nil; auto.
    apply in_emb_tree_upto_0 with x; auto.
  Qed.

End emb_tree_upto.

Fact emb_tree_upto_inc X k (R1 R2 : nat -> relation X) : (forall n, R1 n inc2 R2 n) -> emb_tree_upto k R1 inc2 emb_tree_upto k R2.
Proof.
  intros H0.
  induction 1 as [ s t x ll Hll H IH 
                 | x y ll mm Hll Ha Hb IH
                 | x y ll mm Hll Ha Hb IH 
                 ] using emb_tree_upto_ind.
  apply in_emb_tree_upto_0 with t; auto.
  apply in_emb_tree_upto_1; auto.
  apply in_emb_tree_upto_2; auto.
Qed.
