(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.
Require Import Bool.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import finite.
Require Import intercalate.

Require Import sublist.
Require Import subword.

Require Import good_base.
Require Import bar_t.
Require Import good_bar_t.
Require Import af_t.
Require Import af_bar_t.
Require Import fan_t.
Require Import afs_t.

Require Import af.
Require Import bar.
Require Import good_bar.
Require Import af_bar.
Require Import fan.
Require Import afs.

Require Import higman_lemma. (* subword preserves af_t *)

Require Import tree.
Require Import tree_embeddings.
Require Import tree_kruskal_fix.

Set Implicit Arguments.

Section tree_kruskal_ev4.

  Variable X : Type.
    
  Notation T := (ktree_fix X).

  Variables (k : nat) (P : nat -> T -> Prop)
            (HP : forall n, ~ P n in_ktf_u)
            (Pdec : forall n x, { P n x } + { ~ P n x }) 
            (Peq : forall n, k <= n -> P k = P n)
            (i : nat) (H_k_i : S i < k).

  Definition ktf_pos4 l : { l < i } + { l = i } + { l = S i } + { S i < l }.
  Proof.
    destruct (le_lt_dec i l); try tauto.
    destruct (eq_nat_dec l i); try tauto.
    destruct (eq_nat_dec l (S i)); try tauto.
    right; abstract omega.
  Qed.

  (* we use intercalate instead of list_insert for value of llt
     such that contains extacly on singleton and all other values are nil *)

  Definition wf_ktf4 x llt := length llt = S i /\ lsum (map (@length _) llt) = 1 /\ P (S i) x /\ Forall (Forall (wfptree P)) llt.
  
  Definition wf_ktf4_dec x llt : { wf_ktf4 x llt } + { ~ wf_ktf4 x llt }.
  Proof.
    clear H_k_i.
    unfold wf_ktf4.
    assert (Hd : forall lt, In_t lt llt -> { Forall (wfptree P) lt } + { ~ Forall (wfptree P) lt }).
      intros; apply Forall_dec; intros; apply wfptree_dec, Pdec.
    destruct (eq_nat_dec (length llt) (S i));
    destruct (eq_nat_dec (lsum (map (@length _) llt)) 1);
    destruct (Pdec (S i) x);
    destruct (Forall_dec _ _ Hd); tauto.
  Qed.

  Section ktf_ev4.

    Definition ktf_prop4 l : T -> Prop :=
      match ktf_pos4 l with
        | inleft (inleft (left _))  => P l
        | inleft (inleft (right _)) => 
            fun t => match t with
              | in_ktf_1 x     => P i x
              | in_ktf_3 x llt => wf_ktf4 x llt
              | _ => False
            end
        | inleft (inright _)   => fun _ => False
        | inright _            => P l
      end.  

    Notation P1 := ktf_prop4.

    Section ktf_prop4_eqs.

      Ltac crush H C l i := 
        unfold ktf_prop4;
        destruct (ktf_pos4 l) as [[[C|C]|C]|C]; auto;
        try omega.

      Ltac doit := 
        let C := fresh 
        in match goal with 
          | [ H: ?l <> ?i |- _ ] => crush H C l i
          | [ H: ?l = ?i  |- _ ] => crush H C l i
        end.

      Fact ktf_prop4_neq l : l <> i -> l <> S i -> P1 l = P l.
      Proof. intros; doit. Qed.
      
      Fact ktf_prop4_eqS l : l = S i -> P1 l = fun _ => False.
      Proof. intros; doit. Qed.
      
      Variable (l : nat) (x0 : X) (x : T) (t : tree T) (ll : list (tree T)) (llt : list (list (tree T))) (H : l = i). 

      Fact ktf_prop4_eq_u : P1 l  in_ktf_u = False.                  Proof. doit. Qed.
      Fact ktf_prop4_eq_0 : P1 l (in_ktf_0 x0) = False.              Proof. doit. Qed.
      Fact ktf_prop4_eq_1 : P1 l (in_ktf_1 x) = P i x.               Proof. doit. Qed.
      Fact ktf_prop4_eq_3 : P1 l (in_ktf_3 x llt) = wf_ktf4 x llt.   Proof. doit. Qed.

    End ktf_prop4_eqs.
    
    Fact ktf_prop4_u l : ~ P1 l in_ktf_u.
    Proof.
      destruct (ktf_pos4 l) as [[[|]|]|].
      rewrite ktf_prop4_neq; auto; omega.
      rewrite ktf_prop4_eq_u; auto.
      rewrite ktf_prop4_eqS; auto.
      rewrite ktf_prop4_neq; auto; omega.
    Qed.

    Fact ktf_prop4_dec l x : { P1 l x } + { ~ P1 l x }.
    Proof.
      destruct (ktf_pos4 l) as [[[|]|]|].
      rewrite ktf_prop4_neq; auto; omega.
      destruct x as [ | x | x | x llt ].
      rewrite ktf_prop4_eq_u; auto.
      rewrite ktf_prop4_eq_0; auto.
      rewrite ktf_prop4_eq_1; auto.
      rewrite ktf_prop4_eq_3; auto.
      apply wf_ktf4_dec.
      rewrite ktf_prop4_eqS; auto.
      rewrite ktf_prop4_neq; auto; omega.
    Qed.

    Let st (x : T) mm :=
      if eq_nat_dec (length mm) i then
         match x with
           | in_ktf_1 x     => in_tree x mm
           | in_ktf_3 x llt => in_tree x (intercalate mm llt)

           | _              => in_tree in_ktf_u nil
         end
       else in_tree x mm.

    Definition ktf_ev4 : tree T -> tree T.
    Proof.
      intros t; induction t as [ x ll mm ] using tree_recursion.
      apply (st x mm).
    Defined.
    
    Fact ktf_ev4_fix x ll : ktf_ev4 (in_tree x ll) = st x (map ktf_ev4 ll).
    Proof. apply tree_recursion_fix. Qed.

    (* Remark : for non well-formed inputs, we output an arbitrary 
                non well-formed tree such as in_tree in_ktf_u nil *)
    
    Fact ktf_ev4_fix' x ll : ktf_ev4 (in_tree x ll) = 
       if eq_nat_dec (length ll) i then
         match x with
           | in_ktf_1 x     => in_tree x (map ktf_ev4 ll)
           | in_ktf_3 x llt => in_tree x (intercalate (map ktf_ev4 ll) llt)
           | _              => in_tree in_ktf_u nil
         end
       else in_tree x (map ktf_ev4 ll).
    Proof.
      rewrite ktf_ev4_fix.
      unfold st; rewrite map_length.
      reflexivity.
    Qed. 
    
    Section ktf_ev4_eqs.

      Ltac crush H ll i := 
        rewrite ktf_ev4_fix';
        destruct (eq_nat_dec (length ll) i); simpl; auto;
        contradict H; auto.

      Ltac doit := match goal with 
                     | [ H: length ?ll <> ?i |- _ ] => crush H ll i 
                     | [ H: length ?ll = ?i  |- _ ] => crush H ll i 
                   end.

      Fact ktf_ev4_neq x ll : length ll <> i -> ktf_ev4 (in_tree x ll) = in_tree x (map ktf_ev4 ll). 
      Proof. intro; doit. Qed.
      
      Variables (x0 : X) (x : T) (t : tree T) (ll : list (tree T)) (llt : list (list (tree T))) (H : length ll = i). 
      
      Fact ktf_ev4_eq_u : ktf_ev4 (in_tree in_ktf_u ll) = in_tree in_ktf_u nil.            Proof. doit. Qed.
      Fact ktf_ev4_eq_0 : ktf_ev4 (in_tree (in_ktf_0 x0) ll) = in_tree in_ktf_u nil.       Proof. doit. Qed.
      Fact ktf_ev4_eq_1 : ktf_ev4 (in_tree (in_ktf_1 x) ll) = in_tree x (map ktf_ev4 ll).  Proof. doit. Qed.
      Fact ktf_ev4_eq_3 : ktf_ev4 (in_tree (in_ktf_3 x llt) ll) = in_tree x (intercalate (map ktf_ev4 ll) llt).
      Proof. doit. Qed.

    End ktf_ev4_eqs.

    Let st_wf x lt : P1 (length lt) x -> Forall (wfptree P) lt -> wfptree P (st x lt).
    Proof.
      intros H1 H2.
      rewrite Forall_forall in H2.
      unfold st.
      destruct (eq_nat_dec (length lt) i) as [ H | H ].
      destruct x as [ | x0 | x | x llt ].
      apply ktf_prop4_u in H1; destruct H1.
      rewrite ktf_prop4_eq_0 in H1; auto; destruct H1.
      rewrite ktf_prop4_eq_1 in H1; auto.
      rewrite wfptree_fix, H; split; auto.
      rewrite ktf_prop4_eq_3 in H1; auto.
      destruct H1 as (H1 & H3 & H4 & H5).
      rewrite wfptree_fix; split.
      replace_with H4.
      rewrite intercalate_length_eq; omega.
      intros t Ht.
      rewrite intercalate_In_spec in Ht.
      destruct Ht as [ Ht | (u & Hu1 & Hu2) ].
      apply H2; auto.
      rewrite Forall_forall in H5.
      apply H5 in Hu2.
      rewrite Forall_forall in Hu2; auto.
      destruct (eq_nat_dec (length lt) (S i)).
      rewrite ktf_prop4_eqS in H1; auto; destruct H1.
      rewrite ktf_prop4_neq in H1; auto.
      rewrite wfptree_fix; split; auto.
    Qed.

    Fact ktf_ev4_wf t : wfptree P1 t -> wfptree P (ktf_ev4 t).
    Proof.
      induction 1 as [ x ll H IH ] using wfptree_rect.
      rewrite wfptree_fix in H; destruct H.
      rewrite In_In_t in IH.
      rewrite ktf_ev4_fix.
      apply st_wf.
      rewrite map_length; auto.
      rewrite Forall_forall; auto.
      intros t Ht.
      apply in_map_iff in Ht.
      destruct Ht as (? & ? & ?); subst; auto.
    Qed.

    Fact ktf_ev4_sublist x lx y ly : P1 (length lx) x
                           -> ktf_ev4 (in_tree x lx) = in_tree y ly
                           -> map ktf_ev4 lx <sl ly.
    Proof. 
      destruct (eq_nat_dec (length lx) i).
      destruct x as [ | x0 | x | x lw ].
      rewrite ktf_prop4_eq_u; auto; destruct 1.
      rewrite ktf_prop4_eq_0; auto; destruct 1.
      rewrite ktf_ev4_eq_1; auto. 
      injection 2; intros; subst y ly; apply sl_refl.
      rewrite ktf_prop4_eq_3; auto.
      rewrite ktf_ev4_eq_3; auto.
      intros (H1 & H2 & H3 & H5) H4.
      injection H4; clear H4; intros ? ?; subst y ly.
      apply intercalate_sublist.
      rewrite ktf_ev4_neq; auto.
      injection 2; intros; subst; apply sl_refl.
    Qed.
      
    Corollary ktf_ev4_length x lx y ly : P1 (length lx) x
                           -> ktf_ev4 (in_tree x lx) = in_tree y ly
                           -> length lx <= length ly.
    Proof.
      intros H1 H2.
      rewrite <- (map_length ktf_ev4 lx).
      apply sublist_length.
      revert H1 H2; apply ktf_ev4_sublist.
    Qed.

    Corollary ktf_ev4_In x lx y ly t : wfptree P1 (in_tree x lx) 
                               -> In t lx 
                               -> ktf_ev4 (in_tree x lx) = in_tree y ly
                               -> In (ktf_ev4 t) ly.
    Proof.
      intros H1 H2 H3.
      rewrite wfptree_fix in H1.
      destruct H1 as (H0 & H1).
      generalize (ktf_ev4_sublist _ _ H0 H3); intros H4.
      apply (sl_In _ H4), in_map_iff; exists t; auto.
    Qed.

    Definition ktf_ev4_sig : { t | wfptree P1 t } -> { t | wfptree P t }.
    Proof. intros (t & Ht); exists (ktf_ev4 t); apply ktf_ev4_wf; assumption. Defined.

    Notation analysis4 := (fun t a => wfptree P1 a /\ ktf_ev4 a = t).
   
    Let analysis4_length x lx y ly : analysis4 (in_tree x lx) (in_tree y ly)  
                                   -> length ly <= length lx.
    Proof.
      intros [ ]; rewrite wfptree_fix.
      intros (H1 & H2) H3.
      revert H1 H3; apply ktf_ev4_length.
    Qed.     

    Section analysis4_case1.
  
      Variables (y : T) (ly : list (tree T)) (Hly : length ly < i) (H : wfptree P (in_tree y ly)).
      
      Fact analysis4_spec1 a : analysis4 (in_tree y ly) a <-> exists la, Forall2 analysis4 ly la /\ a = in_tree y la.
      Proof.
        rewrite wfptree_fix in H.
        destruct H as (H1 & H2).
        clear H.
        split.
        
        destruct a as [ x lx ].
        intros H3. 
        generalize (analysis4_length H3); intros H4.
        revert H3.
        rewrite wfptree_fix. 
        rewrite ktf_prop4_neq; try omega.
        rewrite ktf_ev4_neq; try omega.
        intros ((H3 & H5) & H6).
        injection H6; clear H6; intros ? ?; subst y ly.
        exists lx; split; auto.
        rewrite Forall2_exchg, Forall2_analysis, Forall_forall; auto.

        intros (la & H3 & H4).
        rewrite Forall2_exchg, Forall2_analysis in H3.
        destruct H3 as (H3 & H5).
        subst a ly.
        rewrite map_length in Hly.
        rewrite map_length in H1.
        rewrite Forall_forall in H3.
        rewrite wfptree_fix.
        rewrite ktf_prop4_neq; try omega.
        rewrite ktf_ev4_neq; try omega.
        repeat split; auto.
      Qed.
         
      Fact analysis4_eq1 : analysis4 (in_tree y ly) 
                      ~eq1 fun a => exists la, Forall2 analysis4 ly la /\ a = in_tree y la.
      Proof.
        intros; split; intro; rewrite analysis4_spec1; auto.
      Qed.

    End analysis4_case1.
    
    Section analysis4_case2.
  
      Variables (y : T) (ly : list (tree T)) (Hly : length ly = i) (H : wfptree P (in_tree y ly)).
      
      Fact analysis4_spec2 a : analysis4 (in_tree y ly) a <-> exists la, Forall2 analysis4 ly la /\ a = in_tree (in_ktf_1 y) la.
      Proof.
        rewrite wfptree_fix in H.
        destruct H as (H1 & H2).
        clear H.
        split.
        
        destruct a as [ x lx ].
        intros H3. 
        generalize (analysis4_length H3); intros H4.
        revert H3.
        rewrite wfptree_fix.
        destruct (eq_nat_dec (length lx) i).          
        destruct x as [ | x0 | x | x llt ].
        rewrite ktf_prop4_eq_u; tauto.
        rewrite ktf_prop4_eq_0; tauto.
        rewrite ktf_prop4_eq_1; auto.
        rewrite <- Hly.
        rewrite ktf_ev4_eq_1; auto.
        intros ((H3 & H5) & H6).
        injection H6; clear H6; intros ? ?; subst y ly.
        exists lx; split; auto.
        rewrite Forall2_exchg, Forall2_analysis, Forall_forall; auto.
        rewrite ktf_prop4_eq_3; auto.
        rewrite ktf_ev4_eq_3; auto.
        intros ((H3 & H5) & H6).
        unfold wf_ktf4 in H3.
        destruct H3 as (H3 & H8 & H7 & H9).
        injection H6; clear H6; intros H6 ?; subst y.
        rewrite <- H6, intercalate_length_eq, map_length in Hly.
        omega.        
        rewrite ktf_prop4_neq; try omega.
        rewrite ktf_ev4_neq; try omega.
        intros ((H3 & H5) & H6).
        injection H6; clear H6; intros ? ?; subst y ly.
        rewrite map_length in Hly; contradict Hly; auto.

        intros (la & H3 & H4).
        rewrite Forall2_exchg, Forall2_analysis in H3.
        destruct H3 as (H3 & H5).
        subst a ly.
        rewrite map_length in Hly.
        rewrite map_length in H1.
        rewrite Forall_forall in H3.
        rewrite wfptree_fix.
        rewrite ktf_prop4_eq_1; auto.
        rewrite ktf_ev4_eq_1; auto.
        rewrite <- Hly; auto.
      Qed.
         
      Fact analysis4_eq2 : analysis4 (in_tree y ly) 
                      ~eq1 fun a => exists la, Forall2 analysis4 ly la /\ a = in_tree (in_ktf_1 y) la.
      Proof.
        intros; split; intro; rewrite analysis4_spec2; auto.
      Qed.

    End analysis4_case2.
    
    Section analysis4_case3.
  
      Variables (y : T) (ly : list (tree T)) (Hly : length ly = S i) (H : wfptree P (in_tree y ly)).
      
      Fact analysis4_spec3 a : analysis4 (in_tree y ly) a <-> (exists w1 w2 lw, 
                                                                        hig_split ly (w1,lw)
                                                                     /\ Forall2 analysis4 w1 w2 
                                                                     /\ a = in_tree (in_ktf_3 y lw) w2).
      Proof.
        rewrite wfptree_fix in H.
        destruct H as (H1 & H2).
        clear H.
        split.
        
        destruct a as [ x lx ].
        rewrite wfptree_fix.
        intros ((H3 & H4) & H5). 
        destruct (eq_nat_dec (length lx) i).

        destruct x as [ | | x | x llw ].
        rewrite ktf_prop4_eq_u in H3; tauto.
        rewrite ktf_prop4_eq_0 in H3; tauto.

        rewrite ktf_ev4_eq_1 in H5; auto.
        injection H5; clear H5; intros ? ?; subst y ly.
        rewrite map_length in Hly; omega.
               
        rewrite ktf_prop4_eq_3 in H3; auto.
        destruct H3 as (H3 & H6 & H7 & H8).
        rewrite ktf_ev4_eq_3 in H5; auto.        
        injection H5; clear H5; intros ? ?; subst y ly.
        exists (map ktf_ev4 lx), lx, llw; repeat split; auto.
        rewrite intercalate_length_eq, map_length; omega.
        rewrite intercalate_length_eq, map_length; omega.
        rewrite Forall2_exchg, Forall2_analysis, Forall_forall; auto.
        
        destruct (eq_nat_dec (length lx) (S i)).
        rewrite ktf_prop4_eqS in H3; auto; destruct H3.
        
        rewrite ktf_prop4_neq in H3; auto.
        rewrite ktf_ev4_neq in H5; auto.
        injection H5; clear H5; intros ? ?; subst y ly.
        rewrite map_length in Hly.
        contradict Hly; auto.

        intros (w1 & w2 & lw & H3 & H7 & H8).
        subst a.
        generalize (hig_split_length H3); intros H9.
        destruct H3 as (H3 & H4 & H5).
        rewrite Hly in H4.
        injection H4; clear H4; intros H4.        
        generalize (Forall2_length H7); intros H8.
        symmetry in H8.
        rewrite wfptree_fix.
        rewrite H4 in H8.
        rewrite ktf_prop4_eq_3; auto.
        unfold wf_ktf4.
        rewrite ktf_ev4_eq_3; auto.
        rewrite Forall2_exchg, Forall2_analysis in H7.
        destruct H7 as (H7 & ?); subst w1.
        rewrite Forall_forall in H7.
        rewrite H5.
        repeat split; auto.
        omega.
        replace_with H1; auto.
        rewrite Forall_forall.
        intros w Hw; rewrite Forall_forall.
        intros t Ht.
        apply H2.
        rewrite H5.
        rewrite intercalate_In_spec.
        right; exists w; auto.        
      Qed.

      Fact analysis4_eq3 : analysis4 (in_tree y ly)
                       ~eq1 fun a => (exists d, hig_split ly d 
                                        /\ exists w2, Forall2 analysis4 (fst d) w2
                                                   /\ a = in_tree (in_ktf_3 y (snd d)) w2).
      Proof.
        split; intro; rewrite analysis4_spec3.
        intros (w1 & w2 & lw & G1 & G2 & G4).
        exists (w1,lw); split; auto; exists w2; auto.
        intros ((w1 & lw) & G1 & w2 & G2 & G3).
        exists w1, w2, lw; auto.
      Qed.
 
    End analysis4_case3.

    Section analysis4_case4.
  
      Variables (y : T) (ly : list (tree T)) (Hly : S i < length ly) (H : wfptree P (in_tree y ly)).
      
      Fact analysis4_spec4 a : analysis4 (in_tree y ly) a <-> (exists la, Forall2 analysis4 ly la /\ a = in_tree y la).
      Proof.
        rewrite wfptree_fix in H.
        destruct H as (H1 & H2).
        clear H.
        split.
        
        destruct a as [ x lx ].
        rewrite wfptree_fix.
        intros ((H3 & H4) & H5). 
        destruct (eq_nat_dec (length lx) i).

        destruct x as [ | | x | x llw ].
        rewrite ktf_prop4_eq_u in H3; tauto.
        rewrite ktf_prop4_eq_0 in H3; tauto.
        rewrite ktf_ev4_eq_1 in H5; auto.
        injection H5; clear H5; intros ? ?; subst y ly.
        rewrite map_length in Hly; omega.      
        rewrite ktf_prop4_eq_3 in H3; auto.
        destruct H3 as (H3 & H6 & H7 & H8).
        rewrite ktf_ev4_eq_3 in H5; auto.        
        injection H5; clear H5; intros ? ?; subst y ly.
        rewrite intercalate_length_eq, map_length in Hly; omega.
        
        destruct (eq_nat_dec (length lx) (S i)).
        rewrite ktf_prop4_eqS in H3; tauto.
        rewrite ktf_prop4_neq in H3; auto.
        rewrite ktf_ev4_neq in H5; auto.
        injection H5; clear H5; intros ? ?; subst y ly.
        rewrite map_length in H1.
        exists lx.
        rewrite Forall2_exchg, Forall2_analysis, Forall_forall; auto.

        intros (la & H3 & H4).

        rewrite Forall2_exchg, Forall2_analysis in H3.
        destruct H3 as (H3 & H5).
        subst a ly.
        rewrite map_length in Hly.
        rewrite map_length in H1.
        rewrite Forall_forall in H3.
        rewrite wfptree_fix.
        rewrite ktf_prop4_neq; try omega.
        rewrite ktf_ev4_neq; try omega; auto.
      Qed.

      Fact analysis4_eq4 :  analysis4 (in_tree y ly)
                       ~eq1 fun a => (exists la, Forall2 analysis4 ly la /\ a = in_tree y la).
      Proof.
        split; intro; rewrite analysis4_spec4; auto.
      Qed.
 
    End analysis4_case4.     
 
    Fact ktf_ev4_inv_finite_t t : wfptree P t -> finite_t (analysis4 t).
    Proof.
      induction 1 as [ y ly H1 IH ] using wfptree_rect.
      
      destruct (ktf_pos4 (length ly)) as [[[H|H]|H]|H].

      generalize (@analysis4_eq1 y ly H H1); intros H2.
      apply eq1_sym in H2.
      apply finite_t_eq with (1 := H2).
      apply finite_t_comp.
      apply finite_t_rel_inv; auto.
      intros; apply finite_t_img.
      
      generalize (@analysis4_eq2 y ly H H1); intros H2.
      apply eq1_sym in H2.
      apply finite_t_eq with (1 := H2).
      apply finite_t_comp.
      apply finite_t_rel_inv; auto.
      intros; apply finite_t_img.
   
      generalize (@analysis4_eq3 y ly H H1); intros H2.
      apply eq1_sym in H2.
      apply finite_t_eq with (1 := H2).
      apply finite_t_comp.
      apply hig_split_finite_t.
      intros (w,lw) (G1 & G2 & G3); apply finite_t_comp.
      apply finite_t_rel_inv; auto.
      simpl. 
      intros; apply IH.
      rewrite G3.
      apply intercalate_In_t; left; auto.
      intros; apply finite_t_img.
      
      generalize (@analysis4_eq4 y ly H H1); intros H2.
      apply eq1_sym in H2.
      apply finite_t_eq with (1 := H2).
      apply finite_t_comp.
      apply finite_t_rel_inv; auto.
      intros; apply finite_t_img.
    Qed.
      
    (* we have finished analysising analysis4 ... *)
      
    Definition ktf_ev4_inv t Ht := proj1_sig (ktf_ev4_inv_finite_t t Ht).
    
    Fact ktf_ev4_inv_spec t Ht : forall x, In x (ktf_ev4_inv t Ht) <-> analysis4 t x.
    Proof.
      intro; apply (proj2_sig (ktf_ev4_inv_finite_t _ Ht)).
    Qed.

    Section ktf_ev4_inv_sig.
  
      Let ktf_ev4_inv_spec1 t Ht x : In x (ktf_ev4_inv t Ht) -> wfptree P1 x.
      Proof.
        rewrite ktf_ev4_inv_spec; tauto.
      Qed.

      Let ktf_ev4_inv_spec2 t Ht x : ktf_ev4 x = t -> wfptree P1 x -> In x (ktf_ev4_inv t Ht).
      Proof.
        rewrite ktf_ev4_inv_spec; split; auto.
      Qed.
     
      Definition ktf_ev4_inv_sig (k : { x | wfptree P x }) : list { x | wfptree P1 x } :=
        match k with
          | exist t Ht => list_Forall_sig _ _ (ktf_ev4_inv_spec1 t Ht)
        end.

      Fact ktf_ev4_inv_sig_Forall2 ll mm : 
                   Forall2 (@In _) ll (map ktf_ev4_inv_sig mm) 
                -> map (fun x => proj1_sig (ktf_ev4_sig x)) ll 
                 = map (@proj1_sig _ _) mm.
      Proof.
        rewrite Forall2_map_right.
        induction 1 as [ | (x & Hx) (y & Hy) ll mm Q1 Q2 Q3 ]; simpl; auto.
        simpl in Q1.
        rewrite list_Forall_sig_In in Q1.
        destruct Q1 as (a & Ha & Q1).
        f_equal; auto.
        injection Q1; clear Q1; intros ?; subst a.
        rewrite ktf_ev4_inv_spec in Ha; apply Ha.
      Qed.
 
      Fact ktf_ev4_sig_inv x (Hx : wfptree P1 x) 
                           y (Hy : wfptree P  y) :
                                 ktf_ev4 x = y 
                              -> exists x', x = proj1_sig x' 
                                         /\ In x' (ktf_ev4_inv_sig (exist _ y Hy)).
      Proof.
        intros Hxy.
        exists (exist _ _ (ktf_ev4_inv_spec1 _ Hy _ (ktf_ev4_inv_spec2 Hy _ Hxy Hx))).
        split; auto.
        apply list_Forall_sig_In_refl.
      Qed.

    End ktf_ev4_inv_sig.
  
    Let fin_analysis4 : forall x, wfptree P x -> finite (analysis4 x).
    Proof.
      intros; apply finite_t_finite, ktf_ev4_inv_finite_t; auto.
    Qed.

    Variables (m : T) (lT : list (tree T)) (H_m_lt : wfptree P (in_tree m lT)) (HlT : length lT = S i). 

    Variable (R : nat -> T -> T -> Prop). 

    Definition ktf_rel4_i (t1 t2 : T) := 
      match t1, t2 with
        | in_ktf_1 x1     , in_ktf_1 x2     => R i x1 x2
        | in_ktf_3 x1 lw1 , in_ktf_3 x2 lw2 => R (S i) x1 x2
                                            /\ idx_rel (fun t => subword ((emb_tree_upto k R) rlift t)) lT lw1 lw2
        | _          , _                    => False 
      end.

    Notation Ri' := ktf_rel4_i.

    Fact afs_t_ktf_rel4_i :    
                       afs_t (P i) (R i)
                    -> (forall x y, P (S i) x -> P (S i) y -> R (S i) x y)
                    -> (forall t, In_t t lT -> afs_t (wfptree P) (emb_tree_upto k R rlift t)) 
                    -> afs_t (P1 i) Ri'.
    Proof.
      intros H1 H2 H3.
      apply in_afs_t0 in H2.
      assert (H4 : forall l, { Forall (wfptree P) l } + { ~ Forall (wfptree P) l }).
        intros; apply Forall_dec; intros; apply wfptree_dec, Pdec.
      assert (G1 : forall m, { Forall (Forall (wfptree P)) m } + { ~ Forall (Forall (wfptree P)) m }).
        intros; apply Forall_dec; intros; apply H4.
      generalize (@afs_t_idx_rel _ _ _ H4 (fun t => subword ((emb_tree_upto k R) rlift t)) lT).
      intros H5; spec_all H5.
      intros; apply higman_list_afs_t; auto.
      apply wfptree_dec; auto.
      generalize (afs_t_product H2 H5); clear H5; intros H5.
      spec_all H5; auto.

      intros y.
      destruct (G1 y); destruct (eq_nat_dec (length y) (length lT)); tauto.
      
      generalize (afs_t_sum_both H1 H5); intros H6.
      spec_all H6; auto.
      intros (t,y); simpl.
      unfold pred_prod.
      destruct (Pdec (S i) t); destruct (G1 y); destruct (eq_nat_dec (length y) (length lT)); tauto.
      revert H6.
      
      apply afs_t_relmap with (f := fun k t => 
        match k, t with
          | inl x      , in_ktf_1 y    => x = y
          | inr (x,lv) , in_ktf_3 y lw => x = y /\ lv = lw
          | _    , _          => False
        end); auto.
        
      apply ktf_prop4_dec.
      
      intros [ | ? | x | x lw ].
      rewrite ktf_prop4_eq_u; auto; intros [].
      rewrite ktf_prop4_eq_0; auto; intros [].
      rewrite ktf_prop4_eq_1; auto; intros Hx.
      exists (inl x); simpl; auto.
      rewrite ktf_prop4_eq_3; auto; intros (U1 & U2 & U3 & U4).
      exists (inr (x,lw)).
      unfold pred_prod; simpl. 
      rewrite U1; auto.
      
      intros [ x1 | (x1,lw1) ] [ x2 | (x2,lw2) ]
             [ | | x3 | x3 lw3 ] [ | | x4 | x4 lw4 ]; try tauto; simpl.
      intros; subst; auto.
      intros _ _ _ _ (? & ?) (? & ?); subst; auto.
    Qed.

    Fact afs_ktf_rel4_i :    
                       afs (P i) (R i)
                    -> (forall x y, P (S i) x -> P (S i) y -> R (S i) x y)
                    -> (forall t, In t lT -> afs (wfptree P) (emb_tree_upto k R rlift t)) 
                    -> afs (P1 i) Ri'.
    Proof.
      intros H1 H2 H3.
      apply in_afs_0 in H2.
      assert (H4 : forall l, { Forall (wfptree P) l } + { ~ Forall (wfptree P) l }).
        intros; apply Forall_dec; intros; apply wfptree_dec, Pdec.
      assert (G1 : forall m, { Forall (Forall (wfptree P)) m } + { ~ Forall (Forall (wfptree P)) m }).
        intros; apply Forall_dec; intros; apply H4.
      generalize (@afs_idx_rel _ _ (Forall (wfptree P)) (fun t => subword ((emb_tree_upto k R) rlift t)) lT).
      intros H5; spec_all H5.
      intros; apply higman_list_afs; auto.
      generalize (afs_sum_both H1 (afs_product H2 H5)). 
      
      apply afs_relmap with (f := fun k t => 
        match k, t with
          | inl x      , in_ktf_1 y    => x = y
          | inr (x,lv) , in_ktf_3 y lw => x = y /\ lv = lw
          | _    , _          => False
        end); auto.
      
      intros [ | ? | x | x lw ].
      rewrite ktf_prop4_eq_u; auto; intros [].
      rewrite ktf_prop4_eq_0; auto; intros [].
      rewrite ktf_prop4_eq_1; auto; intros Hx.
      exists (inl x); simpl; auto.
      rewrite ktf_prop4_eq_3; auto; intros (U1 & U2 & U3 & U4).
      exists (inr (x,lw)).
      unfold pred_prod; simpl. 
      rewrite U1; auto.
      
      intros [ x1 | (x1,lw1) ] [ x2 | (x2,lw2) ]
             [ | | x3 | x3 lw3 ] [ | | x4 | x4 lw4 ]; try tauto; simpl.
      intros; subst; auto.
      intros _ _ _ _ (? & ?) (? & ?); subst; auto.
    Qed.

    Definition ktf_rel4 (l : nat) : relation T.
    Proof.
      destruct (ktf_pos4 l) as [[[|]|]|].
      exact (R l).
      exact (Ri').
      exact (fun _ _ => False).
      exact (R l).
    Defined.
  
    Notation R' := ktf_rel4.
    
    Section eqs.

      Variable (l : nat).    

      Ltac doit := 
        unfold R'; intros;
        destruct (ktf_pos4 l) as [[[|]|]|]; try omega; auto.

      Fact ktf_rel4_eq_0 : l < i        -> R' l = R l.               Proof. doit. Qed.
      Fact ktf_rel4_eq_1 : l = i        -> R' l = Ri'.               Proof. doit. Qed.
      Fact ktf_rel4_eq_2 : l = S i      -> R' l = fun _ _ => False.  Proof. doit. Qed.
      Fact ktf_rel4_eq_3 : S i < l      -> R' l = R l.               Proof. doit. Qed.
      
    End eqs.

    Definition ktf_ev4_disapointing (t : tree T) :=
      match t with 
        | in_tree x ll  => length ll = S i /\ R (S i) m x
                        \/ length ll = i 
                        /\ match x with
                             | in_ktf_3 x lw => Exists2 (emb_tree_upto k R o @In _) lT lw
                             | _ => False
                           end
      end.

    Definition ktf_ev4_has_disap  b := exists a, wfptree P1 a /\ ktf_ev4_disapointing a /\ sub_tree a b.

    Variable (HR : forall x y, P (S i) x -> P (S i) y -> R (S i) x y). 

    Fact ktf_ev4_disap_prop2 s :
                wfptree P s  
             -> analysis4 s inc1 ktf_ev4_has_disap
             -> emb_tree_upto k R (in_tree m lT) s.
    Proof.
      induction 1 as [ x ll Hs IH ] using wfptree_rect; intros H0.
      rewrite In_In_t in IH.
      generalize Hs; rewrite wfptree_fix; intros (H5 & H6).
      
      destruct (ktf_pos4 (length ll)) as [[[C|C]|C]|C].

      (* case1 of analysis4 *)
      
      set (U (mm : list (tree T)) := exists t, In t mm /\ ktf_ev4_has_disap t).
      set (V (mm : list (tree T)) := ktf_ev4_disapointing (in_tree x mm)).

      destruct (@finite_decide _ (Forall2 analysis4 ll) U V) as [ H | H ].
      apply finite_Forall2; intros; apply fin_analysis4; auto.
      
      intros mm Hmm.     
      specialize (H0 (in_tree x mm)); spec_all H0.
      rewrite analysis4_spec1; auto.
      exists mm; auto.
      destruct H0 as (a & G1 & G2 & G3).
      apply sub_in_tree_inv in G3.
      destruct G3 as [ G3 | (b & G3 & G4) ].
      right; subst; unfold V; auto.
      left; unfold U; exists b; split; auto; exists a; auto.
      
      unfold U in H; clear U V.
      apply finite_analysis in H.
      destruct H as (t & G1 & G2).
      apply in_emb_tree_upto_0 with t; auto.
      intros; apply fin_analysis4; auto.
      
      unfold V in H; clear U V.
      destruct H as (mm & G1 & G2). 
      apply Forall2_length in G1.
      destruct G2 as [ (G2 & _) | (G2 & _) ]; omega.

      (* case2 of analysis4 *)
      
      set (U (mm : list (tree T)) := exists t, In t mm /\ ktf_ev4_has_disap t).
      set (V (mm : list (tree T)) := ktf_ev4_disapointing (in_tree (in_ktf_1 x) mm)).

      destruct (@finite_decide _ (Forall2 analysis4 ll) U V) as [ H | H ].
      apply finite_Forall2; intros; apply fin_analysis4; auto.
      
      intros mm Hmm.     
      specialize (H0 (in_tree (in_ktf_1 x) mm)); spec_all H0.
      rewrite analysis4_spec2; auto.
      exists mm; auto.
      destruct H0 as (a & G1 & G2 & G3).
      apply sub_in_tree_inv in G3.
      destruct G3 as [ G3 | (b & G3 & G4) ].
      right; subst a; unfold V; auto.
      left; unfold U; exists b; split; auto; exists a; auto.
      
      unfold U in H; clear U V.
      apply finite_analysis in H.
      destruct H as (t & G1 & G2).
      apply in_emb_tree_upto_0 with t; auto.
      intros; apply fin_analysis4; auto.
      
      unfold V in H; clear U V.
      destruct H as (mm & G1 & G2).
      destruct G2 as [ (G2 & _) | (G2 & []) ].
      apply Forall2_length in G1; omega.
      
      (* case3 of analysis4 *)
      
      assert (Hmx : R (S i) m x).
        apply HR; auto.
        rewrite <- HlT; rewrite wfptree_fix in H_m_lt; tauto.
        rewrite <- C; rewrite wfptree_fix in Hs; tauto.
      
      (* the last subcase where we already have R (S i) m x 

         we use all the analysis for ll of the form ll = intercalate w lw

       *)       
      
      assert (forall d, hig_split ll d   -> Exists2 (emb_tree_upto k R o In (A:=tree T)) lT (snd d)
                                         \/ emb_tree_upto k R (in_tree m lT) (in_tree x ll) ) as H7.
        intros (w,lw) H. 
        generalize (hig_split_length H); intros F4.
        destruct H as (F1 & F2 & F3).
        rewrite C in F2; injection F2; clear F2; intros F2.
        assert (F5: forall t, In t w -> In t ll).
          intros; rewrite F3, intercalate_In_spec; left; auto.
        set (U (mm : list (tree T)) := exists t, In t mm /\ ktf_ev4_has_disap t).
        set (V (mm : list (tree T)) := ktf_ev4_disapointing (in_tree (in_ktf_3 x lw) mm)).
        
        destruct (@finite_decide _ (Forall2 analysis4 w) U V) as [ H | H ].
        apply finite_Forall2; intros; apply fin_analysis4; auto.
             
        intros mm Hmm.     
        specialize (H0 (in_tree (in_ktf_3 x lw) mm)); spec_all H0.
        rewrite analysis4_spec3; auto.
        exists w, mm, lw; auto.
        unfold hig_split, hig_decomp.
        repeat split; auto.
        rewrite F3, intercalate_length_eq; omega.
        destruct H0 as (a & G1 & G2 & G3).
        apply sub_in_tree_inv in G3.
        destruct G3 as [ G3 | (b & G3 & G4) ].
        right; subst a; unfold V; auto.
        left; unfold U; exists b; split; auto; exists a; auto.
        
        unfold U in H; clear U V.
        apply finite_analysis in H.
        2: intros; apply fin_analysis4, H6; rewrite F3, intercalate_In_spec; left; auto.
        destruct H as (t & G1 & G2).
        right; apply in_emb_tree_upto_0 with t; auto.
            
        unfold V in H; clear U V.
        destruct H as (mm & G1 & G2).
        destruct G2 as [ (G2 & G3) | (G2 & G3) ].
        apply Forall2_length in G1; omega.
        left; auto.
        
      apply finite_decide in H7.

      destruct H7 as [ H7 | H7 ].
      
      apply hig_split_Forall2 in H7; try omega.
      apply in_emb_tree_upto_1; auto; try omega.
      replace_with Hmx; omega.
      
      destruct H7 as (? & ? & H7); auto.
      
      apply hig_split_finite.
      
      (* case 4 of analysis *)
      
      set (U (mm : list (tree T)) := exists t, In t mm /\ ktf_ev4_has_disap t).
      set (V (mm : list (tree T)) := ktf_ev4_disapointing (in_tree x mm)).

      destruct (@finite_decide _ (Forall2 analysis4 ll) U V) as [ H | H ].
      apply finite_Forall2; intros; apply fin_analysis4; auto.
      
      intros mm Hmm.     
      specialize (H0 (in_tree x mm)); spec_all H0.
      rewrite analysis4_spec4; auto.
      exists mm; auto.
      destruct H0 as (a & G1 & G2 & G3).
      apply sub_in_tree_inv in G3.
      destruct G3 as [ G3 | (b & G3 & G4) ].
      right; subst; unfold V; auto.
      left; unfold U; exists b; split; auto; exists a; auto.
      
      unfold U in H; clear U V.
      apply finite_analysis in H.
      destruct H as (t & G1 & G2).
      apply in_emb_tree_upto_0 with t; auto.
      intros; apply fin_analysis4; auto.
      
      unfold V in H; clear U V.
      destruct H as (mm & G1 & G2). 
      apply Forall2_length in G1.
      destruct G2 as [ (G2 & _) | (G2 & _) ]; omega.
    Qed.

    Fact ktf_ev4_emb s1 s2 :  wfptree P1 s1 
                           -> wfptree P1 s2
                           -> emb_tree_upto k R' s1 s2 
                           -> emb_tree_upto k R (ktf_ev4 s1) (ktf_ev4 s2)
                           \/ ktf_ev4_has_disap s1.
    Proof.
      revert s1 s2.
      apply emb_wfptree_upto_ind.
 
      (* case 1 *)
      
      intros s t x ll H1 H2 Hll H IH.
      rewrite wfptree_fix in H2; destruct H2 as [ H2 H3 ].
      
      destruct IH as [ H4 | (d & H5 & H6 & H7) ].
      2: right; exists d; auto.
      
      destruct (eq_nat_dec (length ll) i) as [ Hi' | Hi' ].      
      destruct x as [ | x0 | x | x llt ].
      rewrite ktf_prop4_eq_u in H2; auto; destruct H2.
      rewrite ktf_prop4_eq_0 in H2; auto; destruct H2.
      rewrite ktf_ev4_eq_1; auto.
      left; apply in_emb_tree_upto_0 with (ktf_ev4 t); auto.
      apply in_map_iff; exists t; auto.
      rewrite ktf_ev4_eq_3; auto.
      left; apply in_emb_tree_upto_0 with (ktf_ev4 t); auto.
      rewrite intercalate_In_spec; left.
      apply in_map_iff; exists t; auto.
      destruct (eq_nat_dec (length ll) (S i)).
      rewrite ktf_prop4_eqS in H2; auto; destruct H2.
      rewrite ktf_ev4_neq; auto.
      left; apply in_emb_tree_upto_0 with (ktf_ev4 t); auto.
      apply in_map_iff; exists t; auto.
      
      (* case 2 *)
      
      intros x y ll mm H1 H2 Hll Ha Hb IH.
      
      rewrite wfptree_fix in H2; destruct H2 as [ H3 H4 ].
      rewrite wfptree_fix in H1; destruct H1 as [ H1 H2 ].
      apply Forall2_decide_ind in IH.
      destruct IH as [ H5 | H5 ].
      
      Focus 2.
      rewrite Exists2_exists_left in H5.
      destruct H5 as (_ & u & H6 & d & H7 & H8 & H9).
      right; exists d; repeat split; auto.
      apply sub_tree_trans with (1 := H9).
      constructor 2 with (1 := H6).
      apply sub_tree_refl.
      
      unfold R' in Ha.
      apply Forall2_length in Hb.
      rewrite <- Forall2_map_both in H5.
      destruct (ktf_pos4 (length ll)) as [[[H|H]|H]|H]; try omega.
      
      (* length ll < i *)
      
      do 2 (rewrite ktf_ev4_neq; try omega).
      left; apply in_emb_tree_upto_1; auto;
      rewrite map_length; auto; omega.
      
      (* length ll = i *)
      
      rewrite H in Hb; symmetry in Hb.      
      destruct x as [ | x0 | x | x llx ].
      rewrite ktf_prop4_eq_u in H1; auto; destruct H1.
      rewrite ktf_prop4_eq_0 in H1; auto; destruct H1.
      unfold Ri' in Ha.
      destruct y as [ | | y | ]; try (destruct Ha; fail).
      do 2 (rewrite ktf_ev4_eq_1; auto).
      left; apply in_emb_tree_upto_1; auto.
      rewrite map_length; omega.
      replace_with Ha; rewrite map_length; auto.
      unfold Ri' in Ha.
      destruct y as [ | | | y lly ]; try (destruct Ha; fail).
      rewrite ktf_prop4_eq_3 in H1; auto.
      rewrite ktf_prop4_eq_3 in H3; auto.
      destruct H1 as (F1 & F2 & F3 & F4).
      destruct H3 as (G1 & G2 & G3 & G4).
      do 2 (rewrite ktf_ev4_eq_3; auto).
      destruct Ha as (H6 & H7).
      unfold lift_rel in H7.
      assert (H8: Forall3 (fun t => subword (emb_tree_upto k R) 
                               cup2 (fun lx ly => exists x, In x lx /\ emb_tree_upto k R t x)) lT llx lly).
        revert H7; apply Forall3_inc.
        rewrite Forall_forall.
        intros t Ht lx ly G.
        apply subword_cup in G.
        destruct G as [ G | (u & _ & Q1 & _ & Q2) ].
        left; auto.
        right; exists u; auto.
      apply Forall3_cup in H8.
      destruct H8 as [ H8 | H8 ].

      apply Forall3_Forall2_first in H8.
      left; apply in_emb_tree_upto_1; auto.
      rewrite intercalate_length_eq, map_length; omega.
      replace_with H6.
      rewrite intercalate_length_eq, map_length; omega.
      apply Forall2_intercalate; auto.
      apply Forall2_subword_Forall2; auto.
      
      apply Exists3_Exists2_last in H8.
      destruct H8 as (_ & _ & H8).
      right.
      exists (in_tree (in_ktf_3 x llx) ll).
      split; auto.
      rewrite wfptree_fix; auto.
      rewrite ktf_prop4_eq_3; auto.
      split; auto; red; auto.
      split.
      right; split; auto.
      revert H8; apply Exists2_inc.
      intros ? ? (v & ? & ?); exists v; auto.
      apply sub_tree_refl.
      
      rewrite ktf_prop4_neq in H1; try omega.
      rewrite ktf_prop4_neq in H3; try omega.
      do 2 (rewrite ktf_ev4_neq; try omega).
      
      left; apply in_emb_tree_upto_1; auto.
      rewrite map_length; omega.
      replace_with Ha.
      rewrite map_length; auto.
      
      (* case 3 *)
      
      intros x y ll mm H1 H2 Hll Ha Hb IH.
      
      rewrite wfptree_fix in H2; destruct H2 as [ H3 H4 ].
      rewrite wfptree_fix in H1; destruct H1 as [ H1 H2 ].
      rewrite subword_sublist_eq in IH.
      destruct IH as (pp & H5 & H6).
      apply Forall2_decide_ind in H5.
      destruct H5 as [ H5 | H5 ].
      
      Focus 2.
      rewrite Exists2_exists_left in H5.
      clear H6.
      destruct H5 as (_ & u & H6 & d & H7 & H8 & H9).
      right; exists d; repeat split; auto.
      apply sub_tree_trans with (1 := H9).
      constructor 2 with (1 := H6).
      apply sub_tree_refl.
      
      rewrite <- Forall2_map_both in H5.
      apply sublist_map with (f := ktf_ev4) in H6.
      assert (subword (emb_tree_upto k R) (map ktf_ev4 ll) (map ktf_ev4 mm)) as H7.
        rewrite subword_sublist_eq.
        exists (map ktf_ev4 pp); auto.
      clear H5 H6.
            
      unfold R' in Ha.
      apply subword_length in Hb.
      destruct (ktf_pos4 k) as [ [ [ G0 | G0 ] | G0 ] | G0 ]; try omega.
      do 2 (rewrite ktf_ev4_neq; try omega).

      left; apply in_emb_tree_upto_2; auto.
      rewrite map_length; omega.
    Qed.
    
    Notation ST := (emb_tree_upto k R' <# wfptree P1 #>).

    Fact ktf_ev4_good_or_disap ll : 
                   good ST ll
                -> good (emb_tree_upto k R <# wfptree P#>) (map ktf_ev4_sig ll)
                \/ exists x, In x ll /\ ktf_ev4_has_disap (proj1_sig x).
    Proof.
      induction 1 as [ ll (t1 & Ht1) (t2 & Ht2) H1 H2 
                     | ll t1 t2 [ H1 | (d & H1 & H2) ] ].
                     
      red in H2; simpl in H2.
      apply ktf_ev4_emb in H2; auto.
      destruct H2 as [ H2 | H2 ].
      
      left; simpl map.
      constructor 1 with (ktf_ev4_sig (exist _ t2 Ht2)); auto.
      apply in_map_iff; exists (exist _ t2 Ht2); auto.
      
      right; exists (exist _ t2 Ht2); simpl; tauto.
      
      left; constructor 2; auto.
      
      right; exists d; simpl; tauto.
    Qed.

    Hypothesis HST : af ST.
    Hypothesis HST_t : af_t ST.
    
    Let HST' : bar (good ST) nil.
    Proof.
      revert HST; apply af_eq_bar_good.
    Qed.
    
    Let HST_t' : bar_t (good ST) nil.
    Proof.
      revert HST_t; apply af_t_bar_t.
    Qed.

    Notation U := (fun ws => Forall (good ST) (list_fan (map ktf_ev4_inv_sig ws))).

    Let HU : bar U nil.
    Proof.
      apply bar_map_inv with (f := ktf_ev4_inv_sig) 
                             (Q := fun ll => Forall (good ST) (list_fan ll)); auto.
      simpl; apply fan_on_list; auto.
      constructor; auto.
    Qed.
    
    Let HU_t : bar_t U nil.
    Proof.
      apply bar_t_map_inv with (f := ktf_ev4_inv_sig) 
                               (Q := fun ll => Forall (good ST) (list_fan ll)); auto.
      simpl; apply fan_t_on_list; auto.
      constructor; auto.
    Qed.
  
    Let U_to_good ws : U ws -> good (emb_tree_upto k R <# wfptree P #>) (ws ++ exist _ _ H_m_lt :: nil).
    Proof.
      set (Q ll := good (emb_tree_upto k R <# wfptree P #>) (map ktf_ev4_sig ll)).
      set (B (t : { x | wfptree P1 x }) := ktf_ev4_has_disap (proj1_sig t)).
      
      intros H0.
      destruct (@hig_combi_principle _ (map ktf_ev4_inv_sig ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
      
      (* condition of hig_combi_principle *)
      
      intros ll Hll.
      apply list_fan_spec in Hll.
      rewrite Forall_forall in H0.
      apply H0 in Hll; revert Hll.
      apply ktf_ev4_good_or_disap; auto.
      
      (* there is a good sequence p that evaluates to ws *)
      
      unfold Q in Q2.
      apply ktf_ev4_inv_sig_Forall2 in Q1.
      apply good_Restr in Q2.
      apply good_Restr.
      rewrite map_map in Q2.
      destruct Q2 as (Q2 & Q3).  
      rewrite map_app; simpl; split.
      apply good_app_right.
      replace_with Q2; auto.
      apply Forall_app; split; auto.
      replace_with Q3; auto.
      
      (* there is a fan in which every element contains a dispointing subtree *)

      unfold B in Q2.
      rewrite in_map_iff in Q1.
      destruct Q1 as ((x & Hx) & ? & Q1); subst p.
      apply good_Restr.
      rewrite map_app; simpl; split.
      apply in_split in Q1.
      destruct Q1 as (l1 & l2 & Q1).
      subst ws.
      rewrite map_app; simpl; rewrite <- app_assoc; simpl.
      apply good_app_left.
      constructor 1 with (in_tree m lT).
      apply in_or_app; right; left; auto.
      apply ktf_ev4_disap_prop2; auto. 
      intros u (Hu1 & Hu2).
      destruct (ktf_ev4_sig_inv _ Hu1 Hx Hu2) as (v & Hv1 & Hv2).
      subst u; apply (Q2 _ Hv2).
      
      apply Forall_app; split.
      apply list_sig_Forall.
      constructor; auto.
    Qed.

    Let bar_U_bar_good ws : bar U ws -> bar (good (emb_tree_upto k R <# wfptree P #> rlift exist _ _ H_m_lt)) ws.
    Proof.
      induction 1 using bar_ind.
      apply bar_lift_rel, in_bar_0; auto.
      apply in_bar_1; auto.
    Qed. 

    Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (emb_tree_upto k R <# wfptree P #> rlift exist _ _ H_m_lt)) ws.
    Proof.
      induction 1 using bar_t_rect.
      apply bar_t_lift_rel_rev, in_bar_t0; auto.
      apply in_bar_t1; auto.
    Qed.  

    Theorem ktf_ev4_af : af (emb_tree_upto k R <# wfptree P #> rlift exist _ _ H_m_lt).
    Proof.
      apply af_eq_bar_good; auto.
    Qed. 
   
    Theorem ktf_ev4_af_t : af_t (emb_tree_upto k R <# wfptree P #> rlift exist _ _ H_m_lt).
    Proof.
      apply bar_t_af_t; auto.
    Qed. 

  End ktf_ev4.

End tree_kruskal_ev4.


