(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Relations.
Require Import Wf.
Require Import Wellfounded.

Require Omega.

Require Import tacs.

Require Import rel_utils.

Set Implicit Arguments.

Section wf_basic.

  Variables (X : Type) (R : relation X).
  
  Fact wf_irrefl : well_founded R -> forall x, ~ R x x.
  Proof.
    intros HR x; induction x as [ x IHx ] using (well_founded_ind HR).
    intro; apply (IHx x); auto.
  Qed.

  Fact wf_inv : (forall P : X -> Prop, (forall a, (forall b, R b a -> P b) -> P a) 
                                     -> forall a,                             P a) 
             -> well_founded R.
  Proof.
    intros H x.
    apply H.
    intros; constructor; auto.
  Qed.

End wf_basic.

Section wf_map.

  Variable (X Y : Type) (f : X -> Y) (R : relation X) (S : relation Y).
  
  Hypothesis HRS : forall x y, R x y -> S (f x) (f y).
  
  Fact wf_map : well_founded S -> well_founded R.
  Proof.
    intros H.
    apply wf_inv.
    intros P HP.
    assert (forall y x, y = f x -> P x) as H'.
      induction y as [ y Hy ] using (well_founded_ind H).
      intros a Ha.
      apply HP.
      intros b Hb.
      apply Hy with (2 := eq_refl).
      subst; apply HRS; auto.
    intros x; apply H' with (1 := eq_refl (f x)).
  Qed.

End wf_map.
  
Section Acc_dep_sum.

  Variable (A : Type -> Type) (R : forall X, relation (A X)).
  Hypothesis HR : forall X, well_founded (@R X).

  Definition sum_dep_rel (t1 t2 : sigT A) :=
    match t1, t2 with
      | existT X a, existT Y b => exists E : X = Y, R (eq_rect X _ a _ E) b
    end.
  
  Fact wf_sum_dep_rel : well_founded sum_dep_rel.
  Proof.
    intros ( X & t ).
    generalize (HR t).
    induction 1 as [ t Ht IH ].
    constructor 1; intros (Y & y) (E & Hy).
    subst; simpl in Hy.
    auto.
  Qed.

End Acc_dep_sum.

Section Acc_restr.

  Variable (X : Type) (R : X -> X -> Prop) (A : X -> Prop).

  Section restr_wf_rect.

    Hypothesis HA : forall x, A x -> Acc (R restr A) x.

    Variables (P : X -> Type)
              (HP : forall x, A x -> (forall y, A y -> R y x -> P y) -> P x).
    
    Fact restr_wf_rect : forall x, A x -> P x.
    Proof.
      intros x Hx.
      generalize (HA Hx).
      intros H1.
      revert x H1 Hx.
      induction 1 as [ x Hx IH ]; intros H.
      apply HP.
      apply H.
      intros y H1 H2.
      apply IH; auto.
      split; auto.
    Defined.
    
  End restr_wf_rect.

  Definition restr_wf_rec HA (P : X -> Set) := @restr_wf_rect HA P.
  Definition restr_wf_ind HA (P : X -> Prop) := @restr_wf_rect HA P.

  Fact Acc_r_inv x : A x -> Acc (R restr A) x -> forall y, A y -> R y x -> Acc (R restr A) y.
  Proof.
    induction 2 as [ x Hx _ ].
    intros y H1 H2; apply Hx.
    split; auto.
  Qed.
     
   Fact Acc_r_constr x : (forall y, A y -> R y x -> Acc (R restr A) y) -> Acc (R restr A) x.
   Proof.
     intros H2; constructor.
     intros y H3; apply H2; apply H3.
   Qed.
  
  Section Acc_r_rect.

    Variables (P : X -> Type)
              (HP : forall x, (forall y, A y -> R y x -> Acc (R restr A) y)
                           -> (forall y, A y -> R y x -> P y) 
                           -> A x
                           -> P x).

    Fact Acc_r_rect x : A x -> Acc (R restr A) x -> P x.
    Proof.
      intros H1 H2; revert H2 H1.
      induction 1 as [ x Hx IH ]; intros H.
      apply HP; auto.
      intros; apply Hx; split; auto.
      intros; apply IH; auto; split; auto.
    Defined.

  End Acc_r_rect.
  
  Definition Acc_r_rec (P : X -> Set) := @Acc_r_rect P.
  Definition Acc_r_inc (P : X -> Prop) := @Acc_r_rect P.

End Acc_restr.

Definition well_founded_sub X (P : X -> Prop) (R : relation X) :=
  forall Q, (forall a, P a -> (forall b, P b -> R b a -> Q b) -> Q a) 
          -> forall a, P a                                    -> Q a. 

Section sub_wf_sub.

  Variable (X : Type) (P1 P2 : X -> Prop) (R : relation X).
  
  Fact sub_wf_sub : P1 inc1 P2 -> well_founded_sub P2 R -> well_founded_sub P1 R.
  Proof.
    intros H0 H1 Q HQ a Ha.
    generalize (H0 _ Ha) Ha.
    clear Ha.
    revert a.
    apply (H1 (fun a => P1 a -> Q a)).
    intros; auto.
  Qed.

End sub_wf_sub.

Section Fix_F_bisi.

  (* Bisimulational extensionality for Fix_F in Wf.v *)

  (* Generate the full Acc rect-induction principle *)

  Scheme Acc_rect_all := Induction for Acc Sort Type.

  Variable A : Type.
  Variable (R : A -> A -> Prop) (HR : well_founded R) (P : A -> Type).
  Variable F : forall x:A, (forall y:A, R y x -> P y) -> P x.

  Variable eq_P : forall a, P a -> P a -> Prop.

  Infix "~P" := (@eq_P _) (at level 60). 

  Variable F_ext : forall x f g, (forall z (H1 H2 : R z x), f z H1 ~P g z H2) -> F f ~P F g. 

  Let Q x (Hx : Acc R x) :=  
       forall y (Hy : Acc R y) (E : x = y), 
           eq_rect _ _ (Fix_F P F Hx) _ E ~P Fix_F P F Hy.

  Let Q_prop x (Hx : Acc R x) : Q Hx.
  Proof.
    induction Hx as [ x Hx IH ] using Acc_rect_all.
    unfold Q.
    intros y Hy E; subst; simpl.
    rewrite <- Fix_F_eq.
    apply F_ext.
    intros z H1 H2.
    apply (IH z H1 _ (Acc_inv Hy H2) eq_refl).
  Qed.

  Theorem Fix_F_bisi x (H1 H2 : Acc R x) : Fix_F P F H1 ~P Fix_F P F H2.
  Proof.
    apply (Q_prop H1 H2 eq_refl).
  Qed.

  Theorem Fix_bisi x : Fix HR _ F x ~P F (fun y (_ : R y x) => Fix HR _ F y).
  Proof.
    unfold Fix at 1.
    rewrite <- Fix_F_eq.
    apply F_ext.
    intros; apply Fix_F_bisi.
  Qed.

End Fix_F_bisi.

Section Fix_ext.

  Variable A : Type.
  Variable (R : A -> A -> Prop) (HR : well_founded R) (P : A -> Type).
  Variable F : forall x:A, (forall y:A, R y x -> P y) -> P x.
  Variable F_ext : forall x f g, (forall z (H1 H2 : R z x), f z H1 = g z H2) -> F f = F g.
 
  Theorem Fix_ext_eq x : Fix HR _ F x = F (fun (y:A) (p:R y x) => Fix HR _ F y).
  Proof.
    apply Fix_bisi with (eq_P := fun x => @eq (P x)); auto.
  Qed.

End Fix_ext.

Section lt_measure_induction.

  Variable (X : Type) (f : X -> nat).
  
  Let f_wf : well_founded (fun x y => f x < f y).
  Proof.
    apply wf_inverse_image, lt_wf.
  Qed.

  Section lt_measure_rect.

    Variable (P : X -> Type). 
    Hypothesis Prec : forall x, (forall y, f y < f x -> P y) -> P x.

    Theorem lt_measure_rect : forall x, P x.
    Proof.
      apply (Fix f_wf), Prec.
    Defined.

    Hypothesis Pext : forall x f g, (forall y H1 H2, f y H1 = g y H2) -> @Prec x f = Prec g. 

    Fact lt_measure_rect_fix x : lt_measure_rect x = Prec (fun y Hy => lt_measure_rect y).
    Proof.
      unfold lt_measure_rect.
      apply Fix_ext_eq; auto.
    Qed.

  End lt_measure_rect.

  Definition lt_measure_rec (P : X -> Set) := lt_measure_rect P.
  Definition lt_measure_ind (P : X -> Prop) := lt_measure_rect P.

End lt_measure_induction.
